#include <iostream>
#include "signal.h"
#include <gflags/gflags.h>

#include <glog/logging.h>
#include <pcl/io/pcd_io.h>
#include <pcl/common/transforms.h>

#include "system/syscontroller.h"
#include "input/sensor_subscriber.h"

#ifdef ENABLE_ROS
#include "input/ros_bridge/cloud_subscriber.h"
#include "input/ros_bridge/gnss_subscriber.h"
#include "input/ros_bridge/imu_subscriber.h"
#include "input/ros_bridge/wheel_subscriber.h"
#include "input/ros_bridge/imupose_subscriber.h"
#include "output/ros_publisher.h"
#else
#ifdef ENABLE_BAG
#include "input/rosbag/rosbag_reader.h"
#include "output/ros_publisher.h"
#else
#include "input/file_reader/cloud_receiver.h"
#include "input/file_reader/gnss_receiver.h"
#include "input/file_reader/imu_receiver.h"
#include "input/file_reader/wheel_receiver.h"
#include "input/file_reader/file_reader.h"
#include "output/netpublisher.h"
#endif
#endif

DEFINE_string(base_dir, "", "base_dir");
DEFINE_string(map_dir, "", "map_dir");

using namespace std;
using namespace juefx;

bool requestToEnd = false;
SensorSubscriber::Ptr imuSubscriber;
SensorSubscriber::Ptr gnssSubscriber;
SensorSubscriber::Ptr wheelSubscriber;
SensorSubscriber::Ptr cloudSubscriber;
SensorSubscriber::Ptr imuPoseSubscriber;

void failureHandle(const char* data, int size){
    string str = data;
    LOG(INFO) << "failureHandle: " << str;
}

void signalHandle(int sig){
    requestToEnd = true;
    LOG(INFO) << "requestToEnd";
}

void OnCloseMapCb(const vector<string>& closeMapList){

}

int main(int argc, char *argv[])
{
    gflags::ParseCommandLineFlags(&argc, &argv, true);
    google::InitGoogleLogging(argv[0]);
    google::InstallFailureSignalHandler();
    google::InstallFailureWriter(&failureHandle);
    FLAGS_alsologtostderr = true;
    FLAGS_colorlogtostderr = true;
    FLAGS_stderrthreshold = 0;
    FLAGS_logbufsecs = 0;
    if(boost::filesystem::exists(FLAGS_base_dir)){
        FLAGS_log_dir = FLAGS_base_dir + "/log";
        boost::filesystem::create_directory(FLAGS_log_dir);
    }

    string offlineResultDir = FLAGS_base_dir + "/result";
    string pcdResultDir = offlineResultDir + "/pcd";
    string basePcdResultDir = offlineResultDir + "/basePcd";
    string guessPcdResultDir = offlineResultDir + "/guessPcd";
    string rawFrameResultDir = offlineResultDir + "/raw_frame";
    if(!boost::filesystem::exists(offlineResultDir)){
        boost::filesystem::create_directory(offlineResultDir);
    }
    int unused;
    unused = system(string("rm " + offlineResultDir + "/*").c_str());
    if(!boost::filesystem::exists(pcdResultDir)){
        boost::filesystem::create_directory(pcdResultDir);
    }else{
        unused = system(string("rm " + pcdResultDir + "/*").c_str());
    }
    if(!boost::filesystem::exists(basePcdResultDir)){
        boost::filesystem::create_directory(basePcdResultDir);
    }else{
        unused = system(string("rm " + basePcdResultDir + "/*").c_str());
    }
    if(!boost::filesystem::exists(guessPcdResultDir)){
        boost::filesystem::create_directory(guessPcdResultDir);
    }else{
        unused = system(string("rm " + guessPcdResultDir + "/*").c_str());
    }
    if(!boost::filesystem::exists(rawFrameResultDir)){
        boost::filesystem::create_directory(rawFrameResultDir);
    }else{
        unused = system(string("rm " + rawFrameResultDir + "/*").c_str());
    }
    string restoreParamDir = "/home/juefx/Documents/voxels/";
    if(!boost::filesystem::exists(FLAGS_base_dir + "/imu2base.calib")){
        unused = system(string("cp " + restoreParamDir + "/imu2base.calib " + FLAGS_base_dir).c_str());
    }
    if(!boost::filesystem::exists(FLAGS_base_dir + "/init_pose.txt")){
        unused = system(string("cp " + restoreParamDir + "/init_pose.txt " + FLAGS_base_dir).c_str());
    }
    if(!boost::filesystem::exists(FLAGS_base_dir + "/imu2base_lever_arm.txt")){
        unused = system(string("cp " + restoreParamDir + "/imu2base_lever_arm.txt " + FLAGS_base_dir).c_str());
    }
    if(!boost::filesystem::exists(FLAGS_base_dir + "/eskf_init_pose.txt")){
        unused = system(string("cp " + restoreParamDir + "/eskf_init_pose.txt " + FLAGS_base_dir).c_str());
    }
    if(!boost::filesystem::exists(FLAGS_base_dir + "/qrpconfig_auto.txt")){
        unused = system(string("cp " + restoreParamDir + "/qrpconfig_auto.txt " + FLAGS_base_dir).c_str());
    }

    Parameter::Instance()->init(FLAGS_base_dir);
    State::Instance()->AddState(IMU_CALIBING);
    if(FLAGS_map_dir != ""){
        MapDetector::Instance()->HandleMapFolder(FLAGS_map_dir);
    }else{
        if(PoseGuesser::Instance()->LoadSlamTraj(FLAGS_base_dir + "/traj.txt",
                                                 FLAGS_base_dir + "/gps_traj.txt")){
            State::Instance()->AddState(TRAJ_HAS_GNSS);
        }
    }
    SysController::Ptr sysController(new SysController());
    boost::function<void(const vector<string>&)> closeMapCb
            = boost::bind(&OnCloseMapCb, _1);
    sysController->SetCloseMapCb(closeMapCb);

#ifdef ENABLE_ROS
    string topic_imu = "/imu";
    string topic_gnss = "/fix";
    string topic_wheel = "/odom";
    string topic_cloud = "/points2";
    string topic_output_clouds = "/out_pointcloud";
    string topic_output_odom = "/out_odom";
    string topic_imu_pose = "/imu_pose";
    ros::init(argc, argv, "localize");
    ros::NodeHandle nh;
    if(topic_imu.size()){
        imuSubscriber.reset(new IMUSubscriber(sysController, &nh, topic_imu));
    }
    if(topic_wheel.size()){
        wheelSubscriber.reset(new WheelSubscriber(sysController, &nh, topic_wheel));
    }
    if(topic_gnss.size()){
        gnssSubscriber.reset(new GNSSSubscriber(sysController, &nh, topic_gnss));
    }
    if(topic_cloud.size()){
        cloudSubscriber.reset(new CloudSubscriber(sysController, &nh, topic_cloud));
    }
    if(topic_cloud.size()){
        imuPoseSubscriber.reset(new IMUPoseSubscriber(sysController, &nh, topic_imu_pose));
    }
    dynamic_cast<RosPublisher*>(RosPublisher::Instance().get())->SetRosHandle(&nh);
    RosPublisher::Instance()->init(topic_output_odom, topic_output_clouds);
    ros::AsyncSpinner spinner(4);
    spinner.start();
#elif ENABLE_BAG
    string topic_output_clouds = "/out_pointcloud";
    string topic_output_odom = "/out_odom";
    ros::init(argc, argv, "localize");
    ros::NodeHandle nh;
    dynamic_cast<RosPublisher*>(RosPublisher::Instance().get())->SetRosHandle(&nh);
    RosPublisher::Instance()->init(topic_output_odom, topic_output_clouds);
    string bagPath = FLAGS_base_dir + "/locate.bag";
    RosBagReader bagReader(sysController, bagPath);
    bagReader.Play();
#else
    DataReader::Instance()->Init(FLAGS_base_dir);
    imuSubscriber.reset(new IMUReceiver(sysController));
    wheelSubscriber.reset(new WheelReceiver(sysController));
    gnssSubscriber.reset(new GNSSReceiver(sysController));
    cloudSubscriber.reset(new CloudReceiver(sysController));
    DataReader::Instance()->process();
#endif

    signal(SIGINT, signalHandle);
    while(!requestToEnd){
        usleep(10000);
    }
    spinner.stop();
    Timer::PrintAll();
    return 0;
}
