#include <string>
#include <vector>

#include <pcl/io/pcd_io.h>
#include <pcl/common/transforms.h>

#include "register/ceres_matcher.h"
#include "register/fast_ceres_matcher.h"
#include "register/icp_matcher.h"
#include "register/ndt_matcher.h"
#include "register/raw_icp_matcher.h"
#include "register/utils/util.hpp"

#define init_matcher_data(target_file, source_file)              \
    PointCloudJ::Ptr map(new PointCloudJ);                       \
    pcl::io::loadPCDFile(target_file, *map);                     \
    PointCloudJ::Ptr scan(new PointCloudJ);                      \
    PointCloudJ::Ptr result(new PointCloudJ);                    \
    pcl::io::loadPCDFile(source_file, *scan);                    \
    Eigen::Isometry3f init_pose = Eigen::Isometry3f::Identity(); \
    init_pose.translation() << 0, 0, 0;                          \
    Eigen::Isometry3f final_pose;

#define prepare_result_cloud(scan, pose, result) \
    result->points.clear();                      \
    for (auto &p : scan->points)                 \
    {                                            \
        Eigen::Vector3f v(p.x, p.y, p.z);        \
        Eigen::Vector3f new_v = pose * v;        \
        PointJ new_p;                            \
        new_p.x = new_v.x();                     \
        new_p.y = new_v.y();                     \
        new_p.z = new_v.z();                     \
        new_p.intensity = p.intensity;           \
        result->points.push_back(new_p);         \
    }

void ceres_register(std::string target_file, std::string source_file)
{
    init_matcher_data(target_file, source_file);

    juefx::CeresMatcherOption option;
    juefx::CeresMatcher matcher(option);
    matcher.LoadPointCloudMap(map, 0.05, 0.45);
    matcher.MatchPointCloud(scan, init_pose, final_pose);
    prepare_result_cloud(scan, final_pose, result);
    pcl::io::savePCDFileBinary(source_file + "_ceres_res.pcd", *result);
}

void fast_ceres_register(std::string target_file, std::string source_file, float high_res, float low_res)
{
    init_matcher_data(target_file, source_file);

    juefx::FastCeresMatcherOption option;
    juefx::FastCeresMatcher matcher(option);
    matcher.LoadPointCloudMap(map, high_res, low_res);
    matcher.MatchPointCloud(scan, init_pose, final_pose);
    prepare_result_cloud(scan, final_pose, result);
    pcl::io::savePCDFileBinary(source_file + "_fast_res.pcd", *result);
    Eigen::Isometry3f search_pose = matcher.GetLastSearchPose();
    prepare_result_cloud(scan, search_pose, result);
    pcl::io::savePCDFileBinary(source_file + "_fast_search_res.pcd", *result);
}

void fast_ceres_voxel_register(std::string target_file, std::string source_file, float high_res, float low_res)
{
    init_matcher_data(target_file, source_file);
    // add both map and scan 0.001 m
    for (auto& point : map->points) {
        point.x += 0.001;
        point.y += 0.001;
        point.z += 0.001;
    }
    for (auto &point : scan->points) {
        point.x += 0.001;
        point.y += 0.001;
        point.z += 0.001;
    }

    juefx::FastCeresMatcherOption option;
    juefx::FastCeresMatcher matcher(option);
    matcher.LoadPointCloudMap(map, high_res, low_res, 0.9999);
    matcher.MatchPointCloud(scan, init_pose, final_pose);
    prepare_result_cloud(scan, final_pose, result);
    pcl::io::savePCDFileBinary(source_file + "_fast_res.pcd", *result);
    Eigen::Isometry3f search_pose = matcher.GetLastSearchPose();
    prepare_result_cloud(scan, search_pose, result);
    pcl::io::savePCDFileBinary(source_file + "_fast_search_res.pcd", *result);
}

void icp_register(std::string target_file, std::string source_file)
{
    init_matcher_data(target_file, source_file);

    juefx::ICPMatcher icp_matcher;
    icp_matcher.LoadPointCloudMap(map);
    icp_matcher.MatchPointCloud(scan, init_pose, final_pose);
    prepare_result_cloud(scan, final_pose, result);
    pcl::io::savePCDFileBinary(source_file + "_icp_res.pcd", *result);
}

void ndt_register(std::string target_file, std::string source_file)
{
    init_matcher_data(target_file, source_file);

    juefx::NDTMatcher ndt_matcher;
    ndt_matcher.LoadPointCloudMap(map);
    ndt_matcher.MatchPointCloud(scan, init_pose, final_pose);
    prepare_result_cloud(scan, final_pose, result);
    pcl::io::savePCDFileBinary(source_file + "_ndt_res.pcd", *result);
}

void raw_icp_register(std::string target_file, std::string source_file)
{
    init_matcher_data(target_file, source_file);

    juefx::RawICPMatcher raw_icp_matcher;
    raw_icp_matcher.LoadPointCloudMap(map);
    raw_icp_matcher.MatchPointCloud(scan, init_pose, final_pose);
    prepare_result_cloud(scan, final_pose, result);
    pcl::io::savePCDFileBinary(source_file + "_raw_icp_res.pcd", *result);
}

void multi_raw_icp_register(std::string target_file_path, std::string source_file_path)
{
    std::vector<PointCloudJ::ConstPtr> target_clouds;
    std::vector<PointCloudJ::ConstPtr> source_clouds;
    std::vector<std::string> target_pcds;
    juefx::get_filenames(target_file_path, target_pcds, ".pcd");
    for (auto file : target_pcds) {
        std::cout << "read target pcd file: " << file << std::endl;
        PointCloudJ::Ptr target(new PointCloudJ);
        pcl::io::loadPCDFile(file, *target);
        target_clouds.push_back(target);
    }
    std::vector<std::string> source_pcds;
    juefx::get_filenames(source_file_path, source_pcds, ".pcd");
    for (auto file : source_pcds) {
        std::cout << "read source pcd file: " << file << std::endl;
        PointCloudJ::Ptr source(new PointCloudJ);
        pcl::io::loadPCDFile(file, *source);
        source_clouds.push_back(source);
    }
    Eigen::Isometry3f init_pose = Eigen::Isometry3f::Identity();
    init_pose.translation() << 0, 0, 0;
    Eigen::Isometry3f final_pose;
    juefx::RawICPMatcher raw_icp_matcher;
    raw_icp_matcher.MatchMultiPointCloud(target_clouds, source_clouds, init_pose, final_pose);
}

void print_usage()
{
    printf("Usage:\n");
    printf("pcd_register -i target_file input_source_file:     pcl icp register \n");
    printf("pcd_register -n target_file input_source_file:     pcl ndt register \n");
    printf("pcd_register -r target_file input_source_file:     raw icp register \n");
    printf("pcd_register -m target_file_path source_file_path: multi raw icp register \n");
    printf("pcd_register -c target_file input_source_file:     ceres register \n");
    printf("pcd_register -f1 target_file input_source_file:    fast-ceres register 0.1/0.45\n");
    printf("pcd_register -f2 target_file input_source_file:    fast-ceres register as voxel 0.09/0.36\n");
}

int main(int argc, char *argv[])
{
    if (argc < 4)
    {
        print_usage();
        return 0;
    }

    std::string func = argv[1];
    if (func[0] == '-' && func[1] == 'i')
    {
        std::string target_file = argv[2];
        std::string source_file = argv[3];
        printf("pcd_register ICP from %s to %s start...\n", source_file.c_str(), target_file.c_str());
        icp_register(target_file, source_file);
    }
    else if (func[0] == '-' && func[1] == 'n')
    {
        std::string target_file = argv[2];
        std::string source_file = argv[3];
        printf("pcd_register NDT from %s to %s start...\n", source_file.c_str(), target_file.c_str());
        ndt_register(target_file, source_file);
    }
    else if (func[0] == '-' && func[1] == 'r')
    {
        std::string target_file = argv[2];
        std::string source_file = argv[3];
        printf("pcd_register RAW ICP from %s to %s start...\n", source_file.c_str(), target_file.c_str());
        raw_icp_register(target_file, source_file);
    }
    else if (func[0] == '-' && func[1] == 'm')
    {
        std::string target_file_path = argv[2];
        std::string source_file_path = argv[3];
        printf("pcd_register MULTI RAW ICP from %s to %s start...\n", source_file_path.c_str(), target_file_path.c_str());
        multi_raw_icp_register(target_file_path, source_file_path);
    }
    else if (func[0] == '-' && func[1] == 'c')
    {
        std::string target_file = argv[2];
        std::string source_file = argv[3];
        printf("pcd_register CERES from %s to %s start...\n", source_file.c_str(), target_file.c_str());
        ceres_register(target_file, source_file);
    }
    else if (func[0] == '-' && func[1] == 'f' && func[2] == '1')
    {
        std::string target_file = argv[2];
        std::string source_file = argv[3];
        printf("pcd_register FAST-CERES 0.1/0.45 from %s to %s start...\n", source_file.c_str(), target_file.c_str());
        fast_ceres_register(target_file, source_file, 0.1, 0.45);
    }
    else if (func[0] == '-' && func[1] == 'f' && func[2] == '2')
    {
        std::string target_file = argv[2];
        std::string source_file = argv[3];
        printf("pcd_register VOXEL FAST-CERES 0.09/0.36 from %s to %s start...\n", source_file.c_str(), target_file.c_str());
        fast_ceres_voxel_register(target_file, source_file, 0.09, 0.36);
    }
    printf("process finished.\n");
    return 0;
}
