#include <iostream>
#include <glog/logging.h>
#include <pcl/io/pcd_io.h>

#include "localize/utils/pcl_point_type.h"
#include "localize/utils/timer.h"
#include "localize/utils/gnss.h"

#include "cartographer/mapping/3d/hybrid_grid.h"
#include "cartographer/mapping/internal/3d/scan_matching/interpolated_grid.h"
#include "localize/voxel_map/voxel_map_interface.h"
#include "localize/voxel_map/interpolated_voxel_map.h"

using namespace std;
using namespace cartographer::mapping;

#define HIT_SCORE 0.75
#define LOOP_TIMES 10

HybridGrid* BuildHybridGird(const PointCloudJ::Ptr pcloud_map, float resolution)
{
    HybridGrid *inner_grid = new HybridGrid(resolution);
    const std::vector<uint16_t> hit_table =
        cartographer::mapping::ComputeLookupTableToApplyOdds(cartographer::mapping::Odds(HIT_SCORE));
    for (const auto &point : pcloud_map->points)
    {
        Eigen::Vector3f position(point.x, point.y, point.z);
        const Eigen::Array3i hit_cell = inner_grid->GetCellIndex(position);
        inner_grid->ApplyLookupTable(hit_cell, hit_table);
    }
    inner_grid->FinishUpdate();
    return inner_grid;
}

void test_hybrid_grid()
{
    PointCloudJ::Ptr map(new PointCloudJ);
    pcl::io::loadPCDFile("1map.9.pcd", *map);
//    pcl::io::loadPCDFile("/home/juefx/localize_data/test_data/1map.pcd", *map);
    PointCloudJ::Ptr scan(new PointCloudJ);
    pcl::io::loadPCDFile("1base.pcd", *scan);
//    pcl::io::loadPCDFile("/home/juefx/localize_data/test_data/1base.pcd", *scan);

    float resolution = 0.09f;
    HybridGrid *grid = BuildHybridGird(map, resolution);
    scan_matching::InterpolatedGrid interpolated_grid(*grid);
    {
        double sum1 = 0;
        TIMER_START(start);
        for (int i = 0; i < LOOP_TIMES; i++)
        {
            for (auto p : scan->points)
            {
                sum1 += interpolated_grid.GetProbability((double)p.x, (double)p.y, (double)p.z);
            }
        }
        TIMER_END(start, seconds);
        printf("raw inter_grid test, cost %f, sum %f\n", seconds, sum1);
    }

    {
        double sum2 = 0;
        TIMER_START(start);
        for (int i = 0; i < LOOP_TIMES; i++)
        {
            for (auto p : scan->points)
            {
                const auto index = grid->GetCellIndex(Eigen::Vector3f(p.x, p.y, p.z));
                sum2 += grid->GetProbability(index);
                sum2 += grid->GetProbability(index + Eigen::Array3i(0, 0, 1));
                sum2 += grid->GetProbability(index + Eigen::Array3i(0, 1, 0));
                sum2 += grid->GetProbability(index + Eigen::Array3i(0, 1, 1));
                sum2 += grid->GetProbability(index + Eigen::Array3i(1, 0, 0));
                sum2 += grid->GetProbability(index + Eigen::Array3i(1, 0, 1));
                sum2 += grid->GetProbability(index + Eigen::Array3i(1, 1, 0));
                sum2 += grid->GetProbability(index + Eigen::Array3i(1, 1, 1));
            }
        }
        TIMER_END(start, seconds);
        printf("raw hybrid_grid test, cost %f, sum %f\n", seconds, sum2);
    }
}

void test_dynamic_grid()
{
    PointCloudJ::Ptr scan(new PointCloudJ);
    pcl::io::loadPCDFile("2base.pcd", *scan);
//    pcl::io::loadPCDFile("/home/juefx/localize_data/test_data/2base.pcd", *scan);

    GnssPoint init_pose;
    init_pose.B = 31.2766158865 - 0.000002;
    init_pose.L = 121.6597340747 - 0.000002;
    init_pose.H = 16.4764;
    Eigen::Vector3d gnss_rpy(0.415632, 1.681607, 249.312256);
    gnss_rpy[1] = -gnss_rpy[1];
    gnss_rpy[2] = 90 - gnss_rpy[2];
    init_pose.rpy = gnss_rpy * M_PI / 180;

    juefx::VoxelMap* voxel_map = new juefx::VoxelMap();
    voxel_map->SetMapPath(".");
//    voxel_map->SetMapPath("/home/juefx/localize_data/test_data/");

    while (!voxel_map->IsMapLoaded(init_pose))
        sleep(1);
    Eigen::Isometry3d pose_in_map;
    Eigen::Vector3d trans;
    juefx::InterpolatedVoxelMap interpolated_grid(*voxel_map, VOXEL_RESOLUTION_1X);
    {
        double sum1 = 0;
        TIMER_START(start);
        voxel_map->GuessPosition(init_pose, pose_in_map);
        for (int i = 0; i < LOOP_TIMES; i++)
        {
            for (auto p : scan->points)
            {
                double p_space, p_intensity;
                interpolated_grid.GetProbability((double)p.x, (double)p.y, (double)p.z, p_space, p_intensity);
                sum1 += p_space;
            }
        }
        voxel_map->ConfirmPosition(trans, init_pose);
        TIMER_END(start, seconds);
        printf("dynamic inter_grid test, cost %f, sum %f\n", seconds, sum1);
    }
    {
        int count[10];
        memset(count, 0, sizeof(count));
        TIMER_START(start);
        voxel_map->GuessPosition(init_pose, pose_in_map);
        for (int i = 0; i < LOOP_TIMES; i++)
        {
            for (auto p : scan->points)
            {
                const auto index = voxel_map->GetCellIndex(p.x, p.y);
                uint8_t length;
                auto *v = voxel_map->GetCellValue(VOXEL_RESOLUTION_1X, index, length);
                if (v != NULL)
                    count[length]++;
                v = voxel_map->GetCellValue(VOXEL_RESOLUTION_1X, index.Step(0, GRID_MOVE_LEFT), length);
                if (v != NULL)
                    count[length]++;
                v = voxel_map->GetCellValue(VOXEL_RESOLUTION_1X, index.Step(0, GRID_MOVE_RIGHT), length);
                if (v != NULL)
                    count[length]++;
                v = voxel_map->GetCellValue(VOXEL_RESOLUTION_1X, index.Step(0, GRID_MOVE_UP), length);
                if (v != NULL)
                    count[length]++;
            }
        }
        voxel_map->ConfirmPosition(trans, init_pose);
        TIMER_END(start, seconds);
        printf("dynamic inter_grid test, cost %f, count %d %d %d %d %d %d %d %d %d %d\n",
               seconds, count[0], count[1], count[2], count[3], count[4],
               count[5], count[6], count[7], count[8], count[9]);
    }
}

int main(int argc, char *argv[])
{
    test_hybrid_grid();
    test_dynamic_grid();
    return 0;
}
