#include <iostream>
#include <sys/resource.h>

#include <glog/logging.h>
#include <pcl/io/pcd_io.h>
#include <pcl/common/transforms.h>

#include "register/ceres_matcher.h"
#include "register/fast_ceres_matcher.h"
#include "register/icp_matcher.h"
#ifdef ENABLE_FAST_GICP
#include "register/vicp_matcher.h"
#endif

using namespace std;

#define prepare_result_cloud(scan, pose, result) \
    result->points.clear();                      \
    for (auto &p : scan->points)                 \
    {                                            \
        Eigen::Vector3f v(p.x, p.y, p.z);        \
        Eigen::Vector3f new_v = pose * v;        \
        PointJ new_p;                            \
        new_p.x = new_v.x();                     \
        new_p.y = new_v.y();                     \
        new_p.z = new_v.z();                     \
        new_p.intensity = p.intensity;           \
        result->points.push_back(new_p);         \
    }

#define init_matcher_data()                                      \
    PointCloudJ::Ptr map(new PointCloudJ);                       \
    pcl::io::loadPCDFile("1map.9.pcd", *map);                    \
    PointCloudJ::Ptr scan(new PointCloudJ);                      \
    PointCloudJ::Ptr result(new PointCloudJ);                    \
    pcl::io::loadPCDFile("1base.pcd", *scan);                    \
    Eigen::Isometry3f init_pose = Eigen::Isometry3f::Identity(); \
    init_pose.translation() << -1, 0, 0;                         \
    Eigen::Isometry3f final_pose;

void test_ceres_matcher()
{
    init_matcher_data();

    juefx::CeresMatcherOption option;
    juefx::CeresMatcher matcher(option);
    matcher.LoadPointCloudMap(map, 0.09, 0.45);
    matcher.MatchPointCloud(scan, init_pose, final_pose);
    prepare_result_cloud(scan, final_pose, result);
    pcl::io::savePCDFileBinary("1result_ceres.pcd", *result);
}

void test_fast_ceres_matcher()
{
    init_matcher_data();

    juefx::FastCeresMatcherOption option;
    juefx::FastCeresMatcher matcher(option);
    matcher.LoadPointCloudMap(map, 0.09, 0.45);
    matcher.MatchPointCloud(scan, init_pose, final_pose);
    prepare_result_cloud(scan, final_pose, result);
    pcl::io::savePCDFileBinary("1result_fastceres.pcd", *result);
}

void test_icp_matcher()
{
    init_matcher_data();

    juefx::ICPMatcher icp_matcher;
    icp_matcher.LoadPointCloudMap(map);
    icp_matcher.MatchPointCloud(scan, init_pose, final_pose);
    prepare_result_cloud(scan, final_pose, result);
    pcl::io::savePCDFileBinary("1result_icp.pcd", *result);
}

#ifdef ENABLE_FAST_GICP
void test_vicp_matcher()
{
    init_matcher_data();

    juefx::VicpMatcherOption voption;
    juefx::VicpMatcher vicp_matcher(voption);
    vicp_matcher.LoadPointCloudMap(map, 0.05);
    vicp_matcher.MatchPointCloud(scan, init_pose, final_pose);
    prepare_result_cloud(scan, final_pose, result);
    pcl::io::savePCDFileBinary("1result_vicp.pcd", *result);
}
#endif

int main(int argc, char *argv[])
{
#ifdef ENABLE_FAST_GICP
    std::cout << "ENABLE_FAST_GICP\n";
#endif

    test_ceres_matcher();
    test_fast_ceres_matcher();
//    test_icp_matcher();
#ifdef ENABLE_FAST_GICP
    test_vicp_matcher();
#endif

    return 0;
}
