/*
 * Copyright 2016 The Cartographer Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "cartographer/mapping/internal/3d/scan_matching/ceres_scan_matcher_3d.h"

#include <string>
#include <utility>
#include <vector>
#include <fstream>

#include "absl/memory/memory.h"
#include "cartographer/common/ceres_solver_options.h"
#include "cartographer/mapping/internal/3d/rotation_parameterization.h"
#include "cartographer/mapping/internal/3d/scan_matching/occupied_space_cost_function_3d.h"
#include "cartographer/mapping/internal/3d/scan_matching/rotation_delta_cost_functor_3d.h"
#include "cartographer/mapping/internal/3d/scan_matching/translation_delta_cost_functor_3d.h"
#include "cartographer/mapping/internal/optimization/ceres_pose.h"
#include "cartographer/transform/rigid_transform.h"
#include "cartographer/transform/transform.h"
#include "ceres/ceres.h"
#include "glog/logging.h"

//#define CERES_DEBUG
#define IS_CERES_MODE(x, y) ((x & y) > 0)

namespace cartographer {
namespace mapping {
namespace scan_matching {

proto::CeresScanMatcherOptions3D CreateCeresScanMatcherOptions3D(
    common::LuaParameterDictionary* const parameter_dictionary) {
  proto::CeresScanMatcherOptions3D options;
  for (int i = 0;; ++i) {
    const std::string lua_identifier =
        "occupied_space_weight_" + std::to_string(i);
    if (!parameter_dictionary->HasKey(lua_identifier)) {
      break;
    }
    options.add_occupied_space_weight(
        parameter_dictionary->GetDouble(lua_identifier));
  }
  options.set_translation_weight(
      parameter_dictionary->GetDouble("translation_weight"));
  options.set_translation_weight_z(
      parameter_dictionary->GetDouble("translation_weight_z"));
  options.set_rotation_weight(
      parameter_dictionary->GetDouble("rotation_weight"));
  options.set_rotation_weight_yaw(
      parameter_dictionary->GetDouble("rotation_weight_yaw"));
  options.set_rotation_weight_roll(
      parameter_dictionary->GetDouble("rotation_weight_roll"));
  options.set_odom_rotation_weight(
      parameter_dictionary->GetDouble("odom_rotation_weight"));
  options.set_odom_rotation_weight_yaw(
      parameter_dictionary->GetDouble("odom_rotation_weight_yaw"));
  options.set_only_optimize_yaw(
      parameter_dictionary->GetBool("only_optimize_yaw"));
  *options.mutable_ceres_solver_options() =
      common::CreateCeresSolverOptionsProto(
          parameter_dictionary->GetDictionary("ceres_solver_options").get());
  return options;
}

CeresScanMatcher3D::CeresScanMatcher3D(
    const proto::CeresScanMatcherOptions3D& options)
    : options_(options),
      ceres_solver_options_(
          common::CreateCeresSolverOptions(options.ceres_solver_options())) {
  ceres_solver_options_.linear_solver_type = ceres::DENSE_QR;
//  ceres_solver_options_.minimizer_progress_to_stdout = true;
}

Eigen::Vector4d CeresScanMatcher3D::Match(
    const Eigen::Vector3d& target_translation,
    const transform::Rigid3d& initial_pose_estimate,
    const ComplexOdometryData* ptr_odom_local,
    const std::vector<PointCloudAndHybridGridPointers>&
        point_clouds_and_hybrid_grids,
    transform::Rigid3d* const pose_estimate,
    ceres::Solver::Summary* const summary,
    int mode, int seq) {

  ceres::Problem problem;
  optimization::CeresPose ceres_pose(
      initial_pose_estimate, nullptr /* translation_parameterization */,
      options_.only_optimize_yaw()
          ? std::unique_ptr<ceres::LocalParameterization>(
                absl::make_unique<ceres::AutoDiffLocalParameterization<
                    YawOnlyQuaternionPlus, 4, 1>>())
          : std::unique_ptr<ceres::LocalParameterization>(
                absl::make_unique<ceres::QuaternionParameterization>()),
      &problem);

  CHECK_EQ(options_.occupied_space_weight_size(),
           point_clouds_and_hybrid_grids.size());
  // weiji: 2 pairs, high resolution / low resolution
  for (size_t i = 0; i != point_clouds_and_hybrid_grids.size(); ++i) {
    CHECK_GT(options_.occupied_space_weight(i), 0.);
    const sensor::PointCloud& point_cloud =
        *point_clouds_and_hybrid_grids[i].first;
    const HybridGrid& hybrid_grid = *point_clouds_and_hybrid_grids[i].second;
    // weiji: transfer points in point_cloud using translation and rotation
    //        then compute their probability in hybrid_grid 
    double scale = options_.occupied_space_weight(i) /
                std::sqrt(static_cast<double>(point_cloud.size()));
    problem.AddResidualBlock(
        OccupiedSpaceCostFunction3D::CreateAutoDiffCostFunction(
            scale, point_cloud, hybrid_grid),
        nullptr /* loss function */, ceres_pose.translation(),
        ceres_pose.rotation());
    // weiji: have tried to seperate points which are above ground
    //        but have no effect because they have been already matched
  }
  
  /*int ground_count = 0;
  {
    const sensor::PointCloud& point_cloud =
        *point_clouds_and_hybrid_grids[0].first;
    const HybridGrid& hybrid_grid = *point_clouds_and_hybrid_grids[0].second;
    for (auto& p : point_cloud) {
      if (p.position.z() < -1.3)
        ground_count++;
    }
    double scale = options_.occupied_space_weight(0) /
                   std::sqrt(static_cast<double>(ground_count));
    problem.AddResidualBlock(
        OccupiedSpaceCostFunction3D::CreateAutoDiffCostFunction(
            scale, point_cloud, hybrid_grid),
        nullptr , ceres_pose.translation(),
        ceres_pose.rotation());
  }*/
  
  double predict_scale = 1;    // influence X Y Z and yaw
  double roll_scale = 1, pitch_scale = 1;
  if (ptr_odom_local != NULL) {
    predict_scale =
        ptr_odom_local->odom.variance * ptr_odom_local->odom.variance;
    CHECK_GT(predict_scale, 0.);
  }
  // turning / rematch when bad match
  if (IS_CERES_MODE(mode, CERES_MODE_TURNING)) roll_scale *= 0.618;
  if (IS_CERES_MODE(mode, CERES_MODE_REMATCH)) roll_scale *= 0.618;
  if (IS_CERES_MODE(mode, CERES_MODE_REMATCH)) pitch_scale *= 0.618;

  CHECK_GT(options_.translation_weight(), 0.);
  CHECK_GT(options_.translation_weight_z(), 0.);
  problem.AddResidualBlock(
      TranslationDeltaCostFunctor3D::CreateAutoDiffCostFunction(
          options_.translation_weight() * predict_scale,
          options_.translation_weight_z() * predict_scale,
          target_translation),
      nullptr /* loss function */, ceres_pose.translation());
  CHECK_GT(options_.rotation_weight(), 0.);
  CHECK_GT(options_.rotation_weight_yaw(), 0.);
  problem.AddResidualBlock(
      RotationDeltaCostFunctor3D::CreateAutoDiffCostFunction(
          options_.rotation_weight_roll() * roll_scale,
          options_.rotation_weight() * pitch_scale,
          options_.rotation_weight_yaw() * predict_scale,
          initial_pose_estimate.rotation()),
      nullptr /* loss function */, ceres_pose.rotation());

  // weiji: add odom residual block
  if (ptr_odom_local != NULL && options_.odom_rotation_weight() > 0) {
    double yaw_scale =
        (ptr_odom_local->orientation_type == COMPLEX_ODOM_ORI_NO_YAW)
            ? 0
            : (ptr_odom_local->orientation_type == COMPLEX_ODOM_ORI_ODOM_YAW)? 1: 0.5;
    CHECK_GT(options_.odom_rotation_weight(), 0.);
    CHECK_GE(options_.odom_rotation_weight_yaw(), 0.);
    problem.AddResidualBlock(
        RotationDeltaCostFunctor3D::CreateAutoDiffCostFunction(
            options_.odom_rotation_weight() * roll_scale,
            options_.odom_rotation_weight() * pitch_scale,
            options_.odom_rotation_weight_yaw() * yaw_scale,
            ptr_odom_local->odom.pose.rotation()),
        nullptr /* loss function */, ceres_pose.rotation());
  }

#ifndef CERES_DEBUG
  if (IS_CERES_MODE(mode, CERES_MODE_DEBUG)) {
#endif
    double residual[5] = {0, 0, 0, 0, 0};
    const transform::Rigid3d& transform = initial_pose_estimate;
    // weiji: 2 pairs, high resolution / low resolution
    for (size_t k = 0; k != point_clouds_and_hybrid_grids.size(); ++k) {
      const auto& point_cloud = *point_clouds_and_hybrid_grids[k].first;
      const InterpolatedGrid hybrid_grid(
          *point_clouds_and_hybrid_grids[k].second);
      const double scaling_factor =
          options_.occupied_space_weight(k) /
          std::sqrt(static_cast<double>(point_cloud.size()));
      for (size_t i = 0; i < point_cloud.size(); ++i) {
        transform::Rigid3d::Vector point(point_cloud[i].position.x(),
                                         point_cloud[i].position.y(),
                                         point_cloud[i].position.z());
        const auto world = transform * point;
        const double probability =
            hybrid_grid.GetProbability(world[0], world[1], world[2]);
        double r = scaling_factor * (1. - probability);
        residual[k] += r * r;
      }
    }

    if (ptr_odom_local == NULL) std::cout << "constraint " << seq;
    std::cout << " BEFORE local ceres - high: " << residual[0]
              << " low: " << residual[1];
#ifndef CERES_DEBUG
  }
#endif

  ceres::Solve(ceres_solver_options_, &problem, summary);

  // weiji: pose_estimate means pose_observation_in_submap,
  //        final pose is "submap->local_pose() * pose_estimate"
  *pose_estimate = ceres_pose.ToRigid();

  // weiji: count low resolution loss
  double residual[5] = {0, 0, 0, 0, 0};
  const transform::Rigid3d& transform = *pose_estimate;
#ifndef CERES_DEBUG
  if (IS_CERES_MODE(mode, CERES_MODE_LOSS)) {
#endif
    // weiji: 2 pairs, high resolution / low resolution
    for (size_t k = 0; k != point_clouds_and_hybrid_grids.size(); ++k) {
      const auto& point_cloud = *point_clouds_and_hybrid_grids[k].first;
      const InterpolatedGrid hybrid_grid(
          *point_clouds_and_hybrid_grids[k].second);
      const double scaling_factor =
          options_.occupied_space_weight(k) /
          std::sqrt(static_cast<double>(point_cloud.size()));
      for (size_t i = 0; i < point_cloud.size(); ++i) {
        transform::Rigid3d::Vector point(point_cloud[i].position.x(),
                                         point_cloud[i].position.y(),
                                         point_cloud[i].position.z());
        const auto world = transform * point;
        const double probability =
            hybrid_grid.GetProbability(world[0], world[1], world[2]);
        double r = scaling_factor * (1. - probability);
        residual[k] += r * r;
      }
    }
    
    Eigen::Vector3d delta_trans = transform.translation() - target_translation;
    Eigen::Vector3d res_trans(
        options_.translation_weight() * predict_scale * delta_trans.x(),
        options_.translation_weight() * predict_scale * delta_trans.y(),
        options_.translation_weight_z() * predict_scale * delta_trans.z());
    residual[2] = res_trans.norm() * res_trans.norm();

    Eigen::Quaterniond delta =
        initial_pose_estimate.rotation().inverse() * transform.rotation();
    Eigen::Vector3d axis_angel =
        transform::RotationQuaternionToAngleAxisVector(delta);
    Eigen::Vector3d res_rotat(
        options_.rotation_weight_roll() * roll_scale * axis_angel[0],
        options_.rotation_weight() * pitch_scale * axis_angel[1],
        options_.rotation_weight_yaw() * predict_scale * axis_angel[2]);
    residual[3] = res_rotat.norm() * res_rotat.norm();

    if (ptr_odom_local != NULL) {
    double yaw_scale =
        (ptr_odom_local->orientation_type == COMPLEX_ODOM_ORI_NO_YAW)
            ? 0
            : (ptr_odom_local->orientation_type == COMPLEX_ODOM_ORI_ODOM_YAW)? 1: 0.5;
      delta = ptr_odom_local->odom.pose.rotation().inverse() * transform.rotation();
      axis_angel = transform::RotationQuaternionToAngleAxisVector(delta);
      Eigen::Vector3d res_grav(
          options_.odom_rotation_weight() * roll_scale * axis_angel[0],
          options_.odom_rotation_weight() * pitch_scale * axis_angel[1],
          options_.odom_rotation_weight_yaw() * yaw_scale * axis_angel[2]);
      residual[4] = res_grav.norm() * res_grav.norm();
    }
#ifndef CERES_DEBUG
  }
#endif

  /*double ground_residual = 0;
  {
    const sensor::PointCloud& point_cloud =
        *point_clouds_and_hybrid_grids[0].first;
    const InterpolatedGrid hybrid_grid(
        *point_clouds_and_hybrid_grids[0].second);
    double scale = options_.occupied_space_weight(0) /
                   std::sqrt(static_cast<double>(ground_count));
    for (size_t i = 0; i < point_cloud.size(); ++i) {
      const auto& pos = point_cloud[i].position;
      if (pos.z() < -1.3)
        continue;
      transform::Rigid3d::Vector point(pos.x(), pos.y(), pos.z());
      const auto world = transform * point;
      const double probability =
          hybrid_grid.GetProbability(world[0], world[1], world[2]);
      double r = scale * (1. - probability);
      ground_residual += r * r;
    }
    residual[1] = ground_residual;
  }*/

#ifndef CERES_DEBUG
  if (IS_CERES_MODE(mode, CERES_MODE_DEBUG)) {
#endif
    char buf[100];
    double prefix = seq;
    
    const auto& point_cloud = *point_clouds_and_hybrid_grids[0].first;
    snprintf(buf, 99, "%.3f_scan.txt", prefix);
    std::ofstream of_scan(buf);
    for (const auto& point : point_cloud) {
      Eigen::Vector3d p = transform * point.position.cast<double>();
      snprintf(buf, 99, "%.3f %.3f %.3f\n", p.x(), p.y(), p.z());
      of_scan << buf;
    }
    of_scan.close();

    const HybridGrid& hybrid_grid = *point_clouds_and_hybrid_grids[0].second;
    snprintf(buf, 99, "%.3f_grid.txt", prefix);
    std::ofstream of_grid(buf);
    for (const auto it : hybrid_grid) {
      float probability = ValueToProbability(it.second);
      if (probability > 0.5) {
        Eigen::Vector3f p = hybrid_grid.GetCenterOfCell(it.first);
        snprintf(buf, 99, "%.3f %.3f %.3f\n", p.x(), p.y(), p.z());
        of_grid << buf;
      }
    }
    of_grid.close();

    std::cout << " - detail(" << predict_scale << "/" << mode
              << ") - high: " << residual[0] << " low: " << residual[1]
              << " trans: " << residual[2] << " rotat: " << residual[3]
              << " grav: " << residual[4] << std::endl;
#ifndef CERES_DEBUG
  }
#endif

  return Eigen::Vector4d(residual[0], residual[1], residual[2], residual[3] + residual[4]);
}

}  // namespace scan_matching
}  // namespace mapping
}  // namespace cartographer
