/*
 * Copyright 2016 The Cartographer Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "cartographer/sensor/point_cloud.h"

#include "cartographer/sensor/proto/sensor.pb.h"
#include "cartographer/transform/transform.h"

namespace cartographer {
namespace sensor {

PointCloud TransformPointCloud(const PointCloud& point_cloud,
                               const transform::Rigid3f& transform) {
  PointCloud result;
  result.reserve(point_cloud.size());
  for (const RangefinderPoint& point : point_cloud) {
    result.emplace_back(transform * point);
  }
  return result;
}

TimedPointCloud TransformTimedPointCloud(const TimedPointCloud& point_cloud,
                                         const transform::Rigid3f& transform) {
  TimedPointCloud result;
  result.reserve(point_cloud.size());
  for (const TimedRangefinderPoint& point : point_cloud) {
    result.push_back(transform * point);
  }
  return result;
}

PointCloud CropPointCloud(const PointCloud& point_cloud, const float min_z,
                          const float max_z) {
  PointCloud cropped_point_cloud;
  for (const RangefinderPoint& point : point_cloud) {
    if (min_z <= point.position.z() && point.position.z() <= max_z) {
      cropped_point_cloud.push_back(point);
    }
  }
  return cropped_point_cloud;
}

TimedPointCloud CropTimedPointCloud(const TimedPointCloud& point_cloud,
                                    const float min_z, const float max_z) {
  TimedPointCloud cropped_point_cloud;
  for (const TimedRangefinderPoint& point : point_cloud) {
    if (min_z <= point.position.z() && point.position.z() <= max_z) {
      cropped_point_cloud.push_back(point);
    }
  }
  return cropped_point_cloud;
}

///////////////////////////////////////////////////////////////////////////
// Copy from LOAM

#define LOAM_LABEL_INVALID 30000
#define LOAM_GROUND_TRUE 1
#define LOAM_GROUND_FALSE -1
#define LOAM_GROUND_UNKOWN 0
struct LoamPoint {
  float x;
  float y;
  float z;
  float range;
  int8_t is_ground;
  uint8_t row_idn;
  uint16_t column_idn;
  int16_t label;
};

using LoamPointCloud = std::vector<LoamPoint>;

LoamProjection::LoamProjection(int lidar_type) {
  if (lidar_type == LOAM_LIDAR_TYPE_40P) {
    N_SCAN = 40;
    HORIZON_SCAN = 1800;
    N_GROUND_SCAN = 28;
    N_SCAN_ANGLE = SCAN_ANGLE_40P;
  } else {
    // Default VLP-32E
    N_SCAN = 32;
    HORIZON_SCAN = 1808;
    N_GROUND_SCAN = 24;
    N_SCAN_ANGLE = SCAN_ANGLE_32E;
  }
  HORIZON_ANGLE = 360.0f / HORIZON_SCAN;
  for (size_t i = 0; i < N_SCAN; i++)
    N_SCAN_RAD[i] = N_SCAN_ANGLE[i] / 180.0 * M_PI;
  HORIZON_RAD = HORIZON_ANGLE / 180.0 * M_PI;

  size = N_SCAN * HORIZON_SCAN;
  projCloud = new LoamPoint[size];
  queueRow = new uint16_t[size];
  queueCol = new uint16_t[size];

  std::pair<int8_t, int8_t> neighbor;
  neighbor.first = -1; neighbor.second =  0; neighbors.push_back(neighbor);
  neighbor.first =  0; neighbor.second =  1; neighbors.push_back(neighbor);
  neighbor.first =  0; neighbor.second = -1; neighbors.push_back(neighbor);
  neighbor.first =  1; neighbor.second =  0; neighbors.push_back(neighbor);
}

LoamProjection::~LoamProjection() {
  delete projCloud;
  delete queueRow;
  delete queueCol;
}

// range image projection
void LoamProjection::ProjectPointCloud(const TimedPointCloud& laserCloudIn) {
  size_t cloudSize = laserCloudIn.size();
  for (size_t i = 0; i < cloudSize; ++i) {
    LoamPoint thisPoint;
    memset(&thisPoint, 0, sizeof(thisPoint));
    thisPoint.x = laserCloudIn[i].position.x();
    thisPoint.y = laserCloudIn[i].position.y();
    thisPoint.z = laserCloudIn[i].position.z();

    // find the row and column index in the image for this point
    int ring = (int)laserCloudIn[i].ring;
    thisPoint.row_idn = 39 - ring;
    thisPoint.column_idn =
        (uint16_t)round((laserCloudIn[i].ring - ring) * 10000);
    if (thisPoint.column_idn >= HORIZON_SCAN)
      thisPoint.column_idn -= HORIZON_SCAN;
    if (thisPoint.row_idn < 0 || thisPoint.column_idn < 0 ||
        thisPoint.column_idn >= HORIZON_SCAN)
      continue;

    thisPoint.range = sqrt(thisPoint.x * thisPoint.x + thisPoint.y * thisPoint.y + thisPoint.z * thisPoint.z);
    if (thisPoint.range < SENSOR_MIN_RANGE)
      continue;
    
    size_t index = thisPoint.column_idn + thisPoint.row_idn * HORIZON_SCAN;
    projCloud[index] = thisPoint;
  }
}

void LoamProjection::GroundRemoval() {
  // set ground
  for (size_t j = 0; j < HORIZON_SCAN; ++j) {
    for (size_t i = 0; i < N_GROUND_SCAN - 1; ++i) {
      size_t lowerInd = j + i * HORIZON_SCAN;
      size_t upperInd = j + (i + 1) * HORIZON_SCAN;
      if (projCloud[lowerInd].range < SENSOR_MIN_RANGE ||
          projCloud[upperInd].range < SENSOR_MIN_RANGE)
        continue;

      float diffX = projCloud[upperInd].x - projCloud[lowerInd].x;
      float diffY = projCloud[upperInd].y - projCloud[lowerInd].y;
      float diffZ = projCloud[upperInd].z - projCloud[lowerInd].z;
      float angle =
          atan2(diffZ, sqrt(diffX * diffX + diffY * diffY)) * 180 / M_PI;

      if (abs(angle - sensorMountAngle) <= SEGMENT_GROUND_ANGLE) {
        projCloud[lowerInd].is_ground = LOAM_GROUND_TRUE;
        projCloud[upperInd].is_ground = LOAM_GROUND_TRUE;
      } else {
        if (projCloud[lowerInd].is_ground == LOAM_GROUND_UNKOWN)
          projCloud[lowerInd].is_ground = LOAM_GROUND_FALSE;
        if (projCloud[upperInd].is_ground == LOAM_GROUND_UNKOWN)
          projCloud[upperInd].is_ground = LOAM_GROUND_FALSE;
      }
    }
  }
}

#define Loam_Enqueue(ind, row, col)  \
  projCloud[ind].label = labelCount; \
  lineCount[row]++;                  \
  queueRow[queueEndInd] = row;       \
  queueCol[queueEndInd] = col;       \
  ++queueSize;                       \
  ++queueEndInd;                     \
  ++totalSize;

void LoamProjection::LabelComponents(int row, int col) {
  // use std::queue std::vector std::deque will slow the program down greatly
  int lineCount[LOAM_MAX_RING] = {0};
  queueRow[0] = row;
  queueCol[0] = col;
  int totalSize = 1;
  int queueSize = 1;
  int queueStartInd = 0;
  int queueEndInd = 1;

  while (queueSize > 0) {
    // Pop point
    int fromRow = queueRow[queueStartInd];
    int fromCol = queueCol[queueStartInd];
    --queueSize;
    ++queueStartInd;
    // Mark popped point
    size_t fromInd = fromCol + fromRow * HORIZON_SCAN;
    projCloud[fromInd].label = labelCount;
    // Loop through all the neighboring grids of popped grid
    for (auto iter = neighbors.begin(); iter != neighbors.end(); ++iter) {
      // step for jumping one point
      int step = 1;
      // for searching local plane 
      float angle1 = 0, angle2 = 0;
      size_t firstInd = 0;
      int firstRow = 0, firstCol = 0;
      while (step <= 2) {
        // new index
        int thisRow = fromRow + (*iter).first * step;
        int thisCol = fromCol + (*iter).second * step;
        // index should be within the boundary
        if (thisRow < 0 || thisRow >= (int)N_SCAN) break;
        // at range image margin (left or right side)
        if (thisCol < 0) thisCol += HORIZON_SCAN;
        if (thisCol >= (int)HORIZON_SCAN) thisCol -= HORIZON_SCAN;
        // prevent infinite loop (caused by put already examined point back)
        size_t thisInd = thisCol + thisRow * HORIZON_SCAN;
        if (projCloud[thisInd].label != 0) break;

        float d1 = std::max(projCloud[fromInd].range, projCloud[thisInd].range);
        float d2 = std::min(projCloud[fromInd].range, projCloud[thisInd].range);
        float alpha = ((*iter).first == 0)
                          ? HORIZON_RAD * step
                          : abs(N_SCAN_RAD[fromRow] - N_SCAN_RAD[thisRow]);
        float angle = atan2(d2 * sin(alpha), (d1 - d2 * cos(alpha)));
        // for searching local plane
        if (step == 1) {
          firstInd = thisInd;
          firstRow = thisRow;
          firstCol = thisCol;
          angle1 = angle;
        }
        if (step == 2) angle2 = angle;\
        // enqueue condition
        if (angle > SEGMENT_SAME_LABEL) {
          Loam_Enqueue(thisInd, thisRow, thisCol);
          break;
        } else if (thisRow == fromRow && angle1 > 0 && angle2 > 0 &&
                   abs(angle2 - angle1) < SEGMENT_HORIZON_PLANE) {
          Loam_Enqueue(firstInd, firstRow, firstCol);
          Loam_Enqueue(thisInd, thisRow, thisCol);
          break;
        }
        step++;
      }
    }
  }

  // check if this segment is valid
  bool validSegment = false;
  if (totalSize >= SEGMENT_VALID_NUM) {
    validSegment = true;
  } else {
    int totalLine = 0;
    for (size_t i = 0; i < N_SCAN; ++i) {
      if (lineCount[i] > SEGMENT_VALID_MIN_WIDTH) {
        validSegment = true;
        break;
      }
      if (lineCount[i] > 0) ++totalLine;
    }
    if (totalLine >= SEGMENT_VALID_MIN_LINE) validSegment = true;
  }
  // segment is valid, add total labelCount
  if (validSegment == true) {
    ++labelCount;
  } else {  // segment is invalid, mark these points
    for (int i = 0; i < totalSize; ++i) {
      size_t ind = queueCol[i] + queueRow[i] * HORIZON_SCAN;
      projCloud[ind].label = LOAM_LABEL_INVALID;
    }
  }
}

int LoamProjection::FilterCloud(TimedPointCloud& laserCloud,
                                const float highResRange) {
  memset(projCloud, 0, sizeof(LoamPoint) * size);
  labelCount = 1;

  ProjectPointCloud(laserCloud);
  GroundRemoval();

  size_t ind = 0;
  for (size_t i = 0; i < N_SCAN; ++i) {
    for (size_t j = 0; j < HORIZON_SCAN; ++j, ++ind) {
      if (projCloud[ind].is_ground == LOAM_GROUND_TRUE ||
          projCloud[ind].range < SENSOR_MIN_RANGE)
        projCloud[ind].label = -1;
    }
  }
  ind = 0;
  for (size_t i = 0; i < N_SCAN; ++i) {
    for (size_t j = 0; j < HORIZON_SCAN; ++j, ++ind) {
      if (projCloud[ind].label == 0) LabelComponents(i, j);
    }
  }

  laserCloud.clear();
  int featurePointCount = 0;
  ind = 0;
  for (size_t i = 0; i < N_SCAN; ++i) {
    for (size_t j = 0; j < HORIZON_SCAN; ++j, ++ind) {
      const LoamPoint& point = projCloud[ind];
      if (point.is_ground == LOAM_GROUND_TRUE ||
          (point.label > 0 && point.label < LOAM_LABEL_INVALID)) {
        TimedRangefinderPoint new_point;
        new_point.position = Eigen::Vector3f(point.x, point.y, point.z);
        new_point.time = 0; new_point.ring = 0;
        laserCloud.push_back(new_point);
        if (point.is_ground != LOAM_GROUND_TRUE && point.range < highResRange)
          featurePointCount++;
      }
    }
  }
  return featurePointCount;
}

}  // namespace sensor
}  // namespace cartographer
