#include "ESKF.h"
#include "unrolledFPFt.h"
#include "utils.h"

using namespace Eigen;
using namespace std;

ESKF::ESKF(Eigen::Vector3f a_gravity,
           const Eigen::Matrix<float, STATE_SIZE, 1> &initialState,
           const Eigen::Matrix<float, dSTATE_SIZE, dSTATE_SIZE> &initalP,
           float var_acc, float var_omega, float var_acc_bias, float var_omega_bias,
           Eigen::Vector3f sigma_pos, Eigen::Vector3f sigma_theta,
           Eigen::Vector3f sigma_vel, int lidar_status, float wheel_vel_range,
           int delayHandling, int bufferL, Eigen::Vector3f i2b_lever_arm,
           const QRPConfigAutoData* config_data
           )
    : var_acc_(var_acc),
      var_omega_(var_omega),
      var_acc_bias_(var_acc_bias),
      var_omega_bias_(var_omega_bias),
      a_gravity_(a_gravity),
      nominalState_(initialState),
      P_(initalP)
{
    // Jacobian of the state transition: page 59, eqn 269
    // Precompute constant part only
    F_x_.setZero();
    // dPos row
    F_x_.block<3, 3>(dPOS_IDX, dPOS_IDX) = I_3;
    // dVel row
    F_x_.block<3, 3>(dVEL_IDX, dVEL_IDX) = I_3;
    // dTheta row
    // dAccelBias row
    F_x_.block<3, 3>(dAB_IDX, dAB_IDX) = I_3;
    // dGyroBias row
    F_x_.block<3, 3>(dGB_IDX, dGB_IDX) = I_3;

    // how to handle delayed messurements.
    delayHandling_ = delayHandling;
    sigma_lidar_pos.setZero();
    sigma_wheel_vel.setZero();
    sigma_lidar_theta.setZero();
    
    sigma_wheel_vel.diagonal() = sigma_vel;
    sigma_lidar_pos.diagonal() = sigma_pos;
    sigma_lidar_theta.diagonal() = sigma_theta;

    // qrp params assignment
    qrp_config_data = config_data;
    monitor_gyro_flag = false;
    last_predict_ = 0;  

    //initial imu2base lever arm
     imu2base_lever_arm = i2b_lever_arm;

    //initial imu to zero
    acc_m = Vector3f(0,0,0);
    gyro_m = Vector3f(0,0,0);


#ifdef ENABLE_DELAY_HANDLE
    bufferL_ = bufferL;
    recentPtr = 0;
    firstMeasTime = lTime(INT32_MAX, INT32_MAX);

    //handle time delay methods
    if (delayHandling_ == larsonAverageIMU || delayHandling_ == larsonFull)
    {
        //init circular buffer for IMU
        imuHistoryPtr_ = new std::vector<imuMeasurement>(bufferL_);
        PHistoryPtr_ = new std::vector<std::pair<lTime, Eigen::Matrix<float, dSTATE_SIZE, dSTATE_SIZE>>>(bufferL);
        for (int i = 0; i < bufferL; i++)
        {
            imuHistoryPtr_->at(i).time = lTime(0, 0);
        }
        Mptr = new Eigen::Matrix<float, dSTATE_SIZE, dSTATE_SIZE>;
    }
    if (delayHandling_ == larsonNewestIMU)
    {
        //init newest value
        lastImu_.time = lTime(0, 0);
        Mptr = new Eigen::Matrix<float, dSTATE_SIZE, dSTATE_SIZE>;
    }
    if (delayHandling_ == applyUpdateToNew || delayHandling_ == larsonAverageIMU)
    {
        //init circular buffer for state
        stateHistoryPtr_ = new std::vector<std::pair<lTime, Eigen::Matrix<float, STATE_SIZE, 1>>>(bufferL_);
        for (int i = 0; i < bufferL; i++)
        {
            stateHistoryPtr_->at(i).first = lTime(0, 0);
        }
    }
#endif
}

Matrix<float, STATE_SIZE, 1> ESKF::makeState(
    const Vector3f &p,
    const Vector3f &v,
    const Quaternionf &q,
    const Vector3f &a_b,
    const Vector3f &omega_b)
{
    Matrix<float, STATE_SIZE, 1> out;
    out << p, v, quatToHamilton(q).normalized(), a_b, omega_b;
    return out;
}

Matrix<float, dSTATE_SIZE, dSTATE_SIZE> ESKF::makeP(
    const Matrix3f &cov_pos,
    const Matrix3f &cov_vel,
    const Matrix3f &cov_dtheta,
    const Matrix3f &cov_a_b,
    const Matrix3f &cov_omega_b)
{
    Matrix<float, dSTATE_SIZE, dSTATE_SIZE> P;
    P.setZero();
    P.block<3, 3>(dPOS_IDX, dPOS_IDX) = cov_pos;
    P.block<3, 3>(dVEL_IDX, dVEL_IDX) = cov_vel;
    P.block<3, 3>(dTHETA_IDX, dTHETA_IDX) = cov_dtheta;
    P.block<3, 3>(dAB_IDX, dAB_IDX) = cov_a_b;
    P.block<3, 3>(dGB_IDX, dGB_IDX) = cov_omega_b;
    return P;
}

Matrix3f ESKF::getDCM()
{
    return getQuat().matrix();
}

Quaternionf ESKF::quatFromHamilton(const Vector4f &qHam)
{
    return Quaternionf(
        (Vector4f() << qHam.block<3, 1>(1, 0), // x, y, z
         qHam.block<1, 1>(0, 0)                // w
         )
            .finished());
}

Vector4f ESKF::quatToHamilton(const Quaternionf &q)
{
    return (Vector4f() << q.coeffs().block<1, 1>(3, 0), // w
            q.coeffs().block<3, 1>(0, 0)                // x, y, z
            )
        .finished();
}

Matrix3f ESKF::getSkew(const Vector3f &in)
{
    Matrix3f out;
    out << 0, -in(2), in(1),
        in(2), 0, -in(0),
        -in(1), in(0), 0;
    return out;
}

Matrix3f ESKF::rotVecToMat(const Vector3f &in)
{
    float angle = in.norm();
    Vector3f axis = (angle == 0) ? Vector3f(1, 0, 0) : in.normalized();
    AngleAxisf angAx(angle, axis);
    return angAx.toRotationMatrix();
}

Quaternionf ESKF::rotVecToQuat(const Vector3f &in)
{
    float angle = in.norm();
    Vector3f axis = (angle == 0) ? Vector3f(1, 0, 0) : in.normalized();
    return Quaternionf(AngleAxisf(angle, axis));
}

Vector3f ESKF::quatToRotVec(const Quaternionf &q)
{
    AngleAxisf angAx(q);
    return angAx.angle() * angAx.axis();
}

void ESKF::predictIMU(const Vector3f &acc,
                      const Vector3f &gryo,
                      const double dt,
#ifdef ENABLE_DELAY_HANDLE
                      lTime stamp,
#endif
                      lTime now)
{
    //recentPtr ++;
#ifdef ENABLE_DELAY_HANDLE
    //handle time delay methods
    if (delayHandling_ == larsonAverageIMU || delayHandling_ == larsonFull)
    {
        //store the imu data for later.

        imuMeasurement thisMeas;
        thisMeas.time = stamp;
        thisMeas.acc = acc;
        thisMeas.gyro = gryo;
        imuHistoryPtr_->at(recentPtr % bufferL_) = thisMeas;
    }
    if (delayHandling_ == larsonNewestIMU)
    {
        //store only the newest imu
        imuMeasurement thisMeas;
        thisMeas.time = stamp;
        thisMeas.acc = acc;
        thisMeas.gyro = gryo;
        lastImu_ = thisMeas;
    }
#endif
    // DCM of current state
    Matrix3f Rot = getDCM();
    // Accelerometer measurement
    Vector3f acc_body = acc - getAccelBias();
    Vector3f acc_global = Rot * acc_body;
    // Gyro measruement
    Vector3f omega = gryo - getGyroBias();
    Vector3f delta_theta = omega * dt;
    // accumulate yaw if the vehicle started
    if (monitor_gyro_flag) {
        gyro_data.push_back(omega[2]);
        if (gyro_data.size() > GYRO_CACHE_MAX){
            gyro_data.pop_front();
        }
    }
    Quaternionf q_delta_theta = rotVecToQuat(delta_theta);
    Matrix3f R_delta_theta = q_delta_theta.toRotationMatrix();

    // Nominal state kinematics (eqn 259, pg 58)
    Vector3f delta_pos = getVel() * dt + 0.5f * (acc_global + a_gravity_) * dt * dt;
    nominalState_.block<3, 1>(POS_IDX, 0) += delta_pos;
    nominalState_.block<3, 1>(VEL_IDX, 0) += (acc_global + a_gravity_) * dt;
    nominalState_.block<4, 1>(QUAT_IDX, 0) = quatToHamilton(getQuat() * q_delta_theta).normalized();

    // // Jacobian of the state transition (eqn 269, page 59)
    // // Update dynamic parts only
    // // dPos row
    // F_x_.block<3, 3>(dPOS_IDX, dVEL_IDX).diagonal().fill(dt); // = I_3 * _dt
    // // dVel row
    // F_x_.block<3, 3>(dVEL_IDX, dTHETA_IDX) = -Rot * getSkew(acc_body) * dt;
    // F_x_.block<3, 3>(dVEL_IDX, dAB_IDX) = -Rot * dt;
    // // dTheta row
    // F_x_.block<3, 3>(dTHETA_IDX, dTHETA_IDX) = R_delta_theta.transpose();
    // F_x_.block<3, 3>(dTHETA_IDX, dGB_IDX).diagonal().fill(-dt); // = -I_3 * dt;

    // Predict P and inject variance (with diagonal optimization)
    // P_ = F_x_*P_*F_x_.transpose();

    Matrix<float, dSTATE_SIZE, dSTATE_SIZE> Pnew;
    unrolledFPFt(P_, Pnew, dt,
                 -Rot * getSkew(acc_body) * dt,
                 -Rot * dt,
                 R_delta_theta.transpose());
    P_ = Pnew;

    // Inject process noise
    P_.diagonal().block<3, 1>(dVEL_IDX, 0).array() += var_acc_ * SQ(dt);
    P_.diagonal().block<3, 1>(dTHETA_IDX, 0).array() += var_omega_ * SQ(dt);
    P_.diagonal().block<3, 1>(dAB_IDX, 0).array() += var_acc_bias_ * dt;
    P_.diagonal().block<3, 1>(dGB_IDX, 0).array() += var_omega_bias_ * dt;

#ifdef ENABLE_DELAY_HANDLE
    if (delayHandling_ == applyUpdateToNew || delayHandling_ == larsonAverageIMU)
    {
        //store state for later.
        std::pair<lTime, Eigen::Matrix<float, STATE_SIZE, 1>> thisState;
        thisState.first = stamp;
        thisState.second = nominalState_;
        stateHistoryPtr_->at(recentPtr % bufferL_) = thisState;
    }
    if (delayHandling_ == larsonAverageIMU)
    {
        std::pair<lTime, Eigen::Matrix<float, dSTATE_SIZE, dSTATE_SIZE>> thisP;
        thisP.first = stamp;
        thisP.second = P_;
        PHistoryPtr_->at(recentPtr % bufferL_) = thisP;
    }
#endif
}

// eqn 280, page 62
Matrix<float, 4, 3> ESKF::getQ_dtheta()
{
    Vector4f qby2 = 0.5f * getQuatVector();
    // Assing to letters for readability. Note Hamilton order.
    float w = qby2[0];
    float x = qby2[1];
    float y = qby2[2];
    float z = qby2[3];
    Matrix<float, 4, 3> Q_dtheta;
    Q_dtheta << -x, -y, -z,
        w, -z, y,
        z, w, -x,
        -y, x, w;
    return Q_dtheta;
}

#ifdef ENABLE_DELAY_HANDLE
//get best time from history of state
int ESKF::getClosestTime(std::vector<std::pair<lTime, Eigen::Matrix<float, STATE_SIZE, 1>>> *ptr, lTime stamp)
{
    //we find the first time in the history that is older, or take the oldest one if the buffer does not extend far enough
    int complete = 0;
    int index = recentPtr;
    while (!complete)
    {
        if (ptr->at(index % bufferL_).first <= stamp)
        {
            if (!ptr->at(index % bufferL_).first.isZero())
                return index % bufferL_;

            else
            {
                return recentPtr % bufferL_;
            }
        }
        index--; // scroll back in time.
        if (index <= recentPtr - bufferL_)
            complete = 1;
    }
    return recentPtr % bufferL_;
}

//get best time from history of imu
int ESKF::getClosestTime(std::vector<imuMeasurement> *ptr, lTime stamp)
{
    //we find the first time in the history that is older, or take the oldest one if the buffer does not extend far enough
    int complete = 0;
    int index = recentPtr;
    while (!complete)
    {
        if (ptr->at(index % bufferL_).time <= stamp)
        {
            if (!ptr->at(index % bufferL_).time.isZero())
                return index % bufferL_;

            else
            {
                return recentPtr % bufferL_;
            }
        }
        index--; // scroll back in time.
        if (index <= recentPtr - bufferL_)
            complete = 1;
    }
    return recentPtr % bufferL_;
}
#endif

void ESKF::measurePos(const Vector3f &pos_meas,
                      const Matrix3f &pos_covariance,
#ifdef ENABLE_DELAY_HANDLE
                      lTime stamp,
#endif
                      lTime now)
{
    Vector3f delta_pos;
    if (delayHandling_ == noMethod || delayHandling_ == larsonAverageIMU)
    {
        delta_pos = pos_meas - getPos();
        //        cout << "noMethod delta Pos: "<< delta_pos <<endl;
    }

#ifdef ENABLE_DELAY_HANDLE
    // delta measurement
    if (firstMeasTime == lTime(INT32_MAX, INT32_MAX))
        firstMeasTime = now;

    if (delayHandling_ == applyUpdateToNew)
    {
        if (lastMeasurement < stateHistoryPtr_->at((recentPtr + 1) % bufferL_).first)
            firstMeasTime = now;
        if (stamp > firstMeasTime)
        {
            int bestTimeIndex = getClosestTime(stateHistoryPtr_, stamp);
            delta_pos = pos_meas - stateHistoryPtr_->at(bestTimeIndex).second.block<3, 1>(POS_IDX, 0);
        }
        else
            delta_pos = pos_meas - getPos();
        cout << "UpToNew Pos: " << delta_pos << endl;
    }
    if (delayHandling_ == larsonAverageIMU)
    {
        if (lastMeasurement < imuHistoryPtr_->at((recentPtr + 1) % bufferL_).time)
            firstMeasTime = now;
    }
#endif
    //lastMeasurement = now;
    // H is a trivial observation of purely the position
    Matrix<float, 3, dSTATE_SIZE> H;
    H.setZero();
    H.block<3, 3>(0, dPOS_IDX) = I_3;

    // Apply update
    update_3D(delta_pos,
              pos_covariance,
              H,
#ifdef ENABLE_DELAY_HANDLE
              lTime stamp,
#endif
              now);
}

void ESKF::measureQuat(const Quaternionf &q_gb_meas,
                       const Matrix3f &theta_covariance,
#ifdef ENABLE_DELAY_HANDLE
                       lTime stamp,
#endif
                       lTime now,
                       measureStatus measure_status_theta)
{
    // Transform the quaternion measurement to a measurement of delta_theta:
    // a rotation in the body frame from nominal to measured.
    // This is identical to the form of dtheta in the error_state,
    // so this becomes a trivial measurement of dtheta.
    Quaternionf q_gb_nominal = getQuat();
#ifdef ENABLE_DELAY_HANDLE
    if (firstMeasTime == lTime(INT32_MAX, INT32_MAX))
        firstMeasTime = now;
    if (delayHandling_ == noMethod || delayHandling_ == larsonAverageIMU)
    {
        q_gb_nominal = getQuat();
    }
    if (delayHandling_ == applyUpdateToNew)
    {
        if (stamp > firstMeasTime)
        {
            int bestTimeIndex = getClosestTime(stateHistoryPtr_, stamp);
            q_gb_nominal = quatFromHamilton(stateHistoryPtr_->at(bestTimeIndex).second.block<4, 1>(QUAT_IDX, 0));
        }
        else
            q_gb_nominal = getQuat();
    }
#endif
    Quaternionf q_bNominal_bMeas = q_gb_nominal.conjugate() * q_gb_meas;
    Vector3f delta_theta = quatToRotVec(q_bNominal_bMeas);
    // Because of the above construction, H is a trivial observation of dtheta
    Matrix<float, 3, dSTATE_SIZE> H;
    H.setZero();
    Eigen::Matrix3f H_ = Eigen::Matrix3f::Zero();
    switch(measure_status_theta)
    {
        case THETA_Roll: H_(0,0) = 1;break;
        case THETA_Pitch: H_(1,1) = 1;break;
        case THETA_Yaw: H_(2,2) = 1;break;
        case THETA_ALL: H_ = I_3;break;
    }
    H.block<3, 3>(0, dTHETA_IDX) = H_;

    // Apply update
    update_3D(delta_theta, theta_covariance, H, now);
}

void ESKF::measureVel(const Vector3f &vel_meas,
                      const Eigen::Matrix3f &vel_covariance,
#ifdef ENABLE_DELAY_HANDLE
                      lTime stamp,
#endif
                      lTime now)
{
    // delta measurement
    Matrix3f Rot = getDCM();
    Vector3f delta_vel;
    Vector3f H_theta;
    Matrix3f H_delta_theta;
    //以 base的前左上为坐标系，imu投影在base系的杆臂
    Vector3f omega = gyro_m - getGyroBias();
    Vector3f wheel_imu_meas = vel_meas + getSkew(omega) * imu2base_lever_arm;
    delta_vel = wheel_imu_meas - Rot.transpose() * getVel(); //local系下的误差 

    // H is a trivial observation of purely the velocity
    Matrix<float, 3, dSTATE_SIZE> H;
    H.setZero();
    H.block<3, 3>(0, dVEL_IDX) = Rot.transpose(); //这是对delta_v的H阵系数
    if (fabs(vel_meas[0]) > LOW_SPEED_AS_STOP)    // FIXME 低速时角度更新
    {
        H_theta = Rot.transpose() * getVel();
        H_delta_theta = getSkew(H_theta);
        H.block<3, 3>(0, dTHETA_IDX) = H_delta_theta; //这是对delta_theta的H阵系数
    }
    // Apply update
    update_3D(delta_vel,
              vel_covariance,
              H,
#ifdef ENABLE_DELAY_HANDLE
              lTime stamp,
#endif
              now);
}

void ESKF::update_3D(const Vector3f &delta_measurement,
                     const Matrix3f &meas_covariance,
                     const Matrix<float, 3, dSTATE_SIZE> &H,
#ifdef ENABLE_DELAY_HANDLE
                     lTime stamp,
#endif
                     lTime now)
{
    //generate M matrix for time correction methods
    int bestTimeIndex;
    int normalPass = 1;
#ifdef ENABLE_DELAY_HANDLE
    if (delayHandling_ == larsonAverageIMU)
    {
        if (stamp > firstMeasTime)
            normalPass = 0;
    }
    if (delayHandling_ == larsonAverageIMU && !normalPass)
    {
        imuMeasurement avMeas = getAverageIMU(stamp);
        float dt = (now - stamp).toSec();
        Vector3f acc_body = avMeas.acc - getAccelBias();
        Vector3f omega = avMeas.gyro - getGyroBias();
        Vector3f delta_theta = omega * dt;
        Quaternionf q_delta_theta = rotVecToQuat(delta_theta);
        Matrix3f R_delta_theta = q_delta_theta.toRotationMatrix();
        bestTimeIndex = getClosestTime(stateHistoryPtr_, stamp);

        Matrix3f Rot = quatFromHamilton(stateHistoryPtr_->at(bestTimeIndex).second.block<4, 1>(QUAT_IDX, 0)).matrix();
        // dPos row
        F_x_.block<3, 3>(dPOS_IDX, dVEL_IDX).diagonal().fill(dt); // = I_3 * _dt
        // dVel row
        F_x_.block<3, 3>(dVEL_IDX, dTHETA_IDX) = -Rot * getSkew(acc_body) * dt;
        F_x_.block<3, 3>(dVEL_IDX, dAB_IDX) = -Rot * dt;
        // dTheta row
        F_x_.block<3, 3>(dTHETA_IDX, dTHETA_IDX) = R_delta_theta.transpose();
        F_x_.block<3, 3>(dTHETA_IDX, dGB_IDX).diagonal().fill(-dt); // = -I_3 * dt;
    }
#endif
    // Kalman gain
    Matrix<float, dSTATE_SIZE, 3> PHt = P_ * H.transpose();
    Matrix<float, dSTATE_SIZE, 3> K;
    if ((delayHandling_ == noMethod || delayHandling_ == applyUpdateToNew || delayHandling_ == larsonAverageIMU))
    {
        K = PHt * (H * PHt + meas_covariance).inverse();
    }
#ifdef ENABLE_DELAY_HANDLE
    if (delayHandling_ == larsonAverageIMU && !normalPass)
    {
        K = F_x_ * K;
    }
#endif
    // Correction error state
    // measure_vel is locally frame, but K*(y-H*xt) will tranform the body to global,because  PHt*y is global,PHt*H*xt is global frame
    Matrix<float, dSTATE_SIZE, 1> errorState = K * delta_measurement;
    // Update P (simple form)
    // P = (I_dx - K*H)*P;
    // Update P (Joseph form)
    Matrix<float, dSTATE_SIZE, dSTATE_SIZE> I_KH = I_dx - K * H;
    if (delayHandling_ == noMethod || delayHandling_ == applyUpdateToNew)
    {
        P_ = I_KH * P_ * I_KH.transpose() + K * meas_covariance * K.transpose();
    }
#ifdef ENABLE_DELAY_HANDLE
    if (delayHandling_ == larsonAverageIMU && !normalPass)
    {

        P_ = P_ - K * H * PHistoryPtr_->at(bestTimeIndex).second * F_x_;
    }
#endif

    injectErrorState(errorState);
}
#ifdef ENABLE_DELAY_HANDLE
ESKF::imuMeasurement ESKF::getAverageIMU(lTime stamp)
{
    Vector3f accelAcc(0, 0, 0);
    Vector3f gyroAcc(0, 0, 0);
    int complete = 0;
    int index = recentPtr;
    int count = 0;
    while (!complete)
    {
        if (imuHistoryPtr_->at(index % bufferL_).time >= stamp)
        {
            if (!imuHistoryPtr_->at(index % bufferL_).time.isZero())
            {
                //should acc
                accelAcc += imuHistoryPtr_->at(index % bufferL_).acc;
                gyroAcc += imuHistoryPtr_->at(index % bufferL_).gyro;
                count++;
            }
        }
        else
        {
            break;
        }
        index--; // scroll back in time.
        if (index <= recentPtr - bufferL_)
            complete = 1;
    }
    accelAcc = accelAcc / count;
    gyroAcc = gyroAcc / count;
    ESKF::imuMeasurement ret;
    ret.acc = accelAcc;
    ret.gyro = gyroAcc;
    ret.time = imuHistoryPtr_->at(index % bufferL_).time;
    return ret;
}
#endif

void ESKF::injectErrorState(const Matrix<float, dSTATE_SIZE, 1> &error_state)
{
    // Inject error state into nominal state (eqn 282, pg 62)
    nominalState_.block<3, 1>(POS_IDX, 0) += error_state.block<3, 1>(dPOS_IDX, 0);
    nominalState_.block<3, 1>(VEL_IDX, 0) += error_state.block<3, 1>(dVEL_IDX, 0);
    Vector3f dtheta = error_state.block<3, 1>(dTHETA_IDX, 0);
    Quaternionf q_dtheta = rotVecToQuat(dtheta);
    nominalState_.block<4, 1>(QUAT_IDX, 0) = quatToHamilton(getQuat() * q_dtheta).normalized();
    nominalState_.block<3, 1>(AB_IDX, 0) += error_state.block<3, 1>(dAB_IDX, 0);
    nominalState_.block<3, 1>(GB_IDX, 0) += error_state.block<3, 1>(dGB_IDX, 0);

    // Reflect this tranformation in the P matrix, aka ESKF Reset
    // Note that the document suggests that this step is optional
    // eqn 287, pg 63
    Matrix3f G_theta = I_3 - getSkew(0.5f * dtheta);
    P_.block<3, 3>(dTHETA_IDX, dTHETA_IDX) =
        G_theta * P_.block<3, 3>(dTHETA_IDX, dTHETA_IDX) * G_theta.transpose();
}

Matrix<float, STATE_SIZE, 1> ESKF::getNominalState()
{
    return nominalState_;
}

void ESKF::measureLidar(const Quaternionf &quat, 
                        const Vector3f &pos_meas, 
                        int status, 
                        lTime now,
                        measureStatus measure_status_theta)
{
    if (last_predict_ > 0)
    {
        float dt;
        dt = now.toSec() - last_predict_;
        if (fabs(wheel_vel_[0]) > LOW_SPEED_AS_STOP) {
            predictIMU(acc_m, gyro_m, dt, now);
            //////// FIXME outlier cleaning ////////
            Vector3f predict_pos = nominalState_.block<3, 1>(POS_IDX, 0);
            Vector3f predict_theta = quatToRotVec(getQuat());
            Vector3f meas_theta = quatToRotVec(quat);
            double drift_pos_xy = sqrt(pow(pos_meas[0] - predict_pos[0], 2) + pow(pos_meas[1] - predict_pos[1], 2));
            double drift_theta_yaw = abs(abs(predict_theta[2]) - abs(meas_theta[2])) * 180 / M_PI;
            char buf[128];
            snprintf(buf, 128, "drift_pos_xy %.5f, drift_theta_yaw %.5f", drift_pos_xy, drift_theta_yaw);
            cout << "drift: " << buf << endl;
            if (status != LIDAR_QUALITY_HIGH and (drift_pos_xy > LOW_DRIFT_POS_AS_OUTLIER or drift_theta_yaw > LOW_DRIFT_THETA_AS_OUTLIER))
            {
                // reset lidar smoothing cache if the predict pos/theta drifted
                lidar_smooth_data = std::deque<std::tuple<Eigen::VectorXf, int>>();
                return;
            }
        }
        last_predict_ = now.toSec();

        // update R according to yaw accumulation
        updateR(status);
        if (fabs(wheel_vel_[0]) > LOW_SPEED_AS_STOP and status != LIDAR_QUALITY_LOW)
        {
            //////// FIXME smooth lidar measurement ////////
            std::tuple<Eigen::Quaternionf, Eigen::Vector3f> smooth_meas = smoothMeasurement(quat, pos_meas, status, now);
            Vector3f smooth_pos = get<1>(smooth_meas);
            Quaternionf smooth_quat = get<0>(smooth_meas);
            // ESKF measure and update
            measurePos(smooth_pos, sigma_lidar_pos, now);
            measureQuat(smooth_quat, sigma_lidar_theta, now, measure_status_theta);
        }
        else
            {
            // reset lidar smoothing cache if the vehicle stoped
            lidar_smooth_data = std::deque<std::tuple<Eigen::VectorXf, int>>();
        }       
       
#ifdef ESKF_DEBUG
        char buf[128];
        snprintf(buf, 128, "time %.5f, quat [ w %.6f x %.6f y %.6f z %.6f ]", last_predict_, quat.w(), quat.x(), quat.y(), quat.z());
        cout << "measureLidar() at " << buf << " pos " << pos_meas.transpose() << ", final state: " << nominalState_.transpose() << endl;
#endif
    }
}

void ESKF::measureWheel(const Vector3f &wheel_vel, 
                        double speed_range, 
                        lTime now)
{
    //////// FIXME imu to base yaw error ////////
    static Eigen::AngleAxisf roll_(AngleAxisd(0,Vector3d::UnitX()));
    static Eigen::AngleAxisf pitch_(AngleAxisd(0,Vector3d::UnitY()));
    static Eigen::AngleAxisf yaw_(AngleAxisd(-1.1/57.29,Vector3d::UnitZ()));   //there is a yaw error about 1 degree between base and wheel
    static Eigen::Matrix3f rot_wheel_base(yaw_ * pitch_ * roll_); 
    
    if (last_predict_ > 0)
    {
        wheel_vel_ = wheel_vel;
        if (fabs(wheel_vel_[0]) > LOW_SPEED_AS_STOP) {
            double dt = now.toSec() - last_predict_;
            predictIMU(acc_m, gyro_m, dt, now);
        }  
          
        last_predict_ = now.toSec();
        if (fabs(wheel_vel_[0]) < LOW_SPEED_AS_STOP) {
            wheel_vel_[0] = 0;
        }
        Vector3f rot_wheel_vel = rot_wheel_base * wheel_vel_;
        measureVel(rot_wheel_vel, sigma_wheel_vel, now);
        
    
#ifdef ESKF_DEBUG
        char buf[128];
        snprintf(buf, 128, "time %.5f, wheel_vel %.6f", last_predict_, wheel_vel[0]);
        cout << "measureWheel() at " << buf << ", final state: " << nominalState_.transpose() << endl;
#endif
    }
}

void ESKF::updateImu(const Eigen::Vector3f &acc, const Eigen::Vector3f &gryo, lTime now)
{
    if (last_predict_ > 0)
    {
        double dt;
        acc_m = acc;
        gyro_m = gryo;
        dt = now.toSec() - last_predict_;

        if (fabs(wheel_vel_[0]) > LOW_SPEED_AS_STOP) {
            predictIMU(acc_m, gyro_m, dt, now);
        }
    }
    last_predict_ = now.toSec();

#ifdef ESKF_DEBUG
    char buf[128];
    snprintf(buf, 128, "time %.5f, acc_m [x %.6f y %.6f z %.6f], gyr_m [x %.6f y %.6f z %.6f]", last_predict_,
             acc_m[0], acc_m[1], acc_m[2], gyro_m[0], gyro_m[1], gyro_m[2]);
    cout << "updateImu() at " << buf << ", final state: " << nominalState_.transpose() << endl;
#endif
}

void ESKF::updateNoImu(lTime now)
{
    if (last_predict_ > 0)
    {
        double dt;
        dt = now.toSec() - last_predict_;
        if (fabs(wheel_vel_[0]) > LOW_SPEED_AS_STOP) {
            predictIMU(acc_m, gyro_m, dt, now);
        }  
    }
    last_predict_ = now.toSec();

#ifdef ESKF_DEBUG
    char buf[128];
    snprintf(buf, 128, "time %.5f, acc_m [x %.6f y %.6f z %.6f], gyr_m [x %.6f y %.6f z %.6f]", last_predict_,
             acc_m[0], acc_m[1], acc_m[2], gyro_m[0], gyro_m[1], gyro_m[2]);
    cout << "updateImu() at " << buf << ", final state: " << nominalState_.transpose() << endl;
#endif
}

void ESKF::resetR(int status)
{
    if (status == LIDAR_QUALITY_HIGH){
    Eigen::Matrix3f new_sigma_lidar_pos;
    new_sigma_lidar_pos.setZero();
    Vector3f new_sigma_lidar_pos_(pow(qrp_config_data->sigma_pos_x[0],2),pow(qrp_config_data->sigma_pos_y[0],2),pow(qrp_config_data->sigma_pos_z[0],2));
    new_sigma_lidar_pos.diagonal() = new_sigma_lidar_pos_;
    sigma_lidar_pos = new_sigma_lidar_pos;
    
    Eigen::Matrix3f new_sigma_lidar_theta;
    new_sigma_lidar_theta.setZero();
    Vector3f new_sigma_lidar_theta_(pow(qrp_config_data->sigma_theta_roll[0],2),pow(qrp_config_data->sigma_theta_pitch[0],2),pow(qrp_config_data->sigma_theta_yaw[0],2));
    new_sigma_lidar_theta.diagonal() = new_sigma_lidar_theta_;
    sigma_lidar_theta = new_sigma_lidar_theta;
        
    Eigen::Matrix3f new_sigma_wheel_vel;
    new_sigma_wheel_vel.setZero();
    Vector3f new_sigma_wheel_vel_(pow(qrp_config_data->sigma_vel_x[0],2),pow(qrp_config_data->sigma_vel_y[0],2),pow(qrp_config_data->sigma_vel_z[0],2));
    new_sigma_wheel_vel.diagonal() = new_sigma_wheel_vel_;
    sigma_wheel_vel = new_sigma_wheel_vel;
    }else if (status == LIDAR_QUALITY_LOW){
        Eigen::Matrix3f new_sigma_lidar_pos;
        new_sigma_lidar_pos.setZero();
        Vector3f new_sigma_lidar_pos_(pow(qrp_config_data->sigma_pos_x[1],2),pow(qrp_config_data->sigma_pos_y[1],2),pow(qrp_config_data->sigma_pos_z[1],2));
        new_sigma_lidar_pos.diagonal() = new_sigma_lidar_pos_;
        sigma_lidar_pos = new_sigma_lidar_pos;
        
        Eigen::Matrix3f new_sigma_lidar_theta;
        new_sigma_lidar_theta.setZero();
        Vector3f new_sigma_lidar_theta_(pow(qrp_config_data->sigma_theta_roll[1],2),pow(qrp_config_data->sigma_theta_pitch[1],2),pow(qrp_config_data->sigma_theta_yaw[1],2));
        new_sigma_lidar_theta.diagonal() = new_sigma_lidar_theta_;
        sigma_lidar_theta = new_sigma_lidar_theta;
            
        Eigen::Matrix3f new_sigma_wheel_vel;
        new_sigma_wheel_vel.setZero();
        Vector3f new_sigma_wheel_vel_(pow(qrp_config_data->sigma_vel_x[1],2),pow(qrp_config_data->sigma_vel_y[1],2),pow(qrp_config_data->sigma_vel_z[1],2));
        new_sigma_wheel_vel.diagonal() = new_sigma_wheel_vel_;
        sigma_wheel_vel = new_sigma_wheel_vel;
    }
}

void ESKF::updateR(int status)
    {
    if (status != LIDAR_QUALITY_MEDIUM){
        resetR(status);
    }
    else if (monitor_gyro_flag && qrp_config_data != NULL){
        // get average gyro 
        double sum_gyro = 0.0;
        int gyro_data_size = gyro_data.size();
        for(int i = 0; i < gyro_data_size; i++) {
            double gyro_ = gyro_data.front();
            sum_gyro += gyro_;
            gyro_data.pop_front();
        }
        double avg_gyro = gyro_data_size == 0 ? 0.0 : sum_gyro / gyro_data_size;

        /* f(x) = (R2-R1)/(1+e^k(a-x)) + R1 */
        double new_sigma_pos_x = (qrp_config_data->sigma_pos_x[1] - qrp_config_data->sigma_pos_x[0]) / (1.0+exp(qrp_config_data->sigma_pos_x[3] * (qrp_config_data->sigma_pos_x[2] - abs(avg_gyro)))) + qrp_config_data->sigma_pos_x[0];
        double old_sigma_pos_x = sqrt(sigma_lidar_pos.diagonal()[0]);
        new_sigma_pos_x = 0.5 * new_sigma_pos_x + 0.5 * old_sigma_pos_x;

        double new_sigma_pos_y = (qrp_config_data->sigma_pos_y[1] - qrp_config_data->sigma_pos_y[0]) / (1.0+exp(qrp_config_data->sigma_pos_y[3] * (qrp_config_data->sigma_pos_y[2] - abs(avg_gyro)))) + qrp_config_data->sigma_pos_y[0];
        double old_sigma_pos_y = sqrt(sigma_lidar_pos.diagonal()[1]);
        new_sigma_pos_y = 0.5 * new_sigma_pos_y + 0.5 * old_sigma_pos_y;

        double new_sigma_pos_z = (qrp_config_data->sigma_pos_z[1] - qrp_config_data->sigma_pos_z[0]) / (1.0+exp(qrp_config_data->sigma_pos_z[3] * (qrp_config_data->sigma_pos_z[2] - abs(avg_gyro)))) + qrp_config_data->sigma_pos_z[0];
        
        double new_sigma_theta_roll = (qrp_config_data->sigma_theta_roll[1] - qrp_config_data->sigma_theta_roll[0]) / (1+exp(qrp_config_data->sigma_theta_roll[3] * (qrp_config_data->sigma_theta_roll[2] - abs(avg_gyro)))) + qrp_config_data->sigma_theta_roll[0];
        double new_sigma_theta_pitch = (qrp_config_data->sigma_theta_pitch[1] - qrp_config_data->sigma_theta_pitch[0]) / (1+exp(qrp_config_data->sigma_theta_pitch[3] * (qrp_config_data->sigma_theta_pitch[2] - abs(avg_gyro)))) + qrp_config_data->sigma_theta_pitch[0];
        double new_sigma_theta_yaw = (qrp_config_data->sigma_theta_yaw[1] - qrp_config_data->sigma_theta_yaw[0]) / (1+exp(qrp_config_data->sigma_theta_yaw[3] * (qrp_config_data->sigma_theta_yaw[2] - abs(avg_gyro)))) + qrp_config_data->sigma_theta_yaw[0];
        
        double new_sigma_vel_x = (qrp_config_data->sigma_vel_x[1] - qrp_config_data->sigma_vel_x[0]) / (1+exp(qrp_config_data->sigma_vel_x[3] * (qrp_config_data->sigma_vel_x[2] - abs(avg_gyro)))) + qrp_config_data->sigma_vel_x[0];
        double new_sigma_vel_y = (qrp_config_data->sigma_vel_y[1] - qrp_config_data->sigma_vel_y[0]) / (1+exp(qrp_config_data->sigma_vel_y[3] * (qrp_config_data->sigma_vel_y[2] - abs(avg_gyro)))) + qrp_config_data->sigma_vel_y[0];
        double new_sigma_vel_z = (qrp_config_data->sigma_vel_z[1] - qrp_config_data->sigma_vel_z[0]) / (1+exp(qrp_config_data->sigma_vel_z[3] * (qrp_config_data->sigma_vel_z[2] - abs(avg_gyro)))) + qrp_config_data->sigma_vel_z[0];
        
        Eigen::Matrix3f new_sigma_lidar_pos;
        new_sigma_lidar_pos.setZero();
        Vector3f new_sigma_lidar_pos_(pow(new_sigma_pos_x,2),pow(new_sigma_pos_y,2),pow(new_sigma_pos_z,2));
        new_sigma_lidar_pos.diagonal() = new_sigma_lidar_pos_;
        sigma_lidar_pos = new_sigma_lidar_pos;
                    
        Eigen::Matrix3f new_sigma_lidar_theta;
        new_sigma_lidar_theta.setZero();
        Vector3f new_sigma_lidar_theta_(pow(new_sigma_theta_roll,2),pow(new_sigma_theta_pitch,2),pow(new_sigma_theta_yaw,2));
        new_sigma_lidar_theta.diagonal() = new_sigma_lidar_theta_;
        sigma_lidar_theta = new_sigma_lidar_theta;
        
        Eigen::Matrix3f new_sigma_wheel_vel;
        new_sigma_wheel_vel.setZero();
        Vector3f new_sigma_wheel_vel_(pow(new_sigma_vel_x,2),pow(new_sigma_vel_y,2),pow(new_sigma_vel_z,2));
        new_sigma_wheel_vel.diagonal() = new_sigma_wheel_vel_;
        sigma_wheel_vel = new_sigma_wheel_vel;
    } else {
        monitor_gyro_flag = true;
    }
}

std::tuple<Eigen::Quaternionf,Eigen::Vector3f> ESKF::smoothMeasurement(const Eigen::Quaternionf &quat,const Eigen::Vector3f &pos_meas,int quality_level,lTime now)
{
    Eigen::VectorXf pos_angle_meas(6);
    Eigen::Vector3f angle_meas = RotationQuaternionToEulerVector(quat);
    pos_angle_meas.block<3,1>(0,0) = pos_meas;
    pos_angle_meas.block<3,1>(3,0) = angle_meas;
    std::tuple<Eigen::VectorXf,int> data(pos_angle_meas,quality_level);
    lidar_smooth_data.push_back(data);
    while (lidar_smooth_data.size() > LIDAR_CACHE_MAX){
        lidar_smooth_data.pop_front();
    }
    
    if (quality_level != LIDAR_QUALITY_HIGH and lidar_smooth_data.size() == LIDAR_CACHE_MAX){
        Eigen::Vector2f smoothXY = polynomialFitting(lidar_smooth_data,SMOOTH_POS_Y);
        float smoothZ = polynomialFitting(lidar_smooth_data,SMOOTH_POS_Z)[1];
        float smoothRoll = polynomialFitting(lidar_smooth_data,SMOOTH_THETA_Roll)[1];
        float smoothPitch = polynomialFitting(lidar_smooth_data,SMOOTH_THETA_Pitch)[1];
        float smoothYaw = polynomialFitting(lidar_smooth_data,SMOOTH_THETA_Yaw)[1];
        
        Eigen::Quaternionf smoothQuat;
        Eigen::AngleAxisf roll(AngleAxisd(smoothRoll,Vector3d::UnitX()));
        Eigen::AngleAxisf pitch(AngleAxisd(smoothPitch,Vector3d::UnitY()));
        Eigen::AngleAxisf yaw(AngleAxisd(smoothYaw,Vector3d::UnitZ())); 
        smoothQuat = yaw*pitch*roll;
        Eigen::Vector3f smoothPos(smoothXY[0],smoothXY[1],smoothZ);
        return std::tuple<Eigen::Quaternionf,Eigen::Vector3f> (smoothQuat,smoothPos);
    }else{
        return std::tuple<Eigen::Quaternionf,Eigen::Vector3f> (quat,pos_meas);
    }
}

Eigen::Vector2f ESKF::polynomialFitting(const std::deque<std::tuple<Eigen::VectorXf,int>> data, smoothStatus smooth_status)
{
    std::vector<float> X,Y;
    std::vector<int> status;
    Eigen::Quaternionf quat = getQuat();
    Eigen::Vector3f euler_angle = RotationQuaternionToEulerVector(quat);
    Eigen::Vector3f origin_pos = get<0>(data.back()).block<3,1>(0,0);
    Eigen::Vector3f origin_angle = get<0>(data.back()).block<3,1>(3,0);
    Eigen::Matrix3f rotation = Eigen::Matrix3f::Identity();
    bool same_sign_angle_flag = true;
    int window = 1;
    
    for (size_t i = 0; i < data.size(); i++){
        Eigen::VectorXf pos_angle = get<0>(data[i]);
        Eigen::Vector3f pos = pos_angle.block<3,1>(0,0);
        Eigen::Vector3f angle = pos_angle.block<3,1>(3,0);
        
        int quality_status = get<1>(data[i]);
        int weight = 0;
        int index = 1;
        switch (smooth_status){
            case SMOOTH_POS_Y: {
                index = 1;
                window = LIDAR_SMOOTH_WINDOW_POS;
                static Eigen::AngleAxisf roll(AngleAxisd(0,Vector3d::UnitX()));
                static Eigen::AngleAxisf pitch(AngleAxisd(0,Vector3d::UnitY()));
                Eigen::AngleAxisf yaw(AngleAxisd(-euler_angle[2],Vector3d::UnitZ())); 
                rotation = roll*pitch*yaw;
                pos_angle.block<3,1>(0,0) = rotation * (pos - origin_pos); 
                break;
            }
            case SMOOTH_POS_Z: {
                index = 2;
                window = LIDAR_SMOOTH_WINDOW_POS;
                static Eigen::AngleAxisf roll(AngleAxisd(0,Vector3d::UnitX()));
                Eigen::AngleAxisf pitch(AngleAxisd(-euler_angle[1],Vector3d::UnitY()));
                static Eigen::AngleAxisf yaw(AngleAxisd(0,Vector3d::UnitZ()));
                rotation = roll*pitch*yaw;
                pos_angle.block<3,1>(0,0) = rotation * (pos - origin_pos);
                break;
            }
            case SMOOTH_THETA_Roll: {
                index = 3;
                window = LIDAR_SMOOTH_WINDOW_THETA;;
                pos_angle[0] = i * 0.1;
                if (angle[0] * origin_angle[0] < 0) same_sign_angle_flag = false;
                break;
            }
            case SMOOTH_THETA_Pitch: {
                index = 4;
                window = LIDAR_SMOOTH_WINDOW_THETA;
                pos_angle[0] = i * 0.1;
                if (angle[1] * origin_angle[1] < 0) same_sign_angle_flag = false;
                break;
            }
            case SMOOTH_THETA_Yaw: {
                index = 5;
                window = LIDAR_SMOOTH_WINDOW_THETA;
                pos_angle[0] = i * 0.1;
                if (angle[2] * origin_angle[2] < 0) same_sign_angle_flag = false;
                break;
            }
            default: break;
        }
        switch (quality_status){
            case LIDAR_QUALITY_HIGH: weight = LIDAR_SMOOTH_QUALITY_HIGH_WEIGHT;break;
            case LIDAR_QUALITY_MEDIUM: weight = LIDAR_SMOOTH_QUALITY_MEDIUM_WEIGHT;break;
            case LIDAR_QUALITY_LOW: weight = LIDAR_SMOOTH_QUALITY_LOW_WEIGHT;break;
            default: break;
        }
        for (int j = 0; j < weight; j++){
            X.push_back(pos_angle[0]);
            Y.push_back(pos_angle[index]);
        }
    }
    
    // if the angle list contains different sign value, return an original angle
    if (!same_sign_angle_flag or window == 1){
        return Eigen::Vector2f (X.back(),Y.back());
    }
    
    // get sample data in the window
    vector<float> windowX(X.end() - window, X.end());
    vector<float> windowY(Y.end() - window, Y.end());

    // map sample data from STL vector to eigen vector
    Eigen::Map<Eigen::VectorXf> sampleX(windowX.data(), windowX.size());
    Eigen::Map<Eigen::VectorXf> sampleY(windowY.data(), windowY.size());

    Eigen::MatrixXf mtxVandermonde(windowX.size(), LIDAR_SMOOTH_DEGREE + 1);  // Vandermonde matrix of X-axis coordinate vector of sample data
    Eigen::VectorXf colVandermonde = sampleX;                           // Vandermonde column

    // construct Vandermonde matrix column by column
    for (size_t i = 0; i < LIDAR_SMOOTH_DEGREE + 1; ++i)
    {
        if (0 == i)
        {
            mtxVandermonde.col(0) = Eigen::VectorXf::Constant(windowX.size(), 1, 1);
            continue;
        }
        if (1 == i)
        {
            mtxVandermonde.col(1) = colVandermonde;
            continue;
        }
        colVandermonde = colVandermonde.array() * sampleX.array();
        mtxVandermonde.col(i) = colVandermonde;
    }
    
    // calculate coefficients vector of fitted polynomial
    // Eigen::VectorXf k = (mtxVandermonde.transpose()*mtxVandermonde).inverse()*(mtxVandermonde.transpose())*sampleY;
    Eigen::VectorXf k = mtxVandermonde.colPivHouseholderQr().solve(sampleY);
    
    // fix lastPos/Quat
    float lastX = X.back();
    float lastY = Y.back();
    float fitX = lastX;
    float fitY = 0.0;
    for (int n = 0; n < k.size(); n++){
        fitY += k[n] * pow(fitX,n);
    }
    
    // Pos/Quat's rotation from base to map
    Eigen::Vector2f ret;
    Eigen::Vector3f rot_fit_pos;
    switch (smooth_status){
        case SMOOTH_POS_Y: {
            Eigen::Vector3f fit_pos_xy(fitX,fitY,0);
            rot_fit_pos = rotation.transpose() * fit_pos_xy + origin_pos;
            ret << rot_fit_pos[0],rot_fit_pos[1];
            break;
        }
        case SMOOTH_POS_Z:{
            Eigen::Vector3f fit_pos_xz(fitX,0,fitY);
            rot_fit_pos = rotation.transpose() * fit_pos_xz + origin_pos;
            ret << rot_fit_pos[0],rot_fit_pos[2];
            break;
        }
        case SMOOTH_THETA_Roll:case SMOOTH_THETA_Pitch:case SMOOTH_THETA_Yaw: ret << fitX,fitY ;break;
        default: break;
    }
    
    return ret;
}

void ESKF::getFullState(FullState &state)
{
    state.current_time = last_predict_;
    state.makeP = P_;
    state.makeState = nominalState_;
}

/*void ESKF::setFullState(const FullState &state)
{
    last_predict_ = state.current_time;
    P_ = state.makeP;
    nominalState_ = state.makeState;
}*/
