#ifndef ESKF_H
#define ESKF_H

// Malloc is really bad on embedded platform
//#define EIGEN_NO_MALLOC
#include <Eigen/Core>
#include <Eigen/Geometry>
#include <iostream>
#include <vector>
#include <deque>
#include <math.h>

#include "lTime.h"
#include "qrpconfig.h"

//#define ESKF_DEBUG
#define SUPPORT_STDIOSTREAM

#define POS_IDX (0)
#define VEL_IDX (POS_IDX + 3)
#define QUAT_IDX (VEL_IDX + 3)
#define AB_IDX (QUAT_IDX + 4)
#define GB_IDX (AB_IDX + 3)
#define STATE_SIZE (GB_IDX + 3)

#define dPOS_IDX (0)
#define dVEL_IDX (dPOS_IDX + 3)
#define dTHETA_IDX (dVEL_IDX + 3)
#define dAB_IDX (dTHETA_IDX + 3)
#define dGB_IDX (dAB_IDX + 3)
#define dSTATE_SIZE (dGB_IDX + 3)

#define SQ(x) (x * x)
#define I_3 (Eigen::Matrix3f::Identity())
#define I_dx (Eigen::Matrix<float, dSTATE_SIZE, dSTATE_SIZE>::Identity())

#define GYRO_CACHE_MAX 100
#define LIDAR_CACHE_MAX 30
#define LIDAR_SMOOTH_DEGREE 2
#define LIDAR_SMOOTH_WINDOW_POS 15
#define LIDAR_SMOOTH_WINDOW_THETA 8
#define LIDAR_SMOOTH_QUALITY_HIGH_WEIGHT 2
#define LIDAR_SMOOTH_QUALITY_MEDIUM_WEIGHT 1
#define LIDAR_SMOOTH_QUALITY_LOW_WEIGHT 1
#define LIDAR_QUALITY_HIGH 1
#define LIDAR_QUALITY_MEDIUM 2
#define LIDAR_QUALITY_LOW 3
#define LOW_SPEED_AS_STOP -0.1
#define LOW_DRIFT_POS_AS_OUTLIER 0.3
#define LOW_DRIFT_THETA_AS_OUTLIER 2

// the main ESKF class
class ESKF
{
public:
    struct FullState
    {
        double current_time;
        Eigen::Matrix<float, STATE_SIZE, 1> makeState;
        Eigen::Matrix<float, dSTATE_SIZE, dSTATE_SIZE> makeP;

        Eigen::Vector3f getPosition()
        {
            return makeState.block<3, 1>(POS_IDX, 0);
        };
        Eigen::Quaterniond getRotation()
        {
            Eigen::Quaterniond quat;
            quat.w() = makeState.block<4, 1>(QUAT_IDX, 0)[0];
            quat.x() = makeState.block<4, 1>(QUAT_IDX, 0)[1];
            quat.y() = makeState.block<4, 1>(QUAT_IDX, 0)[2];
            quat.z() = makeState.block<4, 1>(QUAT_IDX, 0)[3];
            return quat;
        };
    };
    enum delayTypes
    {
        noMethod,         //apply updates  as if they are new.
        applyUpdateToNew, //Keep buffer of states, calculate what the update would have been, and apply to current state.
        larsonAverageIMU, //Method as described by Larson et al. Though a buffer of IMU values is kept, and a single update taking the average of these values is used.
        larsonNewestIMU,  //As above, though no buffer kept, use most recent value as representing the average.
        larsonFull        //As above, though the buffer is applied with the correct time steps, fully as described by Larson.
    };
    enum measureStatus 
    {   
        THETA_ALL,
        THETA_Roll,
        THETA_Pitch,
        THETA_Yaw
    };
    enum smoothStatus 
    {   
        SMOOTH_POS_Y,
        SMOOTH_POS_Z,
        SMOOTH_THETA_Roll,
        SMOOTH_THETA_Pitch,
        SMOOTH_THETA_Yaw
    };
    
    struct imuMeasurement
    {
        Eigen::Vector3f acc;
        Eigen::Vector3f gyro;
        lTime time;
    };
    ESKF(){};
    // takes as input the  variance of the acceleration and gyro, where _n is the measurement noise, and _w is the pertibations of the system.
    ESKF(Eigen::Vector3f a_gravity,
         const Eigen::Matrix<float, STATE_SIZE, 1> &initialState,
         const Eigen::Matrix<float, dSTATE_SIZE, dSTATE_SIZE> &initalP,
         float var_acc, float var_omega, float var_acc_bias, float var_omega_bias,
         Eigen::Vector3f sigma_pos, Eigen::Vector3f sigma_theta,
         Eigen::Vector3f sigma_vel, int lidar_status, float wheel_vel_range,
         int delayHandling, int bufferL, 
         Eigen::Vector3f i2b_lever_arm,
         const QRPConfigAutoData* config_data = NULL
         );

    // Concatenates relevant vectors to one large vector.
    static Eigen::Matrix<float, STATE_SIZE, 1> makeState(
        const Eigen::Vector3f &p,
        const Eigen::Vector3f &v,
        const Eigen::Quaternionf &q,
        const Eigen::Vector3f &a_b,
        const Eigen::Vector3f &omega_b);
    // Inserts relevant parts of the block-diagonal of the P matrix
    static Eigen::Matrix<float, dSTATE_SIZE, dSTATE_SIZE> makeP(
        const Eigen::Matrix3f &cov_pos,
        const Eigen::Matrix3f &cov_vel,
        const Eigen::Matrix3f &cov_dtheta,
        const Eigen::Matrix3f &cov_a_b,
        const Eigen::Matrix3f &cov_omega_b);

    // The quaternion convention in the document is "Hamilton" convention.
    // Eigen has a different order of components, so we need conversion
    static Eigen::Quaternionf quatFromHamilton(const Eigen::Vector4f &qHam);
    static Eigen::Vector4f quatToHamilton(const Eigen::Quaternionf &q);
    static Eigen::Matrix3f rotVecToMat(const Eigen::Vector3f &in);
    static Eigen::Quaternionf rotVecToQuat(const Eigen::Vector3f &in);
    static Eigen::Vector3f quatToRotVec(const Eigen::Quaternionf &q);
    static Eigen::Matrix3f getSkew(const Eigen::Vector3f &in);

    // Acessors of nominal state
    inline Eigen::Vector3f getPos() { return nominalState_.block<3, 1>(POS_IDX, 0); }
    inline Eigen::Vector3f getVel() { return nominalState_.block<3, 1>(VEL_IDX, 0); }
    inline Eigen::Vector4f getQuatVector() { return nominalState_.block<4, 1>(QUAT_IDX, 0); }
    inline Eigen::Quaternionf getQuat() { return quatFromHamilton(getQuatVector()); }
    inline Eigen::Vector3f getAccelBias() { return nominalState_.block<3, 1>(AB_IDX, 0); }
    inline Eigen::Vector3f getGyroBias() { return nominalState_.block<3, 1>(GB_IDX, 0); }
    inline Eigen::Matrix3f getDCM();
    Eigen::Matrix<float, STATE_SIZE, 1> getNominalState();

    void getFullState(FullState &state);

    // comment by weiji: 
    //      disable setFullState,
    //      the caller will clone whole object to store the full state
    // void setFullState(const FullState &state);

    void resetR(int status);
    void updateR(int status);
    void updateImu(const Eigen::Vector3f &acc, const Eigen::Vector3f &gryo, lTime now);
    void updateNoImu(lTime now);
    void measureLidar(const Eigen::Quaternionf &quat, 
                      const Eigen::Vector3f &pos_meas, 
                      int status, 
                      lTime now,
                      measureStatus measure_status_theta = THETA_ALL);
    void measureWheel(const Eigen::Vector3f &wheel_vel, 
                      double speed_range, 
                      lTime now);

private:
    // Called when there is a new measurment from the IMU.
    // dt is the integration time of this sample, nominally the IMU sample period
    void predictIMU(const Eigen::Vector3f &acc,
                    const Eigen::Vector3f &gryo,
                    const double dt,
#ifdef ENABLE_DELAY_HANDLE
                    lTime stamp,
#endif
                    lTime now);

    // Called when there is a new measurment from an absolute position reference.
    // Note that this has no body offset, i.e. it assumes exact observation of the center of the IMU.
    void measurePos(const Eigen::Vector3f &pos_meas,
                    const Eigen::Matrix3f &pos_covariance,
#ifdef ENABLE_DELAY_HANDLE
                    lTime stamp,
#endif
                    lTime now);

    std::tuple<Eigen::Quaternionf,Eigen::Vector3f> smoothMeasurement(const Eigen::Quaternionf &quat,const Eigen::Vector3f &pos_meas,int quality_level,lTime now);
    Eigen::Vector2f polynomialFitting(const std::deque<std::tuple<Eigen::VectorXf,int>> data, smoothStatus smooth_status);

    // Called when there is a new measurment from an absolute position reference.
    // The measurement is with respect to some location on the body that is not at the IMU center in general.
    // pos_ref_body should specify the reference location in the body frame.
    // For example, this would be the location of the GPS antenna on the body.
    // NOT YET IMPLEMENTED
    // void measurePosWithOffset(Eigen::Vector3f pos_meas, Matrix3f pos_covariance,
    //        Eigen::Vector3f pos_ref_body);

    // Called when there is a new measurment from an absolute orientation reference.
    // The uncertianty is represented as the covariance of a rotation vector in the body frame
    void measureQuat(const Eigen::Quaternionf &q_meas,
                     const Eigen::Matrix3f &theta_covariance,
#ifdef ENABLE_DELAY_HANDLE
                     lTime stamp,
#endif
                     lTime now,
                     measureStatus measure_status_theta = THETA_ALL);
    void measureVel(const Eigen::Vector3f &vel_meas, 
                    const Eigen::Matrix3f &vel_covariance, 
                    lTime now);
    Eigen::Matrix<float, 4, 3> getQ_dtheta(); // eqn 280, page 62
    void update_3D(const Eigen::Vector3f &delta_measurement,
                   const Eigen::Matrix3f &meas_covariance,
                   const Eigen::Matrix<float, 3, dSTATE_SIZE> &H,
#ifdef ENABLE_DELAY_HANDLE
                   lTime stamp,
#endif
                   lTime now);
    void injectErrorState(const Eigen::Matrix<float, dSTATE_SIZE, 1> &error_state);

    int delayHandling_;

#ifdef ENABLE_DELAY_HANDLE
    //get best time from history of state
    int getClosestTime(std::vector<std::pair<lTime, Eigen::Matrix<float, STATE_SIZE, 1>>> *ptr, lTime stamp);

    //get best time from history of imu
    int getClosestTime(std::vector<imuMeasurement> *ptr, lTime stamp);
    imuMeasurement getAverageIMU(lTime stamp);

    int bufferL_;
    int recentPtr;
    //pointers to structures that are allocated only after choosing a time delay handling method.
    std::vector<std::pair<lTime, Eigen::Matrix<float, STATE_SIZE, 1>>> *stateHistoryPtr_;
    std::vector<std::pair<lTime, Eigen::Matrix<float, dSTATE_SIZE, dSTATE_SIZE>>> *PHistoryPtr_;
    std::vector<imuMeasurement> *imuHistoryPtr_;
    imuMeasurement lastImu_;
    lTime firstMeasTime;
    lTime lastMeasurement;
    Eigen::Matrix<float, dSTATE_SIZE, dSTATE_SIZE> *Mptr;
#endif
    
    // qrp update params
public:
    const QRPConfigAutoData* qrp_config_data;
private:
    std::deque<double> gyro_data;
    bool monitor_gyro_flag;
    // smooth params
    std::deque<std::tuple<Eigen::VectorXf,int>> lidar_smooth_data;
    
    // IMU Noise values, used in prediction
    float var_acc_;
    float var_omega_;
    float var_acc_bias_;
    float var_omega_bias_;
    // Acceleration due to gravity in global frame
    Eigen::Vector3f a_gravity_; // [m/s^2]
    // State vector of the filter
    Eigen::Matrix<float, STATE_SIZE, 1> nominalState_;
    // Covariance of the (error) state
    Eigen::Matrix<float, dSTATE_SIZE, dSTATE_SIZE> P_;

    // Jacobian of the state transition: page 59, eqn 269
    // Note that we precompute the static parts in the constructor,
    // and update the dynamic parts in the predict function
    Eigen::Matrix<float, dSTATE_SIZE, dSTATE_SIZE> F_x_;

    Eigen::Matrix3f sigma_lidar_pos;
    Eigen::Matrix3f sigma_lidar_theta;
    Eigen::Matrix3f sigma_wheel_vel;

    double last_predict_ = 0;
    Eigen::Vector3f acc_m;
    Eigen::Vector3f gyro_m;  

    Eigen::Vector3f wheel_vel_;
    Eigen::Vector3f imu2base_lever_arm;
};

#endif /* ESKF_H */
