#include "voxel_filter.h"

#include <unordered_set>
#include <fstream>
#include <glog/logging.h>
//#define DEBUG_FILTER
#define distance(p) std::sqrt(p.x * p.x + p.y * p.y + p.z * p.z)
#define FIX_CURR_NUM 5000
#define TAR_NUM 6000
#define TAR_EXT_NUM 1000

inline uint64_t get_index(const PointJ &point, float freq)
{
    int x = std::lroundf(point.x * freq); // 24 bit, 1cm voxel for 16KM
    int y = std::lroundf(point.y * freq); // 24 bit
    int z = std::lroundf(point.z * freq); // 16 bit, 1cm voxel for 600M
    return (((uint64_t)x & 0xffffff) << 40) | (((uint64_t)y & 0xffffff) << 16) | (z & 0xffff);
}

inline uint64_t get_index(const Eigen::Vector3f &point, float freq)
{
    int x = std::lroundf(point.x() * freq); // 24 bit, 1cm voxel for 16KM
    int y = std::lroundf(point.y() * freq); // 24 bit
    int z = std::lroundf(point.z() * freq); // 16 bit, 1cm voxel for 600M
    return (((uint64_t)x & 0xffffff) << 40) | (((uint64_t)y & 0xffffff) << 16) | (z & 0xffff);
}

void vector_voxel_filter(const PointCloudJ::Ptr pcloud, VectorCloud &vcloud,
                         float resolution, float max_range) {    
    static std::unordered_set<uint64_t> last_voxel_set;
    vcloud.clear();
    vcloud.reserve(pcloud->size() / 5);
    float freq = 1 / resolution;
    VectorCloud extra_cloud;
    extra_cloud.reserve(pcloud->size());
    std::unordered_set<uint64_t> voxel_set;
    voxel_set.reserve(pcloud->size() * 5);
    for (const PointJ &point : pcloud->points) {
        if (max_range > 0 && distance(point) > max_range)
            continue;
        uint64_t key = get_index(point, freq);
        Eigen::Vector3f p(point.x, point.y, point.z);
        if (voxel_set.count(key) == 0) {    //若该体素没被添加过，则添加该点及该体素
            voxel_set.emplace(key);
            vcloud.push_back(p);
        }
        if (p[2] < 0) {
            continue;
        }
        if (last_voxel_set.size() && last_voxel_set.count(key) == 0) {  //若前帧体素集没有该体素，则将该点加入保护队列中
            extra_cloud.push_back(p);
        }
    }
    LOG(INFO) << "extra_cloud.size(): " << extra_cloud.size() << " vcloud.size(): " <<  vcloud.size();
    if (extra_cloud.size() && extra_cloud.size() < pcloud->size() * 0.1) {  //若保护队列点数小于总点数的10%，则认为可能是长廊场景
        if (extra_cloud.size() > TAR_EXT_NUM) {
            fix_num_rand_filter(extra_cloud, TAR_EXT_NUM);
        }
        VectorCloud cloud_without_extra;
        for (const auto& p : vcloud){
            uint64_t key = get_index(p, freq);
            if (last_voxel_set.count(key)) {    //若该点被保护过，则取消添加该点
                cloud_without_extra.emplace_back(p);
            }
        }
        vcloud.swap(cloud_without_extra);
        vcloud.insert(vcloud.end(), extra_cloud.begin(), extra_cloud.end());
    }
    LOG(INFO) << "Read scan " << pcloud->points.size() << " points"
              << ", after voxel filter " << resolution
              << ", left " << vcloud.size() << " points";
    last_voxel_set = voxel_set;

}

void clouds_voxel_filter(const PointCloudJ::Ptr pcloud, PointCloudJ::Ptr &clouds,
                         float resolution, float max_range) {
    clouds->points.clear();
    clouds->points.reserve(pcloud->size() / 5);

    std::unordered_set<uint64_t> voxel_set;
    voxel_set.reserve(pcloud->size() * 5);
    for (const PointJ &point : pcloud->points) {
        if (max_range > 0 && distance(point) > max_range)
            continue;
        uint64_t key = get_index(point, resolution);
        if (voxel_set.count(key) == 0) {
            voxel_set.emplace(key);
            clouds->points.push_back(point);
        }
    }
}

void fix_num_rand_filter(VectorCloud &vector, uint32_t fix_num)
{
    if(vector.size() <= fix_num){
        return;
    }
    VectorCloud ret;
    float radio = (float)vector.size() / fix_num;
    float diff = 0;
    for(size_t i = 0; i < vector.size(); i++){
        diff += 1;
        if(diff > radio){
            diff -= radio;
            ret.emplace_back(vector.at(i));
        }else{
            continue;
        }
    }
    vector.swap(ret);
}
