#pragma once

#include <ceres/ceres.h>
#include <ceres/rotation.h>
#include "../carto/common/math.h"

#include "../utils/vector_point_type.h"

namespace juefx {

class RotationDeltaCostFunction {
public:
    static ceres::CostFunction *CreateAutoDiffCostFunction(
        const double scaling_factor_rollpitch,
        const double scaling_factor_yaw,
        const Eigen::Quaterniond &target_rotation) {
        return new ceres::AutoDiffCostFunction<RotationDeltaCostFunction,
                                               3 /* residuals */,
                                               4 /* rotation variables */>(
            new RotationDeltaCostFunction(scaling_factor_rollpitch,
                                          scaling_factor_yaw, target_rotation));
    }

    template <typename T>
    bool operator()(const T *const rotation_quaternion, T *residual) const {
        std::array<T, 4> delta;
        cartographer::common::QuaternionProduct(target_rotation_inverse_, rotation_quaternion,
                                                delta.data());

        std::array<T, 3> axis_angel;
        ceres::QuaternionToAngleAxis(delta.data(), axis_angel.data());
        residual[0] = scaling_factor_rollpitch_ * axis_angel[0];
        residual[1] = scaling_factor_rollpitch_ * axis_angel[1];
        residual[2] = scaling_factor_yaw_ * axis_angel[2];
        return true;
    }

private:
    explicit RotationDeltaCostFunction(const double scaling_factor_rollpitch,
                                       const double scaling_factor_yaw,
                                       const Eigen::Quaterniond &target_rotation)
        : scaling_factor_rollpitch_(scaling_factor_rollpitch),
          scaling_factor_yaw_(scaling_factor_yaw) {
        target_rotation_inverse_[0] = target_rotation.w();
        target_rotation_inverse_[1] = -target_rotation.x();
        target_rotation_inverse_[2] = -target_rotation.y();
        target_rotation_inverse_[3] = -target_rotation.z();
    }

    RotationDeltaCostFunction(const RotationDeltaCostFunction &) = delete;
    RotationDeltaCostFunction &operator=(const RotationDeltaCostFunction &) = delete;

    const double scaling_factor_rollpitch_;
    const double scaling_factor_yaw_;
    double target_rotation_inverse_[4];
};

} 