#include "fast_voxel_matcher.h"

#define MAX_YAW_WINDOW_SIZE 2   // 2 * 2 + 1 = 5
#define TOP_CANDIDATE_COUNT 3
#define MAX_SCAN_CANDIDATE_COUNT 1

namespace juefx {

struct DiscreteScan {
    int32_t index;
    uint32_t ground_size;
    uint32_t above_ground_size;
    Rigid3f pose;
    std::vector<BlockGridIndex> grid_indexes;
    std::vector<int16_t> z_cm;
    std::vector<bool> above_ground;
};

struct Candidate {
    Candidate(int depth,
              const DiscreteScan *discrete_scan,
              const Eigen::Array3i &offset,
              const float low_resolution_score = 0)
        : search_depth(depth), scan(discrete_scan), offset(offset), score(-1.f),
          low_resolution_score(low_resolution_score) {}

    static Candidate Unsuccessful()
    {
        return Candidate(-1, NULL, Eigen::Array3i::Zero());
    }

    int search_depth;
    const DiscreteScan *scan;
    Eigen::Array3i offset;
    float score;
    float low_resolution_score;

    bool operator<(const Candidate& other) const { return score < other.score; }
    bool operator>(const Candidate& other) const { return score > other.score; }

    Rigid3f GetPose(float resolution) const {
        return Rigid3f(scan->pose.translation() + resolution * offset.matrix().cast<float>(),
                       scan->pose.rotation());
    }
};

FastVoxelMatcher::FastVoxelMatcher(
    const FastVoxelMatcherOption &options)
    : options_(options) {}

FastVoxelMatcher::~FastVoxelMatcher() {}

thread_local Candidate l_best_cand = Candidate::Unsuccessful();
thread_local uint32_t no_map_cnt = 0;

juefx::FastVoxelMatcher::Param FastVoxelMatcher::GetMatchParam(const VoxelMap &voxel_map,
                                                        float movement,
                                                        float rotation,
                                                        int mode) const {
    FastVoxelMatcher::Param param;
    param.mode = mode;
    param.resolution = voxel_map.GetResolution(VOXEL_RESOLUTION_1X);
    param.resolution_cm = (uint8_t)std::lround(param.resolution * 100);

    float XY_search_window, Z_search_window, Yaw_search_window;
    if ((mode & FAST_MODE_INIT) > 0) {
        XY_search_window = options_.XY_search_window * options_.init_XYZ_scale;
        Z_search_window = options_.Z_search_window * options_.init_XYZ_scale;
        Yaw_search_window = options_.Yaw_search_window * options_.init_Yaw_scale;
        param.yaw_step = options_.Yaw_search_window;
        param.yaw_window_size = (uint16_t)options_.init_Yaw_scale;
    }
    else {
        float scale = (mode & FAST_MODE_GUESS) > 0? 1: 0.5f;
        XY_search_window = std::min((float)fabs(movement) * scale, options_.XY_search_window);
        Z_search_window = std::min(XY_search_window * scale, options_.Z_search_window);
        Yaw_search_window = std::min((float)fabs(rotation) * scale, options_.Yaw_search_window);
    }
    // xy search (-window 0 +window: window*2+1 )
    int xy_window_size = std::lround(XY_search_window / param.resolution - 0.5f);
    param.xy_window_size = xy_window_size >= 1 ? (uint16_t)xy_window_size: 1;
    int z_window_size = std::lround(Z_search_window / param.resolution - 0.5f);
    param.z_window_size = z_window_size >= 1 ? (uint16_t)z_window_size : 1;
    // start search depth & search step
    param.start_depth = VOXEL_RESOLUTION_NUM - 1;
    int xy_step_size = 1 << param.start_depth;
    while (xy_step_size >= param.xy_window_size * 2 + 1) {
        param.start_depth --;
        xy_step_size = 1 << param.start_depth;
    }
    // yaw search if not init
    if ((mode & FAST_MODE_NOYAW) > 0) {
        param.yaw_step = 0;
        param.yaw_window_size = 0;
    }
    else if ((mode & FAST_MODE_INIT) == 0) {
        param.yaw_step = options_.min_Yaw_search_step;
        int yaw_window_size = std::lround(Yaw_search_window / param.yaw_step - 0.5f);
        param.yaw_window_size = yaw_window_size >= 1 ? (uint16_t)yaw_window_size : 1;
        while (param.yaw_window_size > MAX_YAW_WINDOW_SIZE) {
            param.yaw_step *= 2;
            int yaw_window_size = std::lround(Yaw_search_window / param.yaw_step - 0.5f);
            param.yaw_window_size = yaw_window_size >= 1 ? (uint16_t)yaw_window_size : 1;
        }
    }
    return param;
}

std::unique_ptr<FastVoxelMatcher::Result>
FastVoxelMatcher::Match(const Rigid3f &init_pose,
                        const VectorCloud &point_cloud,
                        const VoxelMap &voxel_map,
                        const FastVoxelMatcher::Param &param) const
{
    no_map_cnt = 0;

    const std::vector<DiscreteScan> discrete_scans = GenerateDiscreteScans(
        init_pose, point_cloud, voxel_map, param);

    std::vector<Candidate> lowest_resolution_candidates =
        GenerateLowestResolutionCandidates(voxel_map, discrete_scans, param);

    l_best_cand = Candidate::Unsuccessful();
#ifdef VOXEL_MAP_FAST_DEBUG
    printf("fast match branch_bound start, accept_score %f ...\n", options_.accepted_score);
#endif
    const Candidate best_candidate = BranchAndBound(
        voxel_map, lowest_resolution_candidates, param, options_.accepted_score);
    LOG(INFO) << "best_candidate.score: " << best_candidate.score;
    if (best_candidate.score > options_.accepted_score) {
        LOG(INFO) << "index: " << best_candidate.scan->index
                  << " discrete_scans.size(): " << discrete_scans.size()
                  << " best_candidate.offset: " << Eigen::Vector3i(best_candidate.offset.x(),
                                                                   best_candidate.offset.y(),
                                                                   best_candidate.offset.z()).transpose();
        return std::make_unique<Result>(Result{
            best_candidate.score, best_candidate.low_resolution_score,
            best_candidate.GetPose(param.resolution)});
    } else if (l_best_cand.score > 0) {
        if (l_best_cand.search_depth > 0) {
            const int step_size = 1 << l_best_cand.search_depth;
            for (size_t index = 0; index < 3; index++) {
                if (l_best_cand.offset[index] <= 0 && l_best_cand.offset[index] + step_size >= 0) {
                    l_best_cand.offset[index] = 0;
                }else {
                    l_best_cand.offset[index] + l_best_cand.offset[index] < 0 ? step_size * 0.5 : -step_size * 0.5;
                }
            }
        }
        return std::make_unique<Result>(Result{
            l_best_cand.score, l_best_cand.low_resolution_score,
            l_best_cand.GetPose(param.resolution)});
    }
    return NULL;
}

std::vector<DiscreteScan> FastVoxelMatcher::GenerateDiscreteScans(const Rigid3f &init_pose,
                                                                  const VectorCloud &point_cloud,
                                                                  const VoxelMap &voxel_map,
                                                                  const FastVoxelMatcher::Param &param) const {
    std::vector<DiscreteScan> result;
    const int angular_window_size = param.yaw_window_size;
    std::vector<float> angles;
    for (int rz = -angular_window_size; rz <= angular_window_size; ++rz) {
        angles.push_back(rz * param.yaw_step);
    }
    for (size_t i = 0; i != angles.size(); ++i) {
        const Eigen::Vector3f angle_axis(0.f, 0.f, angles[i]);
        // It's important to apply the 'angle_axis' rotation between the translation
        // and rotation of the 'initial_pose', so that the rotation is around the
        // origin of the range data, and yaw is in map frame.
        const Rigid3f pose(
            init_pose.translation(),
            cartographer::transform::AngleAxisVectorToRotationQuaternion(angle_axis) *
                init_pose.rotation());
        result.push_back(
            CreateDiscreteScan(i, pose, point_cloud, voxel_map));
    }
  return result;
}

DiscreteScan FastVoxelMatcher::CreateDiscreteScan(const int index,
                                                  const Rigid3f &trans_pose,
                                                  const VectorCloud &point_cloud,
                                                  const VoxelMap &voxel_map) const {
    DiscreteScan scan;
    scan.index = index;
    scan.ground_size = 0;
    scan.above_ground_size = 0;
    scan.pose = trans_pose;
    scan.grid_indexes.reserve(point_cloud.size());
    scan.z_cm.reserve(point_cloud.size());
    scan.above_ground.reserve(point_cloud.size());

    Eigen::Isometry3f trans;
    rigid3f_to_isometry3f(trans_pose, trans);
    for (const auto &point : point_cloud) {
        Eigen::Vector3f trans_point = trans * point;
        auto index = voxel_map.GetCellIndex(trans_point.x(), trans_point.y());
        scan.grid_indexes.push_back(index);
        scan.z_cm.push_back((int16_t)floor(trans_point.z() * 100));
        // ground & above_ground
        if (point.z() < - options_.imu_height) {
            scan.above_ground.push_back(false);
            scan.ground_size++;
        } else {
            scan.above_ground.push_back(true);
            scan.above_ground_size++;
        }
    }
    CHECK_EQ(scan.ground_size + scan.above_ground_size, point_cloud.size());
    return scan;
}

std::vector<Candidate>
FastVoxelMatcher::GenerateLowestResolutionCandidates(const VoxelMap &voxel_map,
                                                     const std::vector<DiscreteScan> &discrete_scans,
                                                     const FastVoxelMatcher::Param &param) const {

    const int start_depth = param.start_depth;
    const int step_size = 1 << start_depth;
    const int num_xy_candidates = (2 * param.xy_window_size + 1 + step_size) / step_size;
    const int num_z_candidates = (2 * param.z_window_size + 1 + step_size) / step_size;
    const int num_candidates = discrete_scans.size() *
                               num_xy_candidates * num_xy_candidates * num_z_candidates;
    std::vector<Candidate> candidates;
    candidates.reserve(num_candidates);
    for (int scan_index = 0; scan_index != discrete_scans.size(); ++scan_index)
    {
        const DiscreteScan* scan = &(discrete_scans[scan_index]);
        for (int x = -param.xy_window_size; x <= param.xy_window_size; x += step_size)
        {
            for (int y = -param.xy_window_size; y <= param.xy_window_size; y += step_size)
            {
                for (int z = -param.z_window_size; z <= param.z_window_size; z += step_size)
                {
                    candidates.emplace_back(start_depth, scan, Eigen::Array3i(x, y, z));
                }
            }
        }
    }
    CHECK_EQ(candidates.size(), num_candidates);
    ScoreCandidates(voxel_map, &candidates, param, true);
    // buffer current yaw & pose
    for (int i = 0; i < candidates.size(); i++) {
        Candidate& cand = candidates[i];
        if (cand.scan->index == discrete_scans.size() / 2)
            cand.score += 0.01;
        if (cand.offset[0] <= 0 && cand.offset[0] + step_size >= 0
                && cand.offset[1] <= 0 && cand.offset[1] + step_size >= 0
                && cand.offset[2] <= 0 && cand.offset[2] + step_size >= 0)
            cand.score += 0.01;
    }
    // Only keep TOP 3 candidates, and at most 1 cands in one scan
    std::vector<int> scans_count;
    scans_count.resize(candidates.size());
    std::vector<Candidate> top_candidates;
    for (int i = 0; i < candidates.size(); i++) {
        const Candidate& cand = candidates[i];
        ++ scans_count[cand.scan->index];
        if (scans_count[cand.scan->index] <= MAX_SCAN_CANDIDATE_COUNT
            && top_candidates.size() < TOP_CANDIDATE_COUNT) {
            top_candidates.push_back(cand);
        }
    }
    return top_candidates;
}

// [floor_cm, ceil_cm) vs. [d.floor, d.floor + d.height)
inline int CountSpaceProbility(const Distribution *dist, uint8_t length, int floor_cm, int ceil_cm)
{
    if (dist == NULL || length == 0)
        return 0;
    for (size_t i = 0; i < length; i++)
    {
        const Distribution &d = dist[i];
        if (floor_cm < d.floor + d.height && d.floor < ceil_cm)
            return 1;
    }
    return 0;
}

void FastVoxelMatcher::ScoreCandidates(const VoxelMap &voxel_map,
                                       std::vector<Candidate> *const candidates,
                                       const FastVoxelMatcher::Param &param,
                                       bool low_resolution) const {
    if(0 == candidates->size()){
        return;
    }
    std::vector<float> ground_sum;
    std::vector<float> above_ground_sum;
    std::vector<uint32_t> has_map_sum;
    ground_sum.resize(candidates->size());
    above_ground_sum.resize(candidates->size());
    has_map_sum.resize(candidates->front().scan->grid_indexes.size());
    for (int c = 0; c < candidates->size(); c++)
    {
        Candidate &candidate = (*candidates)[c];
        int same_xy = c + 1;
        // process candidate with same xy offset
        for (; same_xy < candidates->size(); same_xy++) {
            Candidate &candidate_next = (*candidates)[same_xy];
            if (candidate_next.scan != candidate.scan
                || candidate_next.offset.x() != candidate.offset.x()
                || candidate_next.offset.y() != candidate.offset.y()) {
                break;
            }
        }
        CHECK_LT(candidate.search_depth, VOXEL_RESOLUTION_NUM);
        for (size_t i = 0; i < candidate.scan->grid_indexes.size(); i++) {
            bool hasMap = false;
            const auto &index = candidate.scan->grid_indexes[i];
            const auto final_index = index.Move(candidate.offset.x(), candidate.offset.y());
            uint8_t length = 0;
            const auto* voxel = voxel_map.GetCellValue(candidate.search_depth,
                                                       final_index, length);
            int height = param.resolution_cm << candidate.search_depth;
            int floor = param.resolution_cm >> 1;
            int ceil = height - floor;
            // process candidate with same xy offset
            for (int k = c; k < same_xy; k++) {
                Candidate &candidate_xy = (*candidates)[k];
                int z = candidate_xy.scan->z_cm[i] + candidate_xy.offset.z() * param.resolution_cm;
                int prob = CountSpaceProbility(voxel, length, z - floor, z + ceil);
                if (prob) {
                    hasMap = true;
                }
                if (candidate_xy.scan->above_ground[i])
                    above_ground_sum[k] += prob;
                else
                    ground_sum[k] += prob;
            }
            if (hasMap) {
                has_map_sum[i]++;
            }
        }
        // process candidate with same xy offset
        for (int k = c; k < same_xy; k++) {
            if (param.mode & FAST_MODE_GROUND == 0) {
                Candidate &candidate_xy = (*candidates)[k];
                candidate_xy.score = (ground_sum[k] + above_ground_sum[k]) /
                        (candidate.scan->ground_size + candidate.scan->above_ground_size);
                if (low_resolution)
                {
                    // if lowest depth is higher than low_resolution, set low_resolution direct to 0.9
                    candidate_xy.low_resolution_score =
                        candidate.search_depth >= options_.low_resolution_depth ? candidate_xy.score : 0.9;
                }
#ifdef VOXEL_MAP_FAST_DEBUG
                printf("score candidate, scan %d, depth %d, offset %d %d %d, score %f, low_score %f\n",
                       candidate.scan->index, candidate.search_depth,
                       candidate.offset.x(), candidate.offset.y(), candidate_xy.offset.z(),
                       candidate_xy.score, candidate_xy.low_resolution_score);
#endif
            }
            else {
                Candidate &candidate_xy = (*candidates)[k];
                float ground_score = ground_sum[k] / candidate.scan->ground_size;
                float above_ground_score = above_ground_sum[k] / candidate.scan->above_ground_size;
                float min_score = std::min(above_ground_score, ground_score);
                float max_score = std::max(above_ground_score, ground_score);
                candidate_xy.score = min_score * 0.8f + std::min(max_score, 0.667f) * 0.2f;
                if (low_resolution) {
                    // if lowest depth is higher than low_resolution, set low_resolution direct to 0.9
                    candidate_xy.low_resolution_score =
                        candidate.search_depth >= options_.low_resolution_depth ? candidate_xy.score : 0.9;
                }
#ifdef VOXEL_MAP_FAST_DEBUG
                printf("score candidate, scan %d, depth %d, offset %d %d %d, score %f/%f=%f, low_score %f\n",
                    candidate.scan->index, candidate.search_depth,
                    candidate.offset.x(), candidate.offset.y(), candidate_xy.offset.z(),
                    above_ground_score, ground_score, candidate_xy.score,
                    candidate_xy.low_resolution_score);
#endif
            }
        }
        c = same_xy - 1;
    }
    if (low_resolution) {
        for (const auto &sum : has_map_sum) {
            if(0 == sum){
                no_map_cnt++;
            }
        }
        LOG(INFO) << "grid_indexes: " << candidates->front().scan->grid_indexes.size() << ", no_map_cnt: " << no_map_cnt;
    }
    if (0 != no_map_cnt) {
        float multi = (float)candidates->front().scan->grid_indexes.size() /
                (candidates->front().scan->grid_indexes.size() - no_map_cnt);
        for (Candidate &candidate : *candidates) {
            candidate.score *= multi;
        }
    }
    std::sort(candidates->begin(), candidates->end(), std::greater<Candidate>());
}

Candidate FastVoxelMatcher::BranchAndBound(const VoxelMap &voxel_map,
                                           const std::vector<Candidate> &candidates,
                                           const FastVoxelMatcher::Param &param,
                                           float min_score) const {
    CHECK_GT(candidates.size(), 0);
    const int candidate_depth = candidates[0].search_depth;
    const float accepted_low_score = (param.mode == FAST_MODE_INIT) ? options_.accepted_score :
                                                                      options_.accepted_low_score;
    // add by weiji: record best candidate
    for (const Candidate &candidate : candidates) {
        if (candidate.search_depth <= 2 &&
                (l_best_cand.search_depth < 0 ||
                 candidate.search_depth < l_best_cand.search_depth ||
                 (candidate.search_depth == l_best_cand.search_depth &&
                        candidate.score > fabs(l_best_cand.score)))) {
            l_best_cand = candidate;
            if (!((param.mode & FAST_MODE_INIT) || (param.mode & FAST_MODE_GUESS))) {
                l_best_cand.score *= -1;
            }
        }
    }

    if (candidate_depth == VOXEL_RESOLUTION_1X)
    {
        for (const Candidate &candidate : candidates) {
            if (candidate.score <= min_score) {
                // Return if the candidate is bad because the following candidate will
                // not have better score.
                return Candidate::Unsuccessful();
            }
            if (candidate.low_resolution_score < accepted_low_score) {
                continue;
            }
#ifdef VOXEL_MAP_FAST_DEBUG
            printf("====>>====>> scan %d, depth %d, offset %d %d %d, score %f, low_score %f\n",
                   candidate.scan->index, candidate.search_depth,
                   candidate.offset.x(), candidate.offset.y(), candidate.offset.z(),
                   candidate.score, candidate.low_resolution_score);
#endif
            return candidate;
        }
        // All candidates have good scores but none passes the low score check.
        return Candidate::Unsuccessful();
    }

    int next_depth = candidate_depth - 1;
    Candidate best_high_resolution_candidate = Candidate::Unsuccessful();
    // every depth, score limit begins with min_score
    best_high_resolution_candidate.score = min_score;
    for (const Candidate& candidate : candidates) {
        if (candidate.score <= best_high_resolution_candidate.score ||
            candidate.low_resolution_score < accepted_low_score) {
            break;
        }
#ifdef VOXEL_MAP_FAST_DEBUG
        printf(">>>> scan %d, depth %d, offset %d %d %d, score %f, low_score %f\n",
               candidate.scan->index, candidate.search_depth,
               candidate.offset.x(), candidate.offset.y(), candidate.offset.z(),
               candidate.score, candidate.low_resolution_score);
#endif
        std::vector<Candidate> higher_resolution_candidates;
        const int half_width = 1 << (candidate_depth - 1);
        // x y z start from -window_size, no need to check lower bound
        for (int x : {0, half_width}) {
            if (candidate.offset.x() + x > param.xy_window_size)
                break;
            for (int y : {0, half_width}) {
                if (candidate.offset.y() + y > param.xy_window_size)
                    break;
                for (int z : {0, half_width})
                {
                    if (candidate.offset.z() + z > param.z_window_size)
                        break;
                    higher_resolution_candidates.emplace_back(
                        next_depth, candidate.scan, candidate.offset + Eigen::Array3i(x, y, z),
                        candidate.low_resolution_score);
                }
            }
        }
        bool low_resolution = next_depth >= options_.low_resolution_depth;
        ScoreCandidates(voxel_map, &higher_resolution_candidates, param, low_resolution);
        best_high_resolution_candidate =
            std::max(best_high_resolution_candidate,
                     BranchAndBound(voxel_map, higher_resolution_candidates, param,
                                    best_high_resolution_candidate.score));
    }
    return best_high_resolution_candidate;
}

}
