#pragma once

#include <cmath>

#include "../utils/vector_point_type.h"
#include "voxel_map_interface.h"

namespace juefx {

class Jet {
public:
    static double GetValue(const double x) { return x; }

    static const double *GetDerivative(const double x, int &n)
    {
        n = 0;
        return NULL;
    }

    template <typename T>
    static double GetValue(const T &jet_x) { return jet_x.a; }

    template <typename T>
    static const double *GetDerivative(const T &jet_x, int &n) 
    {
        n = jet_x.v.size();
        return jet_x.v.data();
    }
};

// Interpolates between HybridGrid probability voxels. We use the tricubic
// interpolation which interpolates the values and has vanishing derivative at
// these points.
//
// This class is templated to work with the autodiff that Ceres provides.
// For this reason, it is also important that the interpolation scheme be
// continuously differentiable.
class InterpolatedVoxelMap {
public:
    explicit InterpolatedVoxelMap(const VoxelMap &voxel_map,
                                  uint8_t type)
        : voxel_map_(voxel_map),
          rtype_(type),
          resolution_(voxel_map.GetResolution(type)) {}

    InterpolatedVoxelMap(const InterpolatedVoxelMap &) = delete;
    InterpolatedVoxelMap &operator=(const InterpolatedVoxelMap &) = delete;

    // Returns the interpolated probability at (x, y, z) of the HybridGrid
    // used to perform the interpolation.
    //
    // This is a piecewise, continuously differentiable function. We use the
    // scalar part of Jet parameters to select our interval below. It is the
    // tensor product volume of piecewise cubic polynomials that interpolate
    // the values, and have vanishing derivative at the interval boundaries.
    template <typename T>
    bool GetProbability(const T &x, const T &y, const T &z, T &p_space, T &p_intensity) const
    {
        const Eigen::Vector3f p = TransferValue(x, y, z);
        auto index = voxel_map_.GetPymdCellIndex(p.x(), p.y());

        double x1, y1, x2, y2;
        int direction = ComputeInterpolationDataPoints(p.x(), p.y(), &x1, &y1, &x2, &y2);
        const T normalized_x = (x - x1) / (x2 - x1);
        const T normalized_y = (y - y1) / (y2 - y1);
        // Compute pow(..., 2) and pow(..., 3). Using pow() here is very expensive.
        const T normalized_xx = normalized_x * normalized_x;
        const T normalized_xxx = normalized_x * normalized_xx;
        const T normalized_yy = normalized_y * normalized_y;
        const T normalized_yyy = normalized_y * normalized_yy;

        // 12 22
        // 11 21
        const Distribution *voxel_11, *voxel_12, *voxel_21, *voxel_22;
        uint8_t length_11, length_12, length_21, length_22;
        if (direction == (GRID_MOVE_LEFT | GRID_MOVE_UP)) {
            voxel_11 = voxel_map_.GetPymdCellValue(rtype_, index.Step(rtype_, GRID_MOVE_LEFT), length_11);
            voxel_21 = voxel_map_.GetPymdCellValue(rtype_, index, length_21);
            voxel_12 = voxel_map_.GetPymdCellValue(rtype_, index.Step(rtype_, GRID_MOVE_LEFT | GRID_MOVE_UP), length_12);
            voxel_22 = voxel_map_.GetPymdCellValue(rtype_, index.Step(rtype_, GRID_MOVE_UP), length_22);
        }
        else if (direction == (GRID_MOVE_RIGHT | GRID_MOVE_UP)) {
            voxel_11 = voxel_map_.GetPymdCellValue(rtype_, index, length_11);
            voxel_21 = voxel_map_.GetPymdCellValue(rtype_, index.Step(rtype_, GRID_MOVE_RIGHT), length_21);
            voxel_12 = voxel_map_.GetPymdCellValue(rtype_, index.Step(rtype_, GRID_MOVE_UP), length_12);
            voxel_22 = voxel_map_.GetPymdCellValue(rtype_, index.Step(rtype_, GRID_MOVE_RIGHT | GRID_MOVE_UP), length_22);
        }
        else if (direction == (GRID_MOVE_RIGHT | GRID_MOVE_DOWN)) {
            voxel_11 = voxel_map_.GetPymdCellValue(rtype_, index.Step(rtype_, GRID_MOVE_DOWN), length_11);
            voxel_21 = voxel_map_.GetPymdCellValue(rtype_, index.Step(rtype_, GRID_MOVE_RIGHT | GRID_MOVE_DOWN), length_21);
            voxel_12 = voxel_map_.GetPymdCellValue(rtype_, index, length_12);
            voxel_22 = voxel_map_.GetPymdCellValue(rtype_, index.Step(rtype_, GRID_MOVE_RIGHT), length_22);
        }
        else if (direction == (GRID_MOVE_LEFT | GRID_MOVE_DOWN)) {
            voxel_11 = voxel_map_.GetPymdCellValue(rtype_, index.Step(rtype_, GRID_MOVE_LEFT | GRID_MOVE_DOWN), length_11);
            voxel_21 = voxel_map_.GetPymdCellValue(rtype_, index.Step(rtype_, GRID_MOVE_DOWN), length_21);
            voxel_12 = voxel_map_.GetPymdCellValue(rtype_, index.Step(rtype_, GRID_MOVE_LEFT), length_12);
            voxel_22 = voxel_map_.GetPymdCellValue(rtype_, index, length_22);
        }
        else {
            DCHECK(false);
        }

        // We first interpolate y, then x. All 3 times this uses the same
        // scheme: A * (2t^3 - 3t^2 + 1) + B * (-2t^3 + 3t^2).
        // The first polynomial is 1 at t=0, 0 at t=1, the second polynomial is 0
        // at t=0, 1 at t=1. Both polynomials have derivative zero at t=0 and t=1.
#define count_interpolated(q11, q12, q21, q22, q)        \
    const T q1 = (q11 - q12) * normalized_yyy * 2. +     \
                 (q12 - q11) * normalized_yy * 3. + q11; \
    const T q2 = (q21 - q22) * normalized_yyy * 2. +     \
                 (q22 - q21) * normalized_yy * 3. + q21; \
    q = (q1 - q2) * normalized_xxx * 2. + (q2 - q1) * normalized_xx * 3. + q1;

        T space_11 = CountSpaceProbility(voxel_11, length_11, z);
        T space_12 = CountSpaceProbility(voxel_12, length_12, z);
        T space_21 = CountSpaceProbility(voxel_21, length_21, z);
        T space_22 = CountSpaceProbility(voxel_22, length_22, z);
        count_interpolated(space_11, space_12, space_21, space_22, p_space);

        if (false && rtype_ == VOXEL_RESOLUTION_1X)
        {
            double intsy_11 = 1; //voxel_11;
            double intsy_12 = 1; //voxel_12;
            double intsy_21 = 1; //voxel_21;
            double intsy_22 = 1; //voxel_22;
            count_interpolated(intsy_11, intsy_12, intsy_21, intsy_22, p_intensity);
        }

#ifdef VOXEL_MAP_CERES_DEBUG
        if (debug_output_count_ > 0)
        {
            int *p_count = const_cast<int *>(&debug_output_count_);
            (*p_count) --;
            printf("grid probability type %u, x: %f, y: %f, z: %f \n", rtype_, p.x(), p.y(), p.z());
            printf("x1: %f, y1: %f / x2: %f, y2: %f \n", x1, y1, x2, y2);
            for (auto d : *voxel_11)
                printf("11 - %d %u %u\n", d.floor, d.height, d.intensity);
            for (auto d : *voxel_12)
                printf("12 - %d %u %u\n", d.floor, d.height, d.intensity);
            for (auto d : *voxel_21)
                printf("21 - %d %u %u\n", d.floor, d.height, d.intensity);
            for (auto d : *voxel_22)
                printf("22 - %d %u %u\n", d.floor, d.height, d.intensity);
            printf("p_11: %f, p_12: %f, p_21: %f, p_22: %f, p_space: %f\n",
                   Jet::GetValue(space_11), Jet::GetValue(space_12), Jet::GetValue(space_21), Jet::GetValue(space_22),
                   Jet::GetValue(p_space));
            int n = 0;
            const double *dv = Jet::GetDerivative(p_space, n);
            if (n > 0) {
                printf("space derivative: ");
                for (int i = 0; i < n; i++)
                    printf("%f ", dv[i]);
                printf("\n");
            }
            printf("\n");
        }
#endif

        return true;
    }

#ifdef VOXEL_MAP_CERES_DEBUG
    int debug_output_count_ = VOXEL_MAP_CERES_OUTPUT_NUM;
    void ResetDebugCount() const { 
        int* p_count = const_cast<int*>(&debug_output_count_);
        *p_count = VOXEL_MAP_CERES_OUTPUT_NUM; 
    }
#endif

private:
    Eigen::Vector3f TransferValue(const double x, const double y,
                                  const double z) const
    {
        return Eigen::Vector3f(x, y, z);
    }

    template <typename T>
    Eigen::Vector3f TransferValue(const T &jet_x, const T &jet_y,
                                  const T &jet_z) const
    {
        return TransferValue(jet_x.a, jet_y.a, jet_z.a);
    }

    int ComputeInterpolationDataPoints(const float x, const float y,
                                       double *x1, double *y1,
                                       double *x2, double *y2) const
    {
        // Center of the cell containing (x, y).
        Eigen::Vector2f center = voxel_map_.GetCenterOfCell(rtype_, x, y);
        int direction = 0;
        if (center.x() > x) {
            *x1 = center.x() - resolution_;
            *x2 = center.x();
            direction |= GRID_MOVE_LEFT;
        } else {
            *x1 = center.x();
            *x2 = center.x() + resolution_;
            direction |= GRID_MOVE_RIGHT;
        }
        if (center.y() > y) {
            *y1 = center.y() - resolution_;
            *y2 = center.y();
            direction |= GRID_MOVE_DOWN;
        } else {
            *y1 = center.y();
            *y2 = center.y() + resolution_;
            direction |= GRID_MOVE_UP;
        }
        return direction;
    }

    struct DistributionResult {
        const Distribution *dist = NULL;
        int min_distance_cm = 10000;
    };

    template <typename T>
    T CountSpaceProbility(const Distribution *dist, uint8_t length, const T &z) const
    {
        if (dist == NULL || length == 0)
            return (T)0.;
        DistributionResult result = GetNearestDistruibution(dist, length, z);
        if (result.min_distance_cm == 0)
            return (T)1.;
        T x;
        if (result.min_distance_cm < 0)
            x = (result.dist->floor / 100.0 - z) / resolution_;
        else
            x = (z - (result.dist->floor + result.dist->height) / 100.0) / resolution_;
        if (x > 1.)
            return 1. / (5. * x);
        else if (x > 0.)
            return 7/5. * x * x * x - 11/5. * x * x + 1.;
        else {
            DCHECK(false);
            return (T)0.;
        }
    }

    inline DistributionResult GetNearestDistruibution(const Distribution *dist, uint8_t length, const double z) const
    {
        DistributionResult result;
        int z_cm = lround(z * 100);
        for (size_t i = 0; i < length; i++)
        {
            const Distribution &d = dist[i];
            int distance_cm = z_cm < d.floor ? z_cm - d.floor: 
                    (d.floor <= z_cm && z_cm <= d.floor + d.height ? 0 : z_cm - d.floor - d.height);
            if (abs(distance_cm) < abs(result.min_distance_cm)) {
                result.min_distance_cm = distance_cm;
                result.dist = &d;
            }
        }
        DCHECK(result.dist != NULL);
        return result;
    }

    template <typename T>
    DistributionResult GetNearestDistruibution(const Distribution *dist, uint8_t length, const T &jet_z) const
    {
        return GetNearestDistruibution(dist, length, jet_z.a);
    }

    const VoxelMap voxel_map_;
    const uint8_t rtype_;
    const double resolution_;
};

}  // namespace 
