#pragma once
#ifndef SLAM_INPUTS_H
#define SLAM_INPUTS_H 

#include <string>
#include <vector>
#include <dirent.h>

#include <Eigen/Dense>
#include <pcl/io/pcd_io.h>
#include <pcl/point_types.h>
#include "../util.hpp"
#include "localize_utils/utils.h"
#include "../../localize_utils/pcl_point_type.h"
using namespace std;

#define MODE_PANDAR40P 1
#define MODE_PANDAR40P_RAW 2
#define MODE_HESAIXT32 3
#define MODE_HESAIXT32_RAW 4
#define MODE_HESAIXT32_SIMPLE 5
#define MODE_PANDAR40P_OLD 6
#define MODE_VELODYNE32E 7
#define BY_FRAME(type)                                              \
    (type == MODE_PANDAR40P || type == MODE_HESAIXT32 ||            \
     type == MODE_HESAIXT32_SIMPLE || type == MODE_PANDAR40P_OLD)
#define IS_PANDAR40P(type) (type == MODE_PANDAR40P || type == MODE_PANDAR40P_RAW || type == MODE_PANDAR40P_OLD)
#define IS_HESAIXT32(type) (type == MODE_HESAIXT32 || type == MODE_HESAIXT32_RAW || type == MODE_HESAIXT32_SIMPLE)
// MEMS
#define MODE_ROBOSM1 8
#define MODE_INNOVUSION 9
#define IS_MEMS(type) (type == MODE_ROBOSM1 || type == MODE_INNOVUSION)
// THIRD BAG
#define MODE_JD 10
#define MODE_JD_SELF 11
#define IS_JD(type) (type == MODE_JD || type == MODE_JD_SELF || type == MODE_LIO_JD)
#define MODE_UDI 12
#define IS_FREEZE_STOP(type) (type == MODE_JD || type == MODE_JD_SELF || type == MODE_UDI)
// THIRD ODOM FILE
#define MODE_INCEPTIO 20
#define MODE_INCEPTIO_WHEEL 21
#define IS_INCEPTIO(type) (type == MODE_INCEPTIO || type == MODE_INCEPTIO_WHEEL)
// LIO
#define MODE_LIO_JD 50
#define IE_SPLIT_SEC_LENGTH 1000
double get_unix_time(int week, double seconds);

///////////////////////////////////////////////////////////
// Gps File

struct IERawData
{
    int week;
    double seconds;
    double unix_time;
    double latitude, longitude, height;
    double n_velocity, e_velocity, u_velocity;
    double roll, pitch, yaw;                    // radius
    int status;

    // the orientation is described in ros frame: front - left - up
    Eigen::Matrix3d getOrientation() const {
        return (Eigen::AngleAxisd(-yaw, Eigen::Vector3d::UnitZ()) *
                    Eigen::AngleAxisd(-pitch, Eigen::Vector3d::UnitY()) *
                    Eigen::AngleAxisd(roll, Eigen::Vector3d::UnitX()))
                .toRotationMatrix();
    }

    static bool parseFromString(std::string line, IERawData &ie_data);
};

class IEFileReader
{
public:
    IEFileReader(){}
    ~IEFileReader(){}

    int loadIEFile(std::string ieFile, double start_time, double end_time,
                   bool split = false, std::string outputPath = "");
    int loadIEStartFile(std::string ieStartFile);

    IERawData getIEData(double unix_time);
    inline size_t cursor() { return cursor_; }

    std::vector<IERawData> raw_data;
    IERawData start_data;

private:
    size_t cursor_ = 0;
};

///////////////////////////////////////////////////////////
// Imu File

struct IMURawData
{
    int week;
    double seconds;
    double unix_time;
    double x_acc, y_acc, z_acc;
    double x_angular, y_angular, z_angular; // radian

    static bool parseFromString(std::string line, int type, IMURawData &ie_data);
};

class IMUFileReader
{
public:
    IMUFileReader() {}
    ~IMUFileReader() {}

    int loadIMUFile(std::string imuFile, int type, double start_time, double end_time);

    std::vector<IMURawData> raw_data;

private:
    int makeFakeIMU(double start_time, double end_time);
};

///////////////////////////////////////////////////////////
// Wheel File

struct WheelRawData
{
    int week;
    double seconds;
    double unix_time;
    double movement, total_movement;
};

class WheelFileReader
{
public:
    WheelFileReader() : total_movement(0) {}
    ~WheelFileReader() {}

    int loadWheelFile(std::string wheelFile, double start_time, double end_time);
    std::vector<WheelRawData> raw_data;

private:
    bool parseFromString(std::string line, WheelRawData &wheel_data);
    double total_movement;
};


class IEReader
{
public:
    using Ptr = boost::shared_ptr<IEReader>;
    IEReader()
    {

    }

    ~IEReader()
    {
        ie_reader.reset();
    }
    static IEReader::Ptr Instance();

    void Init(const std::string &path)
    {
        ie_reader.reset(new IEFileReader);
        LoadIEFile(path);

    }
    int LoadIEFile(std::string ieFile);


    std::vector<IERawData> GetData(){
        return ie_reader->raw_data;
    }
private:
    std::shared_ptr<IEFileReader> ie_reader = nullptr;
    double ie_start_time = 0;
    double ie_end_time = 0;

    int count_ie = 0;

};

class DataReader
{
public:
    using Ptr = boost::shared_ptr<DataReader>;
    DataReader()
    {

    }

    ~DataReader() 
    {
        imu_reader.reset();
        wheel_reader.reset();
    }
    static DataReader::Ptr Instance();

    void Init(const std::string &baseDir)
    {
        imu_reader.reset(new IMUFileReader);
        wheel_reader.reset(new WheelFileReader);
        LoadIMUFile(baseDir + "/imu.txt", MODE_INNOVUSION);
        LoadWheelFile(baseDir + "/wheel.txt");
#ifdef TEST_ESKF
        GetFileNames(baseDir + "/CloudPredictPose.txt");
#else
        GetFileNames(baseDir + "/pcd/");
#endif
    }
    int LoadIMUFile(std::string imuFile, int type);
    int LoadWheelFile(std::string wheelFile);
    int GetFileNames(std::string path);

    void registImuCb(
            const boost::function<void(const IMURawData&)>& cb){
        imuCallback_ = cb;
    }

    void registWheelCb(
            const boost::function<void(const WheelRawData&)>& cb){
        wheelCallback_ = cb;
    }

    void registCloudCb(
            const boost::function<void(const PointCloudInter&, double)>& cb){
        cloudCallback_ = cb;
    }

    void process();
private:
    std::shared_ptr<IMUFileReader> imu_reader = nullptr;
    std::shared_ptr<WheelFileReader> wheel_reader = nullptr;
    double imu_start_time = 0;
    double imu_end_time = 0;
    double wheel_start_time = 0;
    double wheel_end_time = 0;

    int count_imu = 0;
    int count_wheel = 0;

    map<double, std::string> timePcdMap_;

    boost::function<void(const IMURawData&)> imuCallback_;
    boost::function<void(const WheelRawData&)> wheelCallback_;
    boost::function<void(const PointCloudInter&, double)> cloudCallback_;
};


#endif
