#ifndef RING_QUEUE_H
#define RING_QUEUE_H

#include <queue>
#include "glog/logging.h"

enum Direction{
    Back,
    Front
};

template <class T>
class RingQueue
{
public:

    RingQueue(size_t size){
        _size = 0;
        _maxsize = size;
        _curr = _rear = -1;
        _arr = new T[_maxsize];
    }

    ~RingQueue(){
        delete[] _arr;
    }

    RingQueue(const RingQueue &);

    void reset(){
        _curr = _rear = -1;
        _size = 0;
    }

    void enque(const T &value){
        if(is_full()){
            _size--;
        }
        _size++;
        auto dummy_rear = _rear;
        dummy_rear++;
        if(dummy_rear == _curr){
            _curr++;
            _curr %= _maxsize;
        }
        dummy_rear %= _maxsize;
        _arr[dummy_rear] = value;
        _rear = dummy_rear;
    }

    bool peek(double timestamp, T &value){
        T lastValue, currValue;
        lastValue.timestamp = -1;
        for(int i = 0; i <_size; i++){
            currValue = at_from_back(i);
            if(lastValue.timestamp > timestamp &&
                    currValue.timestamp <= timestamp){
                value = lastValue;
                return true;
            }
            lastValue = currValue;
        }
        return false;
    }

    bool getNext(T &value){
        if(_rear == _curr){
            return false;
        }
        _curr++;
        _curr %= _maxsize;
        value = _arr[_curr];
        return true;
    }

    bool getCurr(T &value){
        if(_curr < 0){
            return false;
        }
        value = _arr[_curr];
        return true;
    }

    void SwitchBack(const double &timestamp){
        if(_curr < 0){
            return;
        }
        size_t switch_size = 0;
        while(switch_size < _size){
            if(_arr[_curr].timestamp < timestamp){
                return;
            }
            _curr--;
            _curr += _maxsize;
            _curr %= _maxsize;
            switch_size++;
        }
    }

    void Shrink(const double &timestamp){
        if(_rear < 0){
            return;
        }
        size_t shrink_size = 0;
        while(shrink_size < _size){
            if(_arr[_rear].timestamp < timestamp){
                return;
            }
            _rear--;
            _rear += _maxsize;
            _rear %= _maxsize;
        }
    }

    size_t size() const{
        return _size;
    }

    size_t rear() const{
        return _rear;
    }

    T at_from_back(size_t index) const{
        DCHECK(index < _size);
        auto shiftedRear = _rear;
        if(_rear < index){
            shiftedRear = _rear + _maxsize;
        }
        auto queueIndex = shiftedRear - index;
        queueIndex %= _maxsize;
//        LOG(INFO) << "queueIndex: " << queueIndex;
        return  _arr[queueIndex];
    }

    T at_from_front(size_t index) const{
        DCHECK(index < _size);
        long front = _rear + _maxsize - _size;
        return  _arr[(front + index) % _maxsize];
    }

    bool is_empty() const{
        return _size == 0;
    }


    bool is_full() const{
        return _size == _maxsize;
    }

private:
    long _rear, _curr;

    size_t _size;
    size_t _maxsize;
    T *_arr;

};

#endif // RING_QUEUE_H
