﻿#ifndef UTILS_H
#define UTILS_H
#pragma once
#include <boost/shared_ptr.hpp>
#include <boost/thread.hpp>
#include <eigen3/Eigen/Geometry>
#include "glog/logging.h"
#include <map>
#include <unordered_map>
#include <iomanip>
#include <queue>
#include <thread>
#include <boost/make_shared.hpp>
#include <chrono>
#include <fstream>
#include <numeric>

using namespace std;
using namespace Eigen;
using namespace chrono;

#define TIMER_START(start) \
    auto start = std::chrono::system_clock::now();

#define TIMER_END(start, seconds)                                                       \
    auto end = std::chrono::system_clock::now();                                        \
    auto duration = std::chrono::duration_cast<std::chrono::microseconds>(end - start); \
    double seconds = double(duration.count()) / std::chrono::microseconds::period::den;


typedef boost::shared_ptr<boost::thread> ThreadPtr;

enum ListenerType{
    GNSS_SUBSCRIBER = 1,
    IMU_SUBSCRIBER,
    WHEEL_SUBSCRIBER,
    CLOUD_SUBSCRIBER
};

enum CommunacatorType{
    CALIB_ONLINE = 100,
    PREDICTER,
    CLOUD_LOCALIZER,
    INTERGRATED_POSE_CONTAINER,
    SYSTEM_CONTROLER,
    POSE_CONTAINER,
    ROS_PUBLISHER
};

enum MsgDataType{
    SensorMsg,
    ResultMsg,
    EventMsg
};

struct MsgData
{
    MsgData() {}
    virtual ~MsgData() {}
    MsgDataType msgDataType;
    double timestamp = -1;
};
typedef boost::shared_ptr<MsgData> MsgDataPtr;

enum SensorDataType{
    GNSS_DATA,
    IMU_DATA,
    WHEEL_DATA,
    CLOUD_DATA,
    IMU_POSE_DATA
};

struct SensorData : public MsgData
{
    SensorData() {}
    virtual ~SensorData() {}
    SensorDataType sensorDataType;
};
typedef boost::shared_ptr<SensorData> SensorDataPtr;

struct GNSSPacket : public SensorData
{
    GNSSPacket() {}
    Vector3d blh;
    uint8_t status;
};
typedef boost::shared_ptr<GNSSPacket> GNSSPacketPtr;

struct IMUPacket : public SensorData
{
    IMUPacket() {}
    Vector3f angular_velocity;
    Vector3f linear_acceleration;
};
typedef boost::shared_ptr<IMUPacket> IMUPacketPtr;

struct WheelPacket : public SensorData
{
    WheelPacket() {}
    Vector3d linear;
};
typedef boost::shared_ptr<WheelPacket> WheelPacketPtr;

struct ImuPosePacket : public SensorData
{
    ImuPosePacket() {}
    Quaterniond rotation;
};
typedef boost::shared_ptr<ImuPosePacket> ImuPosePacketPtr;



enum ResultDataType{
    ISOMETRY_DATA,
    CLOUDINFO_DATA
};

struct ResultData : public MsgData
{
    ResultData() {
        msgDataType = ResultMsg;
    }
    virtual ~ResultData() {}
    ResultDataType resultDataType;
};
typedef boost::shared_ptr<ResultData> ResultDataPtr;

struct IsometryData : public ResultData
{
    IsometryData() {
        resultDataType = ISOMETRY_DATA;
    }
    Isometry3d pose = Isometry3d::Identity();
};
typedef boost::shared_ptr<IsometryData> IsometryDataPtr;

enum EventType{
    MAP_LOADED,
    FST_MATCHED,
    LOCATE_FAIL_HIGH,
    LOCATE_FAIL_LOW,
    LOCATE_SUCCESS
};

struct EventData : public MsgData
{
    EventData() {
        msgDataType = EventMsg;
    }
    EventType eventType;
};
typedef boost::shared_ptr<EventData> EventDataPtr;

enum PoseUpdateType{
    FST_MATCH,
    UPDATE_INCREASE
};

enum ThreadMode{
    PRE_THREAD,
    WORK_THREAD,
    BACKEND_THREAD
};

enum MatchPrecisionState{
    HIGH_PRECISION = 1,
    FAST_PRECISION = 2,
    LOW_PRECISION = 3
};

template <typename T>
Eigen::Matrix<T, 3, 1> RotationQuaternionToEulerVector(
    const Eigen::Quaternion<T>& quaternion) {
  Eigen::Quaternion<T> n_q = quaternion.normalized();
  // We choose the quaternion with positive 'w', i.e., the one with a smaller
  // angle that represents this orientation.
  if (n_q.w() < 0.) {
    // Multiply by -1. http://eigen.tuxfamily.org/bz/show_bug.cgi?id=560
    n_q.w() = -1. * n_q.w();
    n_q.x() = -1. * n_q.x();
    n_q.y() = -1. * n_q.y();
    n_q.z() = -1. * n_q.z();
  }
  // We convert the normalized_quaternion(n_q) into a vector along the rotation
  // axis with length of the rotation angle.
  const T roll = atan2(2 * (n_q.w() * n_q.x() + n_q.y() * n_q.z()),
                       1 - 2 * (n_q.x() * n_q.x() + n_q.y() * n_q.y()));
  const T pitch = asin(2 * (n_q.w() * n_q.y() - n_q.z() * n_q.x()));
  const T yaw = atan2(2 * (n_q.w() * n_q.z() + n_q.x() * n_q.y()),
                      1 - 2 * (n_q.y() * n_q.y() + n_q.z() * n_q.z()));
  return Eigen::Matrix<T, 3, 1>(roll, pitch, yaw);
}

template <typename T>
Eigen::Quaternion<T> AngleAxisVectorToRotationQuaternion(
    const Eigen::Matrix<T, 3, 1>& angle_axis) {
  T scale = T(0.5);
  T w = T(1.);
  constexpr double kCutoffAngle = 1e-8;  // We linearize below this angle.
  if (angle_axis.squaredNorm() > kCutoffAngle) {
    const T norm = angle_axis.norm();
    scale = sin(norm / 2.) / norm;
    w = cos(norm / 2.);
  }
  const Eigen::Matrix<T, 3, 1> quaternion_xyz = scale * angle_axis;
  return Eigen::Quaternion<T>(w, quaternion_xyz.x(), quaternion_xyz.y(),
                              quaternion_xyz.z());
}

/// timer
class Timer {
   public:
    struct TimerRecord {
        TimerRecord() = default;
        TimerRecord(const std::string& name, double time_usage) {
            func_name_ = name;
            time_usage_in_ms_.emplace_back(time_usage);
        }
        std::string func_name_;
        std::vector<double> time_usage_in_ms_;
    };

    /**
     * call F and save its time usage
     * @tparam F
     * @param func
     * @param func_name
     */
    template <class F>
    static void Evaluate(F&& func, const std::string& func_name) {
        auto t1 = std::chrono::high_resolution_clock::now();
        std::forward<F>(func)();
        auto t2 = std::chrono::high_resolution_clock::now();
        auto time_used = std::chrono::duration_cast<std::chrono::duration<double>>(t2 - t1).count() * 1000;

        if (records_.find(func_name) != records_.end()) {
            records_[func_name].time_usage_in_ms_.emplace_back(time_used);
        } else {
            records_.insert({func_name, TimerRecord(func_name, time_used)});
        }
    }

    /// print the run time
    static void PrintAll() {
        LOG(INFO) << ">>> ===== Printing run time =====";
        for (const auto& r : records_) {
            LOG(INFO) << "> [ " << r.first << " ] average time usage: "
                      << std::accumulate(r.second.time_usage_in_ms_.begin(), r.second.time_usage_in_ms_.end(), 0.0) /
                             double(r.second.time_usage_in_ms_.size())
                      << " ms , called times: " << r.second.time_usage_in_ms_.size();
        }
        LOG(INFO) << ">>> ===== Printing run time end =====";
    }

    /// dump to a log file
    static void DumpIntoFile(const std::string& file_name) {
        std::ofstream ofs(file_name, std::ios::out);
        if (!ofs.is_open()) {
            LOG(ERROR) << "Failed to open file: " << file_name;
            return;
        } else {
            LOG(INFO) << "Dump Time Records into file: " << file_name;
        }

        size_t max_length = 0;
        for (const auto& iter : records_) {
            ofs << iter.first << ", ";
            if (iter.second.time_usage_in_ms_.size() > max_length) {
                max_length = iter.second.time_usage_in_ms_.size();
            }
        }
        ofs << std::endl;

        for (size_t i = 0; i < max_length; ++i) {
            for (const auto& iter : records_) {
                if (i < iter.second.time_usage_in_ms_.size()) {
                    ofs << iter.second.time_usage_in_ms_[i] << ",";
                } else {
                    ofs << ",";
                }
            }
            ofs << std::endl;
        }
        ofs.close();
    }

    /// get the average time usage of a function
    static double GetMeanTime(const std::string& func_name) {
        if (records_.find(func_name) == records_.end()) {
            return 0.0;
        }

        auto r = records_[func_name];
        return std::accumulate(r.time_usage_in_ms_.begin(), r.time_usage_in_ms_.end(), 0.0) /
               double(r.time_usage_in_ms_.size());
    }

    /// clean the records
    static void Clear() { records_.clear(); }

private:
    static std::map<std::string, TimerRecord> records_;
};

#endif // UTILS_H
