#include "filter.h"

namespace juefx{

Filter::Ptr filterInstance;

Filter::Filter()
    :eskfQueue_(3000)
{

}

void Filter::Reset()
{
    filterInstance.reset();
}

Filter::Ptr Filter::Instance()
{
    if(!filterInstance){
        LOG(INFO) << "new filterInstance";
        filterInstance.reset(new Filter());
    }
    return filterInstance;
}

bool Filter::Init()
{
    auto qrpconfig_loc = Parameter::Instance()->GetBaseDir() + "/qrpconfig_auto.txt";
    QRPConfigFileReader file_qrpconfig;
    auto fp_qrp = fopen(qrpconfig_loc.c_str(), "r");
    if(fp_qrp == NULL){
        LOG(INFO) << qrpconfig_loc + " open fail!";
        return false;
    }
    file_qrpconfig.loadQRPConfigAutoFile(fp_qrp);
    qrpconfig_data = file_qrpconfig.auto_data;

//    Vector3f gravity = {0, 0, -9.8015};
    Vector3f gravity = Parameter::Instance()->GetGravity();

    Vector3f init_pos = {0, 0, 0};
    Vector3f init_velo = {0, 0, 0};
    Quaternionf init_q(Parameter::Instance()->GetEskfInitPose().linear());
    // 141 测试过 (X=0.000208758, Y=0.001560938 Z=0.001652822, 静态陀螺零偏）
    Vector3f init_acc_bias(0.399642704,-0.08396834,0.047);//
    Vector3f init_gyr_bias(0.00000348,0.000153993,-0.00007214);

    /*////////Q阵//////////////////////////////////////*/
    float sigma_accl_noise; // [m/s^2] (value derived from Noise Spectral Density in datasheet)
    float sigma_gyro_noise; // [rad/s] (value derived from Noise Spectral Density in datasheet)
    float sigma_accl_drift; // [m/s^2 sqrt(s)] (Educated guess, real value to be measured)
    float sigma_gyro_drift; // [rad/s sqrt(s)] (Educated guess, real value to be measured)
    /*///////P阵////*/
    float sigma_init_pos;//(0.1*0.1,0.1*0.1,0.3*0.3); // [m]
    float sigma_init_vel; // [m/s]
    float sigma_init_dtheta; // [rad]
    float sigma_init_accel_bias;//=0.0005*sigma_accel_drift; // [m/s^2]
    float sigma_init_gyro_bias;//=0.0005*sigma_gyro_drift; // [rad/s]
    ////////R阵/////////
    Vector3f sigma_pos;
    Vector3f sigma_theta;
    Vector3f sigma_vel;
    int lidar_status=1;
    float wheel_vel_range=0;
    ////////imu-陀螺-加速度计零偏///////////////////////--(X=0.146914237, Y=0.028846492 Z=0.009 静态加速度数据求得零偏）
    //////////imu-陀螺-加速度计零偏///////////////////////////////////////////

    ////////////量测R阵的初始化///////////////////////////////
    sigma_pos[0]  = pow(qrpconfig_data[0].sigma_pos_x[0],2);
    sigma_pos[1]  = pow(qrpconfig_data[0].sigma_pos_y[0],2);
    sigma_pos[2]  = pow(qrpconfig_data[0].sigma_pos_z[0],2);

    sigma_theta[0]  = pow(qrpconfig_data[0].sigma_theta_roll[0],2);
    sigma_theta[1]  = pow(qrpconfig_data[0].sigma_theta_pitch[0],2);
    sigma_theta[2]  = pow(qrpconfig_data[0].sigma_theta_yaw[0],2);

    sigma_vel[0]    = pow(qrpconfig_data[0].sigma_vel_x[0],2);
    sigma_vel[1]    = pow(qrpconfig_data[0].sigma_vel_y[0],2);
    sigma_vel[2]    = pow(qrpconfig_data[0].sigma_vel_z[0],2);
    ////////////Q阵/--
    sigma_accl_noise = pow(qrpconfig_data[0].sigma_accl_noise,2);   //0.6   三个方向的Q值还是不一样的设置最好
    sigma_gyro_noise = pow(qrpconfig_data[0].sigma_gyro_noise,2);  //0.01，0.0085
    sigma_accl_drift = pow(0.001,2)*sigma_accl_noise;//
    sigma_gyro_drift = pow(0.001,2)*sigma_gyro_noise;//
     /////////////P阵的初始化////////////////////////////////////////////
    sigma_init_pos        = pow(qrpconfig_data[0].sigma_init_pos,2);    // pow(0.3,2)
    sigma_init_vel        = pow(qrpconfig_data[0].sigma_init_vel,2);    //   pow(5.35,2)
    sigma_init_dtheta     = pow(qrpconfig_data[0].sigma_init_dtheta,2); //pow(0.02,2)
    sigma_init_accel_bias = 10*sigma_accl_drift; //[m/s^2]10*sigma_accl_drift;
    sigma_init_gyro_bias  = 10*sigma_gyro_drift; //[m/s^2]10*sigma_gyro_drift;
    ////////////////////////////////////////////////////

    LOG(INFO) << "init_q: " << init_q.coeffs();
    eskf_.reset(new ESKF(
                gravity, // Acceleration due to gravity in global frame
                ESKF::makeState(
                        init_pos, // init pos
                        init_velo, // init vel
                        init_q, // init quaternion
                        init_acc_bias, // init accel bias
                        init_gyr_bias // init gyro bias
                ),
                ESKF::makeP( //P
                    SQ(sigma_init_pos) * I_3,
                    SQ(sigma_init_vel) * I_3,
                    SQ(sigma_init_dtheta) * I_3,
                    SQ(sigma_init_accel_bias) * I_3,
                    SQ(sigma_init_gyro_bias) * I_3
                ),
                    //Q
                    sigma_accl_noise,
                    sigma_gyro_noise,
                    sigma_accl_drift,
                    sigma_gyro_drift,
                    sigma_pos,
                    sigma_theta,
                    sigma_vel,
                    lidar_status,
                    wheel_vel_range,
                    ESKF::delayTypes::noMethod,100,
                    Parameter::Instance()->GetArm().cast<float>(),
                    &qrpconfig_data[0]));
    LOG(INFO) << "eskf init done";
    return true;
}

void Filter::UpdateImu(
        const Vector3f &acc, const Vector3f &gryo, lTime now)
{
    if(!eskf_){
        return;
    }
    eskf_->updateImu(acc, gryo, now);
}

void Filter::MeasureWheel(
        const Vector3f &wheel_vel, double speed_range, lTime now)
{
    if(!eskf_){
        return;
    }
    eskf_->measureWheel(wheel_vel, speed_range, now);
}

void Filter::MeasureLidar(
        const Isometry3f &pose, double timestamp, MatchPrecisionState precisionState)
{
    if(!SwitchBack(timestamp)){
        LOG(WARNING) << "MeasureLidar fail: SwitchBack fail!";
        return;
    }
    Quaternionf quat(pose.linear());
    ESKF::measureStatus measureStatus = ESKF::measureStatus::THETA_Yaw;
    if(HIGH_PRECISION == precisionState){
        measureStatus = ESKF::measureStatus::THETA_ALL;
    };
    eskf_->measureLidar(quat, pose.translation(),
                        precisionState, lTime(timestamp), measureStatus);
}

bool Filter::GetStateTransform(IsometryData &pose)
{
    ESKF::FullState state;
    eskf_->getFullState(state);
    if(state.current_time <= 0){
        return false;
    }
    EskfState fullState;
    fullState.eskf = *eskf_;
    fullState.timestamp = state.current_time;
    eskfQueue_.enque(fullState);
    pose.timestamp = state.current_time;
    pose.pose.translation() = state.getPosition().cast<double>();
    pose.pose.linear() = state.getRotation().toRotationMatrix();
//    Quaterniond resultQ(pose.pose.linear());
//    auto stateRpy = RotationQuaternionToEulerVector(
//                Quaterniond(resultQ));
//    std::ofstream ofs(Parameter::Instance()->GetBaseDir() + "/predictPose.txt", ios::app);
//    ofs << setprecision(15) << pose.timestamp << ", " << pose.pose.translation().x()
//        << ", " << pose.pose.translation().y() << ", " << pose.pose.translation().z()
//        << ", " << resultQ.w() << ", " << resultQ.x() << ", "
//        << resultQ.y() << ", " << resultQ.z() << ", " << stateRpy.x() << ", "
//        << stateRpy.y() << ", " << stateRpy.z() << endl;
//    ofs.close();
    return true;
}

bool Filter::SwitchBack(double timestamp)
{
    EskfState fullState;
    auto size = eskfQueue_.size();
    bool found = false;
    for(int i = 0; i < size; i++){
        fullState = eskfQueue_.at_from_back(i);
        if(fullState.timestamp < timestamp){
            found = true;
            break;
        }
    }
    if(!found){
        return false;
    }
    *eskf_ = fullState.eskf;
    ESKF::FullState state;
    eskf_->getFullState(state);
    auto stateRpy = RotationQuaternionToEulerVector(
                Quaterniond(state.getRotation().toRotationMatrix()));
//    std::ofstream ofs(Parameter::Instance()->GetBaseDir() + "/predictPose.txt", ios::app);
    LOG(INFO) << "SwitchBack: " << setprecision(15) << state.current_time << ", " << state.getPosition().transpose()
        << ", " << ", " << stateRpy.x() << ", "
        << stateRpy.y() << ", " << stateRpy.z() << endl;
//    ofs.close();
    return true;
}

} // end of namespace
