#include "poseguesser.h"
#include "pcl/registration/gicp.h"

namespace juefx{

PoseGuesser::Ptr poseGuesserInstance;

PoseGuesser::PoseGuesser()
    :octree_(2)
{

}

void PoseGuesser::Reset()
{
    poseGuesserInstance.reset();
}

PoseGuesser::Ptr PoseGuesser::Instance()
{
    if(!poseGuesserInstance){
        poseGuesserInstance.reset(new PoseGuesser());
    }
    return poseGuesserInstance;
}

void PoseGuesser::SetMapName(const string &mapName)
{
    mapName_ = mapName;
}

bool PoseGuesser::LoadSlamTraj(
        const string &slamTrajPath, const string &gnssPath)
{
    vector<IsometryData> trajPoseVec;
    vector<IsometryData> blhVec;
    ifstream ifs(slamTrajPath);
    if(ifs) {
        string line;
        getline(ifs, line);
        while(getline(ifs, line))
        {
            IsometryData pose;
            Quaterniond q;
            std::istringstream iss(line);
            string subline;
            getline(iss, subline, ',');
            std::istringstream(subline) >> pose.timestamp;
            getline(iss, subline, ',');
            std::istringstream(subline) >> pose.pose.translation().x();
            getline(iss, subline, ',');
            std::istringstream(subline) >> pose.pose.translation().y();
            getline(iss, subline, ',');
            std::istringstream(subline) >> pose.pose.translation().z();
            getline(iss, subline, ',');
            std::istringstream(subline) >> q.w();
            getline(iss, subline, ',');
            std::istringstream(subline) >> q.x();
            getline(iss, subline, ',');
            std::istringstream(subline) >> q.y();
            getline(iss, subline, ',');
            std::istringstream(subline) >> q.z();
            pose.pose.linear() = q.toRotationMatrix();
            trajPoseVec.push_back(pose);
        }
        ifs.close();
    }else{
        return false;
    }

    ifs.open(gnssPath);
    if(ifs) {
        string line;
        getline(ifs, line);
        while(getline(ifs, line))
        {
            IsometryData blh;
            std::istringstream iss(line);
            string subline;
            getline(iss, subline, ',');
            getline(iss, subline, ',');
            std::istringstream(subline) >> blh.pose.translation().y();
            getline(iss, subline, ',');
            std::istringstream(subline) >> blh.pose.translation().x();
            getline(iss, subline, ',');
            std::istringstream(subline) >> blh.pose.translation().z();
            blhVec.push_back(blh);
        }
        ifs.close();
    }else{
        return false;
    }

    if(0 == blhVec.size()){
        return false;
    }
    ConstractLocalTraj(blhVec, trajPoseVec);
    CalGnssLocal2Map();
    return true;
}

float PoseGuesser::GetMapPoseByBlh(
        const Vector3d &blh, Isometry3d& mapPose)
{
    if(!proj_){
        return numeric_limits<float>::max();
    }
    Vector3d localT;
    proj_->Forward(blh.x(), blh.y(), blh.z(),
                   localT.x(),
                   localT.y(),
                   localT.z());
    localT.z() = 0;
    pcl::PointXYZ query;
    query.x = localT.x();
    query.y = localT.y();
    query.z = localT.z();
    int result_index;
    float sqr_distance;
    octree_.approxNearestSearch(query, result_index, sqr_distance);
    double distance = sqrt(sqr_distance);
    mapPose.translation() = gnssLocal2Map_ * localT;
    mapPose.linear() = interpedTrajPoseVec_.at(result_index).linear()
            * Matrix3d(AngleAxisd(M_PI / 2, Vector3d::UnitZ()));    //flu to rfu

    auto rpy = RotationQuaternionToEulerVector(
                Quaterniond(mapPose.linear()));
    LOG(INFO) << " mapPose: " << mapPose.translation().transpose()
              << " rpy: " << rpy.transpose()
              << " distance: " << distance;
    return distance;
}

bool PoseGuesser::InterpBlh(
        const RingQueue<GNSSPacket> *gnssQueue,
        double timestamp,
        Vector3d& interpedBlh)
{
    size_t size = gnssQueue->size();
    if(size < 2){
        return false;
    }
    GNSSPacket lastGnss, currGnss;
    bool found = false;
    for(size_t i = 0; i < size; i++){
        currGnss = gnssQueue->at_from_back(i);
        if(currGnss.timestamp <= timestamp){
            if(i == 0){
                lastGnss = currGnss;
                currGnss = gnssQueue->at_from_back(1);
            }
            found = true;
            break;
        }
        lastGnss = currGnss;
    }
    if(found){
        double deltTime = lastGnss.timestamp - currGnss.timestamp;
        double timeDiff = timestamp - currGnss.timestamp;
        double radio = timeDiff / deltTime;
        interpedBlh = currGnss.blh + (lastGnss.blh - currGnss.blh) * radio;
    }
    return found;
}

size_t periodPoseId = 1;
inline Isometry3d interpFromMapPoseVec(
        const double &time,
        const vector<IsometryData> *periodPoseVec){
    size_t size = periodPoseVec->size();
    for(; periodPoseId < size; periodPoseId++){
        const IsometryData& pre = periodPoseVec->at(periodPoseId - 1);
        const IsometryData& pro = periodPoseVec->at(periodPoseId);
        if(periodPoseId == 1 && time < pre.timestamp){
            return pre.pose;
        }else if(time >= pre.timestamp &&
                  time < pro.timestamp){
              double deltTime = pro.timestamp - pre.timestamp;
              double timeDiff = time - pre.timestamp;
              double radio = timeDiff / deltTime;
              Isometry3d interpedMapPose;
              interpedMapPose.translation() = pre.pose.translation() + (pro.pose.translation() - pre.pose.translation()) * radio;
              interpedMapPose.linear() =
                      Quaterniond(pre.pose.linear()).slerp(radio, Quaterniond(pro.pose.linear())).toRotationMatrix();
              return interpedMapPose;
        }
    }
    return periodPoseVec->back().pose;
}

void PoseGuesser::ConstractLocalTraj(const vector<IsometryData> &blhVec,
                                     const vector<IsometryData> &trajPoseVec)
{
    proj_.reset(new LocalCartesian(blhVec.front().pose.translation().x(),
                                   blhVec.front().pose.translation().y(),
                                   blhVec.front().pose.translation().z()));
    for(const auto &blh : blhVec){
        Vector3d localT;
        proj_->Forward(blh.pose.translation().x(),
                       blh.pose.translation().y(),
                       blh.pose.translation().z(),
                       localT.x(), localT.y(), localT.z());
        pcl::PointXYZ localTrajPoint(localT.x(), localT.y(), localT.z());
        Isometry3d interpedMapPose = interpFromMapPoseVec(blh.timestamp, &trajPoseVec);
        interpedTrajPoseVec_.push_back(interpedMapPose);
        localTrajPoint.z = interpedMapPose.translation().z();
        gnssLocalTraj_.push_back(localTrajPoint);
    }
    octree_.setInputCloud(boost::make_shared<PointCloudXYZ>(gnssLocalTraj_));
    octree_.addPointsFromInputCloud();
}

void PoseGuesser::CalGnssLocal2Map()
{
    PointCloudXYZ mapTrajCloud;
    for(const auto &trajPose : interpedTrajPoseVec_){
        pcl::PointXYZ mapTraj;
        mapTraj.x = trajPose.translation().x();
        mapTraj.y = trajPose.translation().y();
        mapTraj.z = trajPose.translation().z();
        mapTrajCloud.push_back(mapTraj);
    }
    pcl::GeneralizedIterativeClosestPoint<pcl::PointXYZ, pcl::PointXYZ> icp;
    icp.setTransformationEpsilon(0.0000000001);
    icp.setMaxCorrespondenceDistance(2);
    icp.setCorrespondenceRandomness(40);
    icp.setMaximumIterations(80);
    icp.setRANSACIterations(0);
    icp.setInputSource(boost::make_shared<PointCloudXYZ>(gnssLocalTraj_));
    icp.setInputTarget(boost::make_shared<PointCloudXYZ>(mapTrajCloud));
    pcl::PointCloud<pcl::PointXYZ> unused_result;
    icp.align(unused_result);
    Matrix4d T = icp.getFinalTransformation().cast<double>();
    gnssLocal2Map_.translation() = T.block(0, 3, 3, 1);
    gnssLocal2Map_.linear() = T.block(0, 0, 3, 3);
    auto rpy = RotationQuaternionToEulerVector(
                Quaterniond(gnssLocal2Map_.linear()));
    LOG(INFO) << " gnssLocal2Map_: " << gnssLocal2Map_.translation().transpose()
              << " rpy: " << rpy.transpose();
}

} // end of namespace
