#include "ceres_matcher.h"

#include "cartographer/sensor/internal/voxel_filter.h"
#include "cartographer/mapping/internal/3d/scan_matching/ceres_scan_matcher_3d.h"
#include "cartographer/mapping/probability_values.h"

#include "utils/timer.h"
#include "utils/rigid3.h"

using CeresScanMatcher3D = cartographer::mapping::scan_matching::CeresScanMatcher3D;
using HybridGrid = cartographer::mapping::HybridGrid;

#define HIT_SCORE 0.75

namespace juefx {

CeresMatcher::CeresMatcher(const CeresMatcherOption& option) : options_(option) {
    cartographer::common::proto::CeresSolverOptions option_ceres;
    option_ceres.set_max_num_iterations(50);
    option_ceres.set_use_nonmonotonic_steps(false);
    option_ceres.set_num_threads(option.threads_num);
    cartographer::mapping::scan_matching::proto::CeresScanMatcherOptions3D option_3d;
    *option_3d.mutable_ceres_solver_options() = option_ceres;
    option_3d.add_occupied_space_weight(2.5);
    option_3d.add_occupied_space_weight(4);
    option_3d.set_odom_rotation_weight(0);
    option_3d.set_odom_rotation_weight_yaw(0);
    option_3d.set_only_optimize_yaw(false);
    option_3d.set_rotation_weight(option.R_weight_pitchroll);
    option_3d.set_rotation_weight_roll(option.R_weight_pitchroll);
    option_3d.set_rotation_weight_yaw(option.R_weight_yaw);
    option_3d.set_translation_weight(option.T_weight_xy);
    option_3d.set_translation_weight_z(option.T_weight_z);
    CeresScanMatcher3D *inner_matcher = new CeresScanMatcher3D(option_3d);
    matcher_ = (ScanMatcher *)inner_matcher;
}

CeresMatcher::~CeresMatcher() {
    CeresScanMatcher3D *inner_matcher =(CeresScanMatcher3D*)matcher_;
    delete inner_matcher;
    FreeGrids();
}

void CeresMatcher::LoadPointCloudMap(const PointCloudJ::Ptr pcloud_map, 
                                     float high_resolution, float low_resolution) {
    FreeGrids();
    LOG(INFO) << "Load pcloud map ... resolution " << high_resolution << "/" << low_resolution;
    TIMER_START(start);
    BuildGird(pcloud_map, high_resolution);
    BuildGird(pcloud_map, low_resolution);
    TIMER_END(start, seconds);
    LOG(INFO) << "Load pcloud map finished, cost " << seconds << " sec.";
}

void CeresMatcher::BuildGird(const PointCloudJ::Ptr pcloud_map, float resolution) {
    HybridGrid *inner_grid = new HybridGrid(resolution);
    const std::vector<uint16_t> hit_table =
        cartographer::mapping::ComputeLookupTableToApplyOdds(cartographer::mapping::Odds(HIT_SCORE));
    for (const auto &point : pcloud_map->points)
    {
        Eigen::Vector3f position(point.x, point.y, point.z);
        const Eigen::Array3i hit_cell = inner_grid->GetCellIndex(position);
        inner_grid->ApplyLookupTable(hit_cell, hit_table);
    }
    inner_grid->FinishUpdate();
    map_grids_.push_back((Grid *)inner_grid);
}

void CeresMatcher::FreeGrids() {
    if (map_grids_.size() > 0)
    {
        for (auto* grid : map_grids_)
        {
            HybridGrid *inner_grid = (HybridGrid *)grid;
            delete inner_grid;
        }
        map_grids_.clear();
    }
}

float CeresMatcher::MatchPointCloud(const PointCloudJ::Ptr pcloud, const Eigen::Isometry3f init_pose,
                                    Eigen::Isometry3f& final_pose) {
    if (map_grids_.size() == 0)
        return -1;

    CeresScanMatcher3D *inner_matcher = (CeresScanMatcher3D *)matcher_;
    const HybridGrid *high_grid = (const HybridGrid *)(map_grids_[0]);
    const HybridGrid *low_grid = (const HybridGrid *)(map_grids_[1]);
    TIMER_START(start);
    int ceres_type = CERES_MODE_BASE;
    Rigid3d pose_predict(init_pose.translation().cast<double>(),
                         Eigen::AngleAxisd(init_pose.rotation().cast<double>()));
    cartographer::sensor::PointCloud raw_cloud;
    raw_cloud.reserve(pcloud->size());
    for (const auto& point : pcloud->points) {
        cartographer::sensor::RangefinderPoint p;
        p.position = Eigen::Vector3f(point.x, point.y, point.z);
        raw_cloud.push_back(p);
    }

    cartographer::sensor::VoxelFilter voxel_filter(options_.voxel_filter_size);
    auto vector_cloud = voxel_filter.Filter(raw_cloud);
    auto low_cloud = cartographer::sensor::FilterByMaxRange(vector_cloud, options_.max_range);

    Rigid3d pose_in_map;
    ceres::Solver::Summary summary;
    Eigen::Vector4d loss = inner_matcher->Match(
        pose_predict.translation(), pose_predict, NULL, {{&low_cloud, high_grid}, {&low_cloud, low_grid}},
        &pose_in_map, &summary, ceres_type);
    rigid3d_to_isometry3f(pose_in_map, final_pose);

    char buf[1024];
    Eigen::Vector3d euler3 = 
            cartographer::transform::RotationQuaternionToEulerVector(pose_in_map.rotation());
    snprintf(buf, 1023, "{ pos: %.3f %.3f %.3f, euler: %.5f %.5f %.5f }",
             pose_in_map.translation().x(), pose_in_map.translation().y(), pose_in_map.translation().z(),
             euler3.x(), euler3.y(), euler3.z());
    TIMER_END(start, seconds);
    LOG(INFO) << "Match scan with map, voxel_filter: " << options_.voxel_filter_size << " loss: " << loss[0]
              << ", transform: " << buf << ", cost " << seconds << " sec.";
    return loss[0];
}

}