#include "fast_ceres_matcher.h"

#include "cartographer/sensor/internal/voxel_filter.h"
#include "cartographer/mapping/internal/3d/scan_matching/ceres_scan_matcher_3d.h"
#include "cartographer/mapping/internal/3d/scan_matching/fast_correlative_scan_matcher_3d.h"
#include "cartographer/mapping/probability_values.h"

#include "utils/timer.h"
#include "utils/rigid3.h"

using FastScanMatcher3D = cartographer::mapping::scan_matching::FastCorrelativeScanMatcher3D;
using CeresScanMatcher3D = cartographer::mapping::scan_matching::CeresScanMatcher3D;
using HybridGrid = cartographer::mapping::HybridGrid;

#define SCORE_SCALE 0.5

namespace juefx {

FastCeresMatcher::FastCeresMatcher(const FastCeresMatcherOption& option)
        : matcher_1_(NULL), options_(option) {
    cartographer::common::proto::CeresSolverOptions option_ceres;
    option_ceres.set_max_num_iterations(50);
    option_ceres.set_use_nonmonotonic_steps(false);
    option_ceres.set_num_threads(option.threads_num);
    cartographer::mapping::scan_matching::proto::CeresScanMatcherOptions3D option_3d;
    *option_3d.mutable_ceres_solver_options() = option_ceres;
    option_3d.add_occupied_space_weight(2.5);
    option_3d.add_occupied_space_weight(4);
    option_3d.set_odom_rotation_weight(0);
    option_3d.set_odom_rotation_weight_yaw(0);
    option_3d.set_only_optimize_yaw(false);
    option_3d.set_rotation_weight(option.R_weight_pitchroll);
    option_3d.set_rotation_weight_roll(option.R_weight_pitchroll);
    option_3d.set_rotation_weight_yaw(option.R_weight_yaw);
    option_3d.set_translation_weight(option.T_weight_xy);
    option_3d.set_translation_weight_z(option.T_weight_z);
    CeresScanMatcher3D *ceres_matcher = new CeresScanMatcher3D(option_3d);
    matcher_2_ = (ScanMatcher *)ceres_matcher;
}

FastCeresMatcher::~FastCeresMatcher()
{
    if (matcher_1_ != NULL) {
        FastScanMatcher3D *fast_matcher = (FastScanMatcher3D *)matcher_1_;
        delete fast_matcher;
    }
    CeresScanMatcher3D *ceres_matcher = (CeresScanMatcher3D *)matcher_2_;
    delete ceres_matcher;
    FreeGrids();
}

void FastCeresMatcher::LoadPointCloudMap(const PointCloudJ::Ptr pcloud_map, 
                                         float high_resolution, float low_resolution, float hit_score) {
    FreeGrids();
    LOG(INFO) << "Load pcloud map ... resolution " << high_resolution << "/" << low_resolution
              << ", hit_score " << hit_score;
    TIMER_START(start);
    hit_score_ = hit_score;
    BuildGird(pcloud_map, high_resolution, hit_score);
    BuildGird(pcloud_map, low_resolution, hit_score);
    if (matcher_1_ != NULL) {
        FastScanMatcher3D *fast_matcher = (FastScanMatcher3D *)matcher_1_;
        delete fast_matcher;
    }
    const HybridGrid *high_grid = (const HybridGrid *)(map_grids_[0]);
    const HybridGrid *low_grid = (HybridGrid *)(map_grids_[1]);
    cartographer::mapping::scan_matching::proto::FastCorrelativeScanMatcherOptions3D fast_option;
    fast_option.set_branch_and_bound_depth(5); // 2 ** 4 = 16 * 0.09 = 1.44
    fast_option.set_full_resolution_depth(3);  // 2 ** 3 = 8 * 0.09 = 0.72+ as low resolution
    fast_option.set_min_low_resolution_score(0.5f * SCORE_SCALE);
    fast_option.set_min_rotational_score(0.f);
    fast_option.set_linear_xy_search_window(options_.XY_search_window);
    fast_option.set_linear_z_search_window(0.5f);   // search z for bad calib
    fast_option.set_angular_search_window(options_.Yaw_search_window);
    fast_option.set_global_search_ratio(1.f);
    Eigen::VectorXf rotational_vector;
    FastScanMatcher3D *fast_matcher = new FastScanMatcher3D(*high_grid, low_grid, &rotational_vector, fast_option);
    matcher_1_ = (ScanMatcher *)fast_matcher;
    TIMER_END(start, seconds);
    LOG(INFO) << "Load pcloud map & build index finished, cost " << seconds << " sec.";
}

void FastCeresMatcher::BuildGird(const PointCloudJ::Ptr pcloud_map, float resolution, float hit_score) {
    HybridGrid *inner_grid = new HybridGrid(resolution);
    const std::vector<uint16_t> hit_table =
        cartographer::mapping::ComputeLookupTableToApplyOdds(cartographer::mapping::Odds(hit_score));
    for (const auto &point : pcloud_map->points)
    {
        Eigen::Vector3f position(point.x, point.y, point.z);
        const Eigen::Array3i hit_cell = inner_grid->GetCellIndex(position);
        inner_grid->ApplyLookupTable(hit_cell, hit_table);
    }
    inner_grid->FinishUpdate();
    map_grids_.push_back((Grid *)inner_grid);
}

void FastCeresMatcher::FreeGrids() {
    if (map_grids_.size() > 0)
    {
        for (auto* grid : map_grids_)
        {
            HybridGrid *inner_grid = (HybridGrid *)grid;
            delete inner_grid;
        }
        map_grids_.clear();
    }
}

float FastCeresMatcher::MatchPointCloud(const PointCloudJ::Ptr pcloud, const Eigen::Isometry3f init_pose,
                                        Eigen::Isometry3f& final_pose) {
    if (map_grids_.size() == 0)
        return -1;

    TIMER_START(start);

    cartographer::sensor::PointCloud raw_cloud;
    raw_cloud.reserve(pcloud->size());
    for (const auto& point : pcloud->points) {
        cartographer::sensor::RangefinderPoint p;
        p.position = Eigen::Vector3f(point.x, point.y, point.z);
        raw_cloud.push_back(p);
    }

    cartographer::sensor::VoxelFilter voxel_filter(options_.voxel_filter_size);
    auto vector_cloud = voxel_filter.Filter(raw_cloud);
    auto low_cloud = cartographer::sensor::FilterByMaxRange(vector_cloud, options_.max_range);
    LOG(INFO) << "origin cloud size: " << raw_cloud.size()
              << ", after downsample " << options_.voxel_filter_size << ": " << vector_cloud.size()
              << ", after limit " << options_.max_range << ": " << low_cloud.size();

    Rigid3f pose_predict(init_pose.translation(),
                         Eigen::AngleAxisf(init_pose.rotation()));

    FastScanMatcher3D *fast_matcher = (FastScanMatcher3D *)matcher_1_;
    const auto param = fast_matcher->GetSearchParameters();
    Eigen::VectorXf rotational_vector;
    auto fast_result =
        fast_matcher->MatchWithSearchParameters(0, param, pose_predict, Rigid3f::Identity(),
                                                low_cloud, rotational_vector, Eigen::Quaterniond::Identity(),
                                                0.5f * SCORE_SCALE);
    {
        if (hit_score_ >= 0.9)
            fast_result->score /= 0.9f;
        char buf[1024];
        Rigid3d &p = fast_result->pose_estimate;    // Trans from estimate to base
        Eigen::Vector3d euler3 =
            cartographer::transform::RotationQuaternionToEulerVector(p.rotation());
        snprintf(buf, 1023, "{ pos: %.3f %.3f %.3f, euler: %.5f %.5f %.5f }",
                 p.translation().x(), p.translation().y(), p.translation().z(),
                 euler3.x(), euler3.y(), euler3.z());
        TIMER_END(start, seconds);
        LOG(INFO) << "prematch finish, score: " << fast_result->score 
                  << ", transform: " << buf << ", cost " << seconds << " sec.";
        rigid3d_to_isometry3f(p, last_search_pose_);
    }

    const HybridGrid *high_grid = (HybridGrid *)(map_grids_[0]);
    const HybridGrid *low_grid = (HybridGrid *)(map_grids_[1]);
    CeresScanMatcher3D *ceres_matcher = (CeresScanMatcher3D *)matcher_2_;
    int ceres_type = CERES_MODE_BASE;

    Rigid3d pose_in_map;
    ceres::Solver::Summary summary;
    Eigen::Vector4d loss = ceres_matcher->Match(
        fast_result->pose_estimate.translation(), fast_result->pose_estimate, NULL,
        {{&low_cloud, high_grid}, {&low_cloud, low_grid}},
        &pose_in_map, &summary, ceres_type);
    rigid3d_to_isometry3f(pose_in_map, final_pose);

    char buf[1024];
    Eigen::Vector3d euler3 = 
            cartographer::transform::RotationQuaternionToEulerVector(pose_in_map.rotation());
    snprintf(buf, 1023, "{ pos: %.3f %.3f %.3f, euler: %.5f %.5f %.5f }",
             pose_in_map.translation().x(), pose_in_map.translation().y(), pose_in_map.translation().z(),
             euler3.x(), euler3.y(), euler3.z());
    TIMER_END(start, seconds);
    LOG(INFO) << "Match scan with map, voxel_filter: " << options_.voxel_filter_size << " loss: " << loss[0] 
              << ", transform: " << buf << ", cost " << seconds << " sec.";
    return loss[0];
}

}