#pragma once

#include <Eigen/Core>
#include <Eigen/Dense>
#include <Eigen/Geometry>
#include <pcl/point_cloud.h>
#include <pcl/point_types.h>

#include "common/nanoflann.hpp"

#define RAW_ICP_DEBUG

namespace juefx
{

struct VectorCloudPoint
{
    Eigen::Vector3d P;
    Eigen::Vector3d N;
    float distance;
    float curvature;
};
typedef std::vector<VectorCloudPoint> VectorCloud;

bool transfer_cloud_to_vector(pcl::PointCloud<pcl::PointXYZI>::ConstPtr pcloud, VectorCloud &vector_cloud);

struct FrameCloud
{
    int8_t frame_index;
    int8_t lidar_id;
    VectorCloud raw_pts;    // raw cloud used in optimization
    VectorCloud pts;        // transfered cloud used in KNN search

    typedef nanoflann::KDTreeSingleIndexAdaptor<
        nanoflann::L2_Simple_Adaptor<double, FrameCloud>, FrameCloud, 3 /* dim */
        >
        MyKdTreeT;
    MyKdTreeT *kdtree;

    FrameCloud(int8_t frame_index, int8_t lidar_id)
        : frame_index(frame_index), lidar_id(lidar_id)
    {
        kdtree = NULL;
    }

    void buildKDTree()
    {
        kdtree = new MyKdTreeT(
            3 /*dim*/, *this,
            nanoflann::KDTreeSingleIndexAdaptorParams(1 /* max leaf */));
        kdtree->buildIndex();
    }

    double getClosestPoint(const Eigen::Vector3d &query_pt, size_t &ret_index) const
    {
        const size_t num_results = 1;
        double out_dist_sqr;
        nanoflann::KNNResultSet<double> resultSet(num_results);
        resultSet.init(&ret_index, &out_dist_sqr);
        kdtree->findNeighbors(resultSet, &query_pt[0],
                                nanoflann::SearchParams(32, 0, false));
        return out_dist_sqr;
    }

    void resetKDTree()
    {
        if (kdtree != NULL)
        {
            delete (kdtree);
            kdtree = NULL;
        }
    }

    // Must return the number of data points
    inline size_t kdtree_get_point_count() const { return pts.size(); }

    // Returns the distance between the vector "p1[0:size-1]" and the data point
    // with index "idx_p2" stored in the class:
    inline double kdtree_distance(const double *p2, const size_t idx_p1,
                                    size_t /*size*/) const
    {
        const double d0 = p2[0] - pts[idx_p1].P.x();
        const double d1 = p2[1] - pts[idx_p1].P.y();
        const double d2 = p2[2] - pts[idx_p1].P.z();
        const double square_d = d0 * d0 + d1 * d1 + d2 * d2;
        if (square_d > 0.4 * 0.4)
            return square_d;
        else
        {
            Eigen::Vector3d d(d0, d1, d2);
            double point_to_plane = d.dot(pts[idx_p1].N);
            return point_to_plane * point_to_plane;
        }
    }

    // Returns the dim'th component of the idx'th point in the class:
    // Since this is inlined and the "dim" argument is typically an immediate
    // value, the
    //  "if/else's" are actually solved at compile time.
    inline double kdtree_get_pt(const size_t idx, int dim) const
    {
        if (dim == 0)
            return pts[idx].P.x();
        else if (dim == 1)
            return pts[idx].P.y();
        else
            return pts[idx].P.z();
    }

    // Optional bounding-box computation: return false to default to a standard
    // bbox computation loop.
    //   Return true if the BBOX was already computed by the class and returned in
    //   "bb" so it can be avoided to redo it again.
    //   Look at bb.size() to find out the expected dimensionality (e.g. 2 or 3
    //   for point clouds)
    template <class BBOX>
    bool kdtree_get_bbox(BBOX & /*bb*/) const
    {
        return false;
    }
};

}