#pragma once

#include <vector>
#include <Eigen/Dense>
#include <Eigen/Geometry>
#include <ceres/ceres.h>

#include "frame_cloud.h"
#include "icp_pair.h"
#include "jet.h"

// 小于0.5cm的CERES LOSS会被忽略
#define CERES_MIN_DISTANCE 0.005

namespace juefx {

class IcpProblem
{
public:
    void AddIcpPairSet(const IcpPairSet &pair_set, Eigen::Quaterniond &q, Eigen::Vector3d &t);
    void AddIcpPairSet(const IcpPairSet &pair_set, Eigen::Quaterniond delta_q, Eigen::Vector3d delta_t,
            Eigen::Quaterniond &q, Eigen::Vector3d &t);
    void AddParameterization(Eigen::Quaterniond &q);
    void Sovle();

    ceres::Problem *GetProblem() { return &problem_; }
    const ceres::Solver::Summary *GetSummary() { return &summary_; }

private:
    ceres::Problem problem_;
    ceres::Solver::Summary summary_;
    std::vector<IcpPairSet> pair_sets_;
};

class PointToPlaneError
{
public:
    EIGEN_MAKE_ALIGNED_OPERATOR_NEW

    const Eigen::Vector3d p_dst;
    const Eigen::Vector3d p_qry;
	const Eigen::Vector3d n_dst;
    Eigen::Quaterniond delta_q;
    Eigen::Vector3d delta_t;
    const double weight;
    const int8_t type;

    PointToPlaneError(const Eigen::Vector3d &dst,
                      const Eigen::Vector3d &qry,
                      const Eigen::Vector3d &nor,
                      const int8_t type,
                      Eigen::Quaterniond delta_q,
                      Eigen::Vector3d delta_t,
                      const double weight)
        : p_dst(dst), p_qry(qry), n_dst(nor), delta_q(delta_q), delta_t(delta_t), weight(weight), type(type) {}

    // Factory to hide the construction of the CostFunction object from the client
	// code.
    static ceres::CostFunction *Create(const Eigen::Vector3d &dst,
                                       const Eigen::Vector3d &qry,
                                       const Eigen::Vector3d &nor,
                                       const int8_t type,
                                       Eigen::Quaterniond delta_q,
                                       Eigen::Vector3d delta_t,
                                       const double weight = 1.0)
    {
        return (new ceres::AutoDiffCostFunction<PointToPlaneError, 1, 4, 3>(
            new PointToPlaneError(dst, qry, nor, type, delta_q, delta_t, weight)));
    }

    template <typename T>
	bool operator() (const T *const rotation, const T *const translation, T *residuals) const {
		// Make sure the Eigen::Vector world point is using the ceres::Jet type as
		// it's Scalar type
		Eigen::Matrix<T, 3, 1> qry;
		qry << T(p_qry[0]), T(p_qry[1]), T(p_qry[2]);
		Eigen::Matrix<T, 3, 1> dst;
		dst << T(p_dst[0]), T(p_dst[1]), T(p_dst[2]);
		Eigen::Matrix<T, 3, 1> nor;
		nor << T(n_dst[0]), T(n_dst[1]), T(n_dst[2]);
		
		Eigen::Matrix<T, 3, 3> move_r = delta_q.toRotationMatrix().cast<T>();
		Eigen::Matrix<T, 3, 1> move_t = delta_t.cast<T>();

        Eigen::Quaternion<T> q = Eigen::Map<const Eigen::Quaternion<T>>(rotation);
        Eigen::Matrix<T, 3, 1> t = Eigen::Map<const Eigen::Matrix<T, 3, 1>>(translation);
        Eigen::Matrix<T, 3, 1> qry_in_dst;

        if (type == PAIR_TYPE_TOP_TOP)
        {
            Eigen::Matrix<T, 3, 1> qry_in_dst_odom = move_r * (q.toRotationMatrix() * qry + t) + move_t;
            qry_in_dst = q.inverse().toRotationMatrix() * (qry_in_dst_odom - t);
        }
        else // if (type == PAIR_TYPE_TOP_BACK)
        {
            qry_in_dst = q.toRotationMatrix() * qry + t;
        }

        T distance = (qry_in_dst - dst).dot(nor);
        residuals[0] = distance > CERES_MIN_DISTANCE ? weight * distance : (T)0;
        return true;
    }
};

} // namespace juefx
