#include "icp_pair.h"

#include <thread>
#include <pcl/features/normal_3d.h>

#define NORMAL_SEARCH_MEAN_K 25
#define NORMAL_DEFINE_GROUND 0.98
#define NORMAL_SIMILAR 0.95
#define CURVATURE_FILTER_MAX 0.05
#define DISTANCE_PLAIN_PROJECTION 0.1

namespace juefx
{

PointCloudN::Ptr cloud_calculate_normal(pcl::PointCloud<pcl::PointXYZI>::ConstPtr cloud_in) {
    PointCloudN::Ptr cloud_out(new PointCloudN);
    pcl::copyPointCloud(*cloud_in, *cloud_out);

    pcl::NormalEstimation<pcl::PointXYZI, PointN> norm_est;
    pcl::search::KdTree<pcl::PointXYZI>::Ptr tree(
        new pcl::search::KdTree<pcl::PointXYZI>());
    norm_est.setSearchMethod(tree);
    norm_est.setKSearch(NORMAL_SEARCH_MEAN_K);
    norm_est.setInputCloud(cloud_in);
    norm_est.compute(*cloud_out);
    return cloud_out;
}

void get_closest_points_(const FrameCloud* p_source_frame, const FrameCloud* p_query_frame,
                         bool need_sample, int l, int r, double distance_thresh,
                         std::vector<IcpPair> &pairs)
{
    const FrameCloud &source_frame = *p_source_frame;
    const FrameCloud& query_frame = *p_query_frame;
    for (int i = l; i <= r; ++i)
    {
        size_t idx_min = 0;
        if (query_frame.pts[i].curvature > CURVATURE_FILTER_MAX)
            continue;
        double dist_squared = source_frame.getClosestPoint(query_frame.pts[i].P, idx_min);
        if (idx_min == 0 || idx_min >= source_frame.pts.size())
            continue;
        if (source_frame.pts[idx_min].curvature > CURVATURE_FILTER_MAX)
            continue;
        // filter by normal
        Eigen::Vector3d normal1 = query_frame.pts[i].N;
        Eigen::Vector3d normal2 = source_frame.pts[idx_min].N;
        double match = normal1.dot(normal2);
        if (fabs(match) < NORMAL_SIMILAR)
            continue;
        double point_dist = sqrt(dist_squared);
        // 长匹配不做抽稀, 短匹配抽稀
        if (!need_sample || (rand() & 0x01) == 0)
        {
            IcpPair pair;
            pair.frame_s = p_source_frame;
            pair.frame_q = p_query_frame;
            pair.index_s = idx_min;
            pair.index_q = i;
            pair.is_ground = fabs(normal1.z()) > NORMAL_DEFINE_GROUND &&
                             fabs(normal2.z()) > NORMAL_DEFINE_GROUND;
            pair.is_long = !need_sample;
            pair.type = (source_frame.lidar_id == query_frame.lidar_id)? PAIR_TYPE_TOP_TOP: PAIR_TYPE_TOP_BACK;
            // Drop too-far-from-normal pairs
            Eigen::Vector3d delta = query_frame.pts[i].P - source_frame.pts[idx_min].P;
            double normal_dis = fabs(delta.dot(normal2));
            if (normal_dis > distance_thresh)
                continue;
            double plain_dis = sqrt(delta.norm() * delta.norm() - normal_dis * normal_dis);
            if (plain_dis > DISTANCE_PLAIN_PROJECTION)
                continue;
            // sample by distance: 5m - keep 5%,  35m - keep 100%, adjust by is_ground
            pair.distance = (int8_t)query_frame.raw_pts[i].distance;
            double limit = (!pair.is_ground && pair.distance < 20) ? pair.distance / 2 + 5 : pair.distance * 3 - 10;
            int r = rand() % 100;
            if (r > limit)
                continue;

            pairs.push_back(pair);
        }
    }
}

void get_closest_points(const FrameCloud* source_frame, const FrameCloud* query_frame,
                        bool need_sample, double distance_thresh, int thread_num, IcpPairSet &pair_set) {

    std::vector<std::vector<IcpPair>> out_pairs;
    out_pairs.resize(thread_num);
    std::vector<std::thread> threads;
    int size = query_frame->pts.size();
    for (int i = 1; i <= thread_num; ++i)
    {
        int l = 0 + (i - 1) * size * 1.0 / thread_num;
        int r = 0 + i * size * 1.0 / thread_num - 1;
        threads.push_back(std::move(std::thread(&get_closest_points_, source_frame, query_frame, need_sample,
                                                l, r, distance_thresh, std::ref(out_pairs[i - 1]))));
    }
    for (int i = 0; i < threads.size(); ++i) {
        threads[i].join();
    }
    for (int i = 0; i < threads.size(); ++i) {
        pair_set.pairs.insert(pair_set.pairs.end(), out_pairs[i].begin(), out_pairs[i].end());
    }
#ifdef RAW_ICP_DEBUG
    printf("get %lu icp pairs after join thread ...\n", pair_set.pairs.size());
#endif
}

}