#include "raw_icp_matcher.h"

#include <glog/logging.h>

#include "raw_icp/icp_ceres.h"
#include "raw_icp/icp_pair.h"
#include "utils/rigid3.h"
#include "utils/timer.h"
#include "utils/util.hpp"

namespace juefx {

void RawICPMatcher::LoadPointCloudMap(const PointCloudJ::Ptr pcloud_map) {
    transfer_cloud_to_vector(pcloud_map, map_frame_.raw_pts);
    map_frame_.pts = map_frame_.raw_pts;
    map_frame_.buildKDTree();
    LOG(INFO) << "Load pcloud map finished, size: " << map_frame_.pts.size();
}

float RawICPMatcher::MatchPointCloud(const PointCloudJ::Ptr pcloud, const Eigen::Isometry3f init_pose,
                                  Eigen::Isometry3f &final_pose) {
    if (map_frame_.pts.size() == 0)
        return -1;

    TIMER_START(start);

    IcpPairSet pair_set;
    FrameCloud query_frame(0, 1);
    transfer_cloud_to_vector(pcloud, query_frame.raw_pts);
    const Eigen::Isometry3d trans = init_pose.cast<double>();
    for (auto p : query_frame.raw_pts) {
        VectorCloudPoint new_p;
        new_p.P = trans.linear() * p.P + trans.translation();
        new_p.N = trans.linear() * p.N;
        new_p.distance = p.distance;
        new_p.curvature = p.curvature;
        query_frame.pts.push_back(new_p);
    }
    get_closest_points(&map_frame_, &query_frame, false, 0.4, 4, pair_set);

    IcpProblem problem;
    Eigen::Quaterniond q(init_pose.rotation().cast<double>());
    Eigen::Vector3d t(init_pose.translation().cast<double>());
    problem.AddIcpPairSet(pair_set, q, t);
    problem.AddParameterization(q);
    problem.Sovle();
    LOG(INFO) << problem.GetSummary()->BriefReport();
    eigen3d_to_isometry3f(q, t, final_pose);

    char buf[1024];
    Eigen::Vector3d trans3 = final_pose.translation().cast<double>();
    Eigen::Vector3d euler3 = juefx::matrix_2_euler_rpy(final_pose.rotation().cast<double>());
    snprintf(buf, 1023, "{ pos: %.3f %.3f %.3f, euler: %.5f %.5f %.5f }",
             trans3.x(), trans3.y(), trans3.z(),
             euler3.x(), euler3.y(), euler3.z());
    TIMER_END(start, seconds);
    LOG(INFO) << "Match scan with map, transform: " << buf << ", cost " << seconds << " sec.";
    return 0;
}

float RawICPMatcher::MatchMultiPointCloud(std::vector<PointCloudJ::ConstPtr> target,
                                          std::vector<PointCloudJ::ConstPtr> source,
                                          const Eigen::Isometry3f init_pose,
                                          Eigen::Isometry3f &final_pose) {
    if (target.size() == 0 || target.size() != source.size())
        return -1;

    TIMER_START(start);

    std::vector<IcpPairSet> pair_set_vector;
    std::vector<FrameCloud> target_frame_vector;
    std::vector<FrameCloud> query_frame_vector;
    for (int i = 0; i < target.size(); i++) {
        // 必须一开始弄好, 否则vector自动增长后内部元素地址可能会变化
        pair_set_vector.emplace_back();
        target_frame_vector.emplace_back(i, 0);
        query_frame_vector.emplace_back(i, 1);
    }

    for (int i = 0; i < target.size(); i++) {
        FrameCloud& target_frame = target_frame_vector[i];
        transfer_cloud_to_vector(target[i], target_frame.raw_pts);
        target_frame.pts = target_frame.raw_pts;
        target_frame.buildKDTree();
        LOG(INFO) << "Load target cloud finished, size: " << target_frame.pts.size();

        FrameCloud &query_frame = query_frame_vector[i];
        transfer_cloud_to_vector(source[i], query_frame.raw_pts);
        const Eigen::Isometry3d trans = init_pose.cast<double>();
        for (auto p : query_frame.raw_pts)
        {
            VectorCloudPoint new_p;
            new_p.P = trans.linear() * p.P + trans.translation();
            new_p.N = trans.linear() * p.N;
            new_p.distance = p.distance;
            new_p.curvature = p.curvature;
            query_frame.pts.push_back(new_p);
        }
        LOG(INFO) << "Load source cloud finished, size: " << query_frame.pts.size();

        IcpPairSet &pair_set = pair_set_vector[i];
        get_closest_points(&target_frame, &query_frame, false, 0.4, 4, pair_set);
        LOG(INFO) << "Search for icp pairs, size: " << pair_set.pairs.size();
    }

    IcpProblem problem;
    Eigen::Quaterniond q(init_pose.rotation().cast<double>());
    Eigen::Vector3d t(init_pose.translation().cast<double>());
    for (int i = 0; i < target.size(); i++) {
        problem.AddIcpPairSet(pair_set_vector[i], q, t);
    }
    problem.AddParameterization(q);
    problem.Sovle();
    LOG(INFO) << problem.GetSummary()->BriefReport();
    eigen3d_to_isometry3f(q, t, final_pose);

    char buf[1024];
    Eigen::Vector3d trans3 = final_pose.translation().cast<double>();
    Eigen::Vector3d euler3 = juefx::matrix_2_euler_rpy(final_pose.rotation().cast<double>());
    snprintf(buf, 1023, "{ pos: %.4f %.4f %.4f, euler: %.6f %.6f %.6f }",
             trans3.x(), trans3.y(), trans3.z(),
             euler3.x(), euler3.y(), euler3.z());
    TIMER_END(start, seconds);
    LOG(INFO) << "Match scan with map, transform: " << buf << ", cost " << seconds << " sec.";
    return 0;
}
}