#include "vicp_matcher.h"

#include <glog/logging.h>

#include "fast_gicp/gicp/fast_vgicp.hpp"
#include "utils/timer.h"
#include "utils/util.hpp"
#include "utils/voxel_filter.h"

using namespace Eigen;
using namespace std;

namespace juefx {

VicpMatcher::VicpMatcher(const VicpMatcherOption& option)
    :option_(option)
{
    pcloudMap_.reset(new PointCloudJ);
}

void VicpMatcher::LoadPointCloudMap(
        const PointCloudJ::Ptr pcloud_map,
        float resolution)
{
    if (!pcloud_map)
        return;
    LOG(INFO) << "Load pcloud map ... resolution " << resolution;
    TIMER_START(start);
    clouds_voxel_filter(pcloud_map, pcloudMap_, resolution);
    TIMER_END(start, seconds);
    LOG(INFO) << "Load pcloud map finished, cost " << seconds << " sec.";
}

float VicpMatcher::MatchPointCloud(
        const PointCloudJ::Ptr pcloud,
        const Isometry3f init_pose,
        Isometry3f &final_pose)
{
    TIMER_START(start);

    // must be same with the resolution of map
    // because vgicp acts as ndt
    float resulotion = 0.05f;
    Matrix4f init_matrix;
    init_matrix.block(0, 0, 3, 3) = Matrix3f(init_pose.linear());
    init_matrix.block(0, 3, 3, 1) = init_pose.translation();

    PointCloudJ::Ptr scan = PointCloudJ::Ptr(new PointCloudJ);
    clouds_voxel_filter(pcloud, scan, resulotion);

    fast_gicp::FastVGICP<PointJ, PointJ> vgicp;
    vgicp.setResolution(option_.register_voxel_resolution);
    vgicp.setNumThreads(option_.threads_num);;
    vgicp.clearTarget();
    vgicp.clearSource();
    vgicp.setInputTarget(pcloudMap_);
    vgicp.setInputSource(scan);
    PointCloudJ::Ptr aligned = PointCloudJ::Ptr(new PointCloudJ);
    vgicp.align(*aligned, init_matrix);
    float fitness_score = vgicp.getFitnessScore();
    final_pose.matrix() = vgicp.getFinalTransformation();

    char buf[1024];
    Eigen::Vector3d trans3 = final_pose.translation().cast<double>();
    Eigen::Vector3d euler3 = juefx::matrix_2_euler_rpy(final_pose.rotation().cast<double>());
    snprintf(buf, 1023, "{ pos: %.3f %.3f %.3f, euler: %.5f %.5f %.5f }",
             trans3.x(), trans3.y(), trans3.z(),
             euler3.x(), euler3.y(), euler3.z());
    TIMER_END(start, seconds);
    LOG(INFO) << "Match scan with map, voxel_filter: " << resulotion
              << ", transform: " << buf << ", cost " << seconds << " sec.";
    return fitness_score;
}

}