#include "blockarea.h"

namespace localize {

BlockArea::BlockArea(
        const string &meshPath,
        const boost::shared_ptr<Distributions> &distributions,
        uint32_t blockInfoFileOffset,
        uint32_t blockFileSize,
        uint32_t gridCntPerBlockSide)
    :meshPath_(meshPath),
      distributions_(distributions),
      blockInfoFileOffset_(blockInfoFileOffset),
      blockFileSize_(blockFileSize),
      gridCntPerBlockSide_(gridCntPerBlockSide)
{

}

void BlockArea::backendFunc(uint8_t type, Vector2d point)
{

}

void BlockArea::load()
{

    ifstream infile(meshPath_, ios::binary);
    if(!infile)
    {
        LOG(WARNING) << meshPath_ << " open error!";
        return ;
    }
    char* blockInfoData = (char*)malloc(blockFileSize_);
    infile.clear();
    infile.seekg(blockInfoFileOffset_);
    infile.read(blockInfoData, blockFileSize_);
    BlockInfo blockInfo;
    blockInfo.ParseFromArray(blockInfoData, blockFileSize_);
    free(blockInfoData);
    infile.close();
    plainParams_.push_back(blockInfo.a());
    plainParams_.push_back(blockInfo.b());
    plainParams_.push_back(blockInfo.c());
    plainParams_.push_back(blockInfo.d());
    plainParamValid_ = fabs(plainParams_[0]) > 1e-3 ||
            fabs(plainParams_[1]) > 1e-3 ||
            fabs(plainParams_[2]) > 1e-3 ||
            fabs(plainParams_[3]) > 1e-3;

    if(blockInfo.bit_map_size()){
        parseBitmapBlock(blockInfo);
    }else{
        parseIndexBlock(blockInfo);

    }
    configMultiLayer();
    unordered_map<int64_t, juefx::DistributionIndex>().swap(hashMap_);

    loaded_ = true;
}

//--------------------------------------------------------------------------


//--------------------------------------------------------------------------

void BlockArea::parseBitmapBlock(
        const BlockInfo &blockInfo)
{
    vector<uint32_t> activeGridIndexes =
            parseBitMap(blockInfo.bit_map(), blockInfo.bit_map_size());
    vector<pair<uint32_t, uint32_t>> shiftedAndFeatureVec;
    distributionVec_.reserve(activeGridIndexes.size() * 15);
    for(size_t i = 0; i < activeGridIndexes.size(); i++){
        uint32_t shiftedGridIndex = activeGridIndexes[i];
        uint32_t gridFeatureIndex = blockInfo.grid_feature_index(i);
        shiftedAndFeatureVec.push_back(make_pair(gridFeatureIndex, shiftedGridIndex));
    }
    sort(shiftedAndFeatureVec.begin(), shiftedAndFeatureVec.end(),
         [](const pair<uint32_t, uint32_t>& query,
         const pair<uint32_t, uint32_t>& target) {
             return query.first > target.first;
         });
    parseFromShiftedAndFeatureVec(shiftedAndFeatureVec);
}
//--------------------------------------------------------------------------

void BlockArea::parseIndexBlock(
        const BlockInfo &blockInfo)
{
    uint32_t gridFeatureIndexSize = blockInfo.grid_feature_index_size();
    distributionVec_.reserve(gridFeatureIndexSize * 15);
    vector<pair<uint32_t, uint32_t>> shiftedAndFeatureVec;
    for(size_t i = 0; i < gridFeatureIndexSize; i++){
        uint32_t gridFeatureIndex = blockInfo.grid_feature_index(i);
        if(gridFeatureIndex == 0){
            continue;
        }
        uint32_t shiftedGridIndex = i;
        shiftedAndFeatureVec.push_back(make_pair(gridFeatureIndex, shiftedGridIndex));
    }
    sort(shiftedAndFeatureVec.begin(), shiftedAndFeatureVec.end(),
         [](const pair<uint32_t, uint32_t>& query,
         const pair<uint32_t, uint32_t>& target) {
             return query.first > target.first;
         });
    parseFromShiftedAndFeatureVec(shiftedAndFeatureVec);
}
//--------------------------------------------------------------------------

void BlockArea::parseFromShiftedAndFeatureVec(
        const vector<pair<uint32_t, uint32_t> > &shiftedAndFeatureVec)
{
    indexTable_[0].resize(gridCntPerBlockSide_);
    for(auto& col : indexTable_[0]){
        col.resize(gridCntPerBlockSide_);
    }

//    static int id = 0;
//    std::ofstream ofs("/home/juefx/grid/grid_" + to_string(id++) + ".txt");

    for(size_t i = 0; i < shiftedAndFeatureVec.size(); i++){
        uint32_t gridFeatureIndex = shiftedAndFeatureVec[i].first;
        uint32_t shiftedGridIndex = shiftedAndFeatureVec[i].second;
        Vector2i index;
        index.x() = shiftedGridIndex / gridCntPerBlockSide_;
        index.y() = shiftedGridIndex % gridCntPerBlockSide_;

        const GridDistribution &gridDistribution =
                distributions_->grid_distribution(gridFeatureIndex);
        size_t oneDistributionCnt = gridDistribution.one_distribution_size();

        juefx::GridData gridData;
        for(size_t j = 0; j < oneDistributionCnt; j++){
            const OneDistribution &oneDistribProto =
                    gridDistribution.one_distribution(j);
            juefx::Distribution juefxDistrib;
            juefxDistrib.floor = oneDistribProto.floor() * 3;
            juefxDistrib.height = oneDistribProto.z_delta_leval() * 3;
            juefxDistrib.intensity = oneDistribProto.intensity() * 8;
            gridData.emplace_back(juefxDistrib);
        }
        gatherGridData(gridData, index);
    }
}
//--------------------------------------------------------------------------

void BlockArea::gatherGridData(
        const juefx::GridData &gridData,
        const Vector2i &index,
        uint8_t layerId)
{
    int64_t hash = 0;
    for(size_t i = 0; i < gridData.size(); i++){
        const juefx::Distribution &juefxDistrib =
                gridData[i];
        hash += juefxDistrib.getScore() * (i * 131313131 + 1);
    }

    auto iter = hashMap_.find(hash);
    if(iter != hashMap_.end()){
        indexTable_[layerId][index.x()][index.y()] = iter->second;
    }else{
        juefx::DistributionIndex distributionIndexIndex;
        distributionIndexIndex.start = distributionVec_.size();
        distributionIndexIndex.length = gridData.size();
        hashMap_.insert(make_pair(hash, distributionIndexIndex));
        indexTable_[layerId][index.x()][index.y()] = move(distributionIndexIndex);
        if(gridData.size() == 1){
            distributionVec_.emplace_back(gridData[0]);

        }else{
            for(size_t k = 0; k < gridData.size(); k++){
                const juefx::Distribution &juefxDistrib =
                        gridData[k];
                int64_t singleHash = juefxDistrib.getScore();
                auto it = hashMap_.find(singleHash);
                if(it == hashMap_.end()){
                    juefx::DistributionIndex singleIndex;
                    singleIndex.start = distributionVec_.size();
                    singleIndex.length = 1;
                    hashMap_.insert(make_pair(singleHash, singleIndex));
                    distributionVec_.emplace_back(juefxDistrib);
                }
            }
        }
    }
}
//--------------------------------------------------------------------------

vector<uint32_t> BlockArea::parseBitMap(
        const BlockArea::ProtoRepeatedUint32 &bitmap,
        const uint32_t size)
{
    vector<uint32_t> activeGridIndexes;
    uint32_t activeGridIndex = 0;
    for(size_t i = 0; i < size; i++){
        uint8_t byte = bitmap.Get(i);
        for(size_t byteCnt = 0; byteCnt < 8; byteCnt++){
            if((byte >> byteCnt) & 1){
                activeGridIndexes.emplace_back(activeGridIndex);
            }
            activeGridIndex++;
        }
    }
//    LOG(INFO) << "activeGridIndexes.size(): " << activeGridIndexes.size();
    return activeGridIndexes;
}
//--------------------------------------------------------------------------
inline juefx::GridData getGridDataFromGridDistrib(
        const juefx::Distribution* gridDistribution, const uint8_t &length){
    juefx::GridData ret;
    if(gridDistribution == nullptr){
        return ret;
    }
    for(size_t i = 0; i < length; i++){
        ret.emplace_back(*(gridDistribution + i));
    }
    return ret;
}

//--------------------------------------------------------------------------

void BlockArea::configMultiLayer(uint8_t layerCnt)
{
    for(uint8_t layerId = 1; layerId < layerCnt; layerId++){
        uint8_t preLayerId = layerId - 1;
        uint8_t skipCnt = 1 << preLayerId;
        indexTable_[layerId].resize(gridCntPerBlockSide_);
        for(auto& col : indexTable_[layerId]){
            col.resize(gridCntPerBlockSide_);
        }
        size_t preGridIndexVec[8];
        for(size_t row = 0; row < gridCntPerBlockSide_; row++){
            for(size_t col = 0; col < gridCntPerBlockSide_; col++){
                juefx::GridData preGridData;
                preGridData.reserve(6);

                preGridIndexVec[0] = row;
                preGridIndexVec[1] = col;
                preGridIndexVec[2] = row + skipCnt;
                preGridIndexVec[3] = col;
                preGridIndexVec[4] = row;
                preGridIndexVec[5] = col + skipCnt;
                preGridIndexVec[6] = row + skipCnt;
                preGridIndexVec[7] = col + skipCnt;
                for(int i = 0; i < 4; i++){
                    size_t gridIndexX = preGridIndexVec[i << 1];
                    size_t gridIndexY = preGridIndexVec[(i << 1) + 1];

                    if(gridIndexX >= gridCntPerBlockSide_ ||
                            gridIndexY >= gridCntPerBlockSide_){
                        continue;
                    }
                    uint8_t length;
                    juefx::Distribution* distribution =
                            getGridDistribution(preLayerId, gridIndexX, gridIndexY, length);
                    if(length == 0){
                        continue;
                    }
                    const juefx::GridData &currPreGridData =
                            getGridDataFromGridDistrib(distribution, length);
                    for(const juefx::Distribution &distrib : currPreGridData){
                        preGridData.emplace_back(distrib);
                    }
                }

                juefx::GridData mergedGridData;
                if(!mergeGridData(preGridData, mergedGridData)){
                    continue;
                }

                gatherGridData(mergedGridData, {row, col}, layerId);

            }
        }

    }
}


bool BlockArea::mergeGridData(
        juefx::GridData &preGridData,
        juefx::GridData &mergedGridData){
    if(preGridData.size() == 0){
        return false;
    }
    sort(preGridData.begin(), preGridData.end(),
         [](const juefx::Distribution& query,
         const juefx::Distribution& target) {
             return query.floor < target.floor;
         });
    vector<uint16_t> mergedCntVec;
    for(const juefx::Distribution &newDstbt : preGridData){
        if(mergedGridData.size() == 0){
            mergedGridData.emplace_back(newDstbt);
            mergedCntVec.push_back(1);
            continue;
        }
        int32_t newDstbtCeil = newDstbt.floor + newDstbt.height;
        bool merged = false;
        for(size_t i = 0; i < mergedGridData.size(); i++){
            juefx::Distribution& dstbt = mergedGridData[i];
            int32_t dstbtCeil = dstbt.floor + dstbt.height;
            if(!(newDstbtCeil < (dstbt.floor - 10) ||
                    newDstbt.floor > (dstbtCeil + 10))){    //10cm threhold!!!
                dstbt.floor = min(newDstbt.floor, dstbt.floor);
                int32_t ceil = max(newDstbt.floor + newDstbt.height,
                                   dstbt.floor + dstbt.height);
                dstbt.height = ceil - dstbt.floor;
                dstbt.intensity = (newDstbt.intensity * mergedCntVec[i] + dstbt.intensity) / (i + 1);
                mergedCntVec[i]++;
                merged = true;
            }
        }
        if(!merged){
            mergedGridData.emplace_back(newDstbt);
            mergedCntVec.push_back(1);
        }
    }
    return true;
}


}
