#include "blockarea.h"

namespace localize {

BlockArea::BlockArea(
        const string &meshPath,
        const boost::shared_ptr<Distributions> &distributions,
        uint32_t blockInfoFileOffset,
        uint32_t blockFileSize,
        uint32_t gridCntPerBlockSide)
    :meshPath_(meshPath),
      distributions_(distributions),
      blockInfoFileOffset_(blockInfoFileOffset),
      blockFileSize_(blockFileSize),
      gridCntPerBlockSide_(gridCntPerBlockSide)
{

}

void BlockArea::backendFunc(uint8_t type, Vector2d point)
{

}

void BlockArea::load()
{
    ifstream infile(meshPath_, ios::binary);
    if(!infile)
    {
        LOG(WARNING) << meshPath_ << " open error!";
        return ;
    }
    char* blockInfoData = (char*)malloc(blockFileSize_);
    infile.clear();
    infile.seekg(blockInfoFileOffset_);
    infile.read(blockInfoData, blockFileSize_);
    BlockInfo blockInfo;
    blockInfo.ParseFromArray(blockInfoData, blockFileSize_);
    free(blockInfoData);
    infile.close();
    plainParams_.emplace_back(blockInfo.a());
    plainParams_.emplace_back(blockInfo.b());
    plainParams_.emplace_back(blockInfo.c());
    plainParams_.emplace_back(blockInfo.d());
    plainParamValid_ = fabs(plainParams_[0]) > 1e-3 ||
            fabs(plainParams_[1]) > 1e-3 ||
            fabs(plainParams_[2]) > 1e-3 ||
            fabs(plainParams_[3]) > 1e-3;

    if(blockInfo.bit_map_size()){
        parseBitmapBlock(blockInfo);
    }else{
        parseIndexBlock(blockInfo);

    }
    configMultiLayer();
    unordered_map<int64_t, juefx::DistributionIndex>().swap(hashMap_);

    loaded_ = true;
}

//--------------------------------------------------------------------------

juefx::Distribution *BlockArea::getGridDistribution(
        uint8_t layerId, uint8_t shiftedX, uint8_t shiftedY, uint8_t &length) const
{
    juefx::Distribution * ret;
    const juefx::DistributionIndex &index = indexTable_[layerId][shiftedX][shiftedY];
    if(index.length == 0){
        return nullptr;
    }
    length = index.length;
    ret = const_cast<juefx::Distribution*>(
                &(distributionVec_.data()[index.start]));
    return ret;
}
//--------------------------------------------------------------------------

void BlockArea::parseBitmapBlock(
        const BlockInfo &blockInfo)
{
    vector<uint32_t> activeGridIndexes =
            parseBitMap(blockInfo.bit_map(), blockInfo.bit_map_size());
    vector<pair<uint32_t, uint32_t>> shiftedAndFeatureVec;
    distributionVec_.reserve(activeGridIndexes.size() * 60);
    hashMap_.reserve(activeGridIndexes.size() * 15);
    shiftedAndFeatureVec.reserve(activeGridIndexes.size());
    for(size_t i = 0; i < activeGridIndexes.size(); i++){
        uint32_t shiftedGridIndex = activeGridIndexes[i];
        uint32_t gridFeatureIndex = blockInfo.grid_feature_index(i);
        shiftedAndFeatureVec.emplace_back(make_pair(gridFeatureIndex, shiftedGridIndex));
    }
    sort(shiftedAndFeatureVec.begin(), shiftedAndFeatureVec.end(),
         [](const pair<uint32_t, uint32_t>& query,
         const pair<uint32_t, uint32_t>& target) {
             return query.first > target.first;
         });
    parseFromShiftedAndFeatureVec(shiftedAndFeatureVec);
}
//--------------------------------------------------------------------------

void BlockArea::parseIndexBlock(
        const BlockInfo &blockInfo)
{
    uint32_t gridFeatureIndexSize = blockInfo.grid_feature_index_size();
    distributionVec_.reserve(gridFeatureIndexSize * 60);
    hashMap_.reserve(gridFeatureIndexSize * 15);
    vector<pair<uint32_t, uint32_t>> shiftedAndFeatureVec;
    shiftedAndFeatureVec.reserve(gridFeatureIndexSize);
    for(size_t i = 0; i < gridFeatureIndexSize; i++){
        uint32_t gridFeatureIndex = blockInfo.grid_feature_index(i);
        if(gridFeatureIndex == 0){
            continue;
        }
        uint32_t shiftedGridIndex = i;
        shiftedAndFeatureVec.emplace_back(make_pair(gridFeatureIndex, shiftedGridIndex));
    }
    sort(shiftedAndFeatureVec.begin(), shiftedAndFeatureVec.end(),
         [](const pair<uint32_t, uint32_t>& query,
         const pair<uint32_t, uint32_t>& target) {
             return query.first > target.first;
         });
    parseFromShiftedAndFeatureVec(shiftedAndFeatureVec);
}
//--------------------------------------------------------------------------

void BlockArea::parseFromShiftedAndFeatureVec(
        const vector<pair<uint32_t, uint32_t> > &shiftedAndFeatureVec)
{
    indexTable_[0].resize(gridCntPerBlockSide_);
    for(auto& col : indexTable_[0]){
        col.resize(gridCntPerBlockSide_);
    }

    for(size_t i = 0; i < shiftedAndFeatureVec.size(); i++){
        uint32_t gridFeatureIndex = shiftedAndFeatureVec[i].first;
        uint32_t shiftedGridIndex = shiftedAndFeatureVec[i].second;
        Vector2i index;
        index.x() = shiftedGridIndex / gridCntPerBlockSide_;
        index.y() = shiftedGridIndex % gridCntPerBlockSide_;

        const GridDistribution &gridDistribution =
                distributions_->grid_distribution(gridFeatureIndex);
        size_t oneDistributionCnt = gridDistribution.one_distribution_size();

        juefx::GridData gridData;
        for(size_t j = 0; j < oneDistributionCnt; j++){
            const OneDistribution &oneDistribProto =
                    gridDistribution.one_distribution(j);
            juefx::Distribution juefxDistrib;
            juefxDistrib.floor = oneDistribProto.floor() * 3;
            juefxDistrib.height = oneDistribProto.z_delta_leval() * 3;
            juefxDistrib.intensity = oneDistribProto.intensity() * 8;
            gridData.emplace_back(juefxDistrib);
        }
        gatherGridData(gridData, index);
    }


//    std::ofstream ofs("/home/juefx/grid/" + to_string((int)rowItemCnt_ - 320) + "_" +
//                      to_string((int)colItemCnt_ - 320) + ".txt");
//    for(size_t row = 0; row < gridCntPerBlockSide_; row++){
//        for(size_t col = 0; col < gridCntPerBlockSide_; col++){
//            uint8_t length;
//            auto gridDistribution = getGridDistribution(0, row, col, length);
//            Vector2f gridCenter;
//            gridCenter.x() = (rowItemCnt_ - 320) * 23.04 + ((float)row + 0.5) * 0.09;
//            gridCenter.y() = (colItemCnt_ - 320) * 23.04 + ((float)col + 0.5) * 0.09;
//            for(int i = 0; i < length; i++){
//                for(int32_t j = 0;
//                    j < (gridDistribution + i)->height / 3;
//                    j++){
//                    int32_t indexZ = j + (gridDistribution + i)->floor / 3;
//                    float point[4];
//                    point[0] = gridCenter.x();
//                    point[1] = gridCenter.y();
//                    point[2] = ((float)indexZ + 0.5) * 3 / 100;
//                    point[3] = (gridDistribution + i)->intensity;
//                    ofs << point[0] << "," << point[1] << ","
//                                    << point[2] << "," << point[3] << std::endl;
//                }
//            }
//        }
//    }


//    exit(0);
}
//--------------------------------------------------------------------------

void BlockArea::gatherGridData(
        const juefx::GridData &gridData,
        const Vector2i &index,
        uint8_t layerId)
{
    int64_t hash = 0;
    for(size_t i = 0; i < gridData.size(); i++){
        const juefx::Distribution &juefxDistrib =
                gridData[i];
        hash += juefxDistrib.getScore() * (i * 131313131 + 1);
    }

    auto iter = hashMap_.find(hash);
    if(iter != hashMap_.end()){
        indexTable_[layerId][index.x()][index.y()] = iter->second;
    }else{
        juefx::DistributionIndex distributionIndexIndex;
        distributionIndexIndex.start = distributionVec_.size();
        distributionIndexIndex.length = gridData.size();
        hashMap_.insert(make_pair(hash, distributionIndexIndex));
        indexTable_[layerId][index.x()][index.y()] = distributionIndexIndex;
        for(size_t k = 0; k < gridData.size(); k++){
            distributionVec_.emplace_back(gridData[k]);
        }
    }
}
//--------------------------------------------------------------------------

vector<uint32_t> BlockArea::parseBitMap(
        const BlockArea::ProtoRepeatedUint32 &bitmap,
        const uint32_t size)
{
    vector<uint32_t> activeGridIndexes;
    uint32_t activeGridIndex = 0;
    for(size_t i = 0; i < size; i++){
        uint8_t byte = bitmap.Get(i);
        for(size_t byteCnt = 0; byteCnt < 8; byteCnt++){
            if((byte >> byteCnt) & 1){
                activeGridIndexes.emplace_back(activeGridIndex);
            }
            activeGridIndex++;
        }
    }
//    LOG(INFO) << "activeGridIndexes.size(): " << activeGridIndexes.size();
    return activeGridIndexes;
}
//--------------------------------------------------------------------------
inline void getGridDataFromGridDistrib(
        const juefx::Distribution* gridDistribution, juefx::GridData &gridData, uint8_t length){
    if(gridDistribution == nullptr){
        return;
    }
    for(size_t i = 0; i < length; i++){
        gridData.emplace_back(*(gridDistribution + i));
    }
}

//--------------------------------------------------------------------------

void BlockArea::configMultiLayer(uint8_t layerCnt)
{
    Vector2i pos;
    for(uint8_t layerId = 1; layerId < layerCnt; layerId++){
        uint8_t preLayerId = layerId - 1;
        uint8_t skipCnt = 1 << preLayerId;
        indexTable_[layerId].resize(gridCntPerBlockSide_);
        for(auto& col : indexTable_[layerId]){
            col.resize(gridCntPerBlockSide_);
        }
        size_t preGridIndexVec[8];
        for(size_t row = 0; row < gridCntPerBlockSide_; row++){
            for(size_t col = 0; col < gridCntPerBlockSide_; col++){
                juefx::GridData preGridData;
                preGridData.reserve(6);

                preGridIndexVec[0] = row;
                preGridIndexVec[1] = col;
                preGridIndexVec[2] = row + skipCnt;
                preGridIndexVec[3] = col;
                preGridIndexVec[4] = row;
                preGridIndexVec[5] = col + skipCnt;
                preGridIndexVec[6] = row + skipCnt;
                preGridIndexVec[7] = col + skipCnt;
                for(int i = 0; i < 4; i++){
                    size_t gridIndexX = preGridIndexVec[i << 1];
                    size_t gridIndexY = preGridIndexVec[(i << 1) + 1];

                    if(gridIndexX >= gridCntPerBlockSide_ ||
                            gridIndexY >= gridCntPerBlockSide_){
                        continue;
                    }
                    uint8_t length;
                    juefx::Distribution* distribution =
                            getGridDistribution(preLayerId, gridIndexX, gridIndexY, length);
                    if(length == 0){
                        continue;
                    }
                    getGridDataFromGridDistrib(distribution, preGridData, length);
                }

                if(preGridData.size() == 0){
                    continue;
                }
                sort(preGridData.begin(), preGridData.end(),
                     [](const juefx::Distribution& query,
                     const juefx::Distribution& target) {
                         return query.floor < target.floor;
                     });
                juefx::GridData mergedGridData;
                mergeGridData(preGridData, mergedGridData);

                pos.x() = row;
                pos.y() = col;
                gatherGridData(mergedGridData, pos, layerId);

//                if(layerId != 1){
//                    continue;
//                }
//                std::ofstream ofs;
//                ofs.open("/home/juefx/map_" + to_string(layerId) + ".txt", std::ios::app);
//                int32_t blockIndexX = rowItemCnt_;
//                int32_t blockIndexY = colItemCnt_;
//                uint32_t gridIndexX = row;
//                uint32_t gridIndexY = col;
//                Vector2f gridCenter;
//                gridCenter.x() = blockIndexX * 23.04 + (0.5 + gridIndexX) * 0.09;
//                gridCenter.y() = blockIndexY * 23.04 + (0.5 + gridIndexY) * 0.09;

//                for(int i = 0; i < mergedGridData.size(); i++){
//                    for(int32_t j = 0;
//                        j < mergedGridData[i].height / 3;
//                        j++){
//                        int32_t indexZ = j + mergedGridData[i].floor / 3;
//                        float point[4];
//                        point[0] = gridCenter.x();
//                        point[1] = gridCenter.y();
//                        point[2] = ((float)indexZ + 0.5) * 3 / 100;
//                        point[3] = mergedGridData[i].intensity;
//                        ofs << point[0] << "," << point[1] << ","
//                                        << point[2] << "," << point[3] << std::endl;
//                    }
//                }
//                ofs.close();
            }
        }
    }
}

void BlockArea::mergeGridData(
        const juefx::GridData &preGridData,
        juefx::GridData &mergedGridData){
    vector<uint32_t> mergedIntensityVec;
    vector<uint32_t> mergedHeightVec;
    mergedGridData.reserve(3);
    mergedIntensityVec.reserve(3);
    mergedHeightVec.reserve(3);
    for(const juefx::Distribution &newDstbt : preGridData){
        auto shiftedIntensity = newDstbt.intensity * newDstbt.height;
        if(mergedGridData.size() == 0){
            mergedGridData.emplace_back(newDstbt);
            mergedIntensityVec.emplace_back(shiftedIntensity);
            mergedHeightVec.emplace_back(newDstbt.height);
            continue;
        }
        int32_t newDstbtCeil = newDstbt.floor + newDstbt.height;

        juefx::Distribution& dstbt = mergedGridData.back();
        int32_t dstbtCeil = dstbt.floor + dstbt.height;
        if(newDstbtCeil <= dstbtCeil){
            mergedIntensityVec.back() += shiftedIntensity;
            mergedHeightVec.back() += newDstbt.height;
            continue;
        }

        if(newDstbt.floor < (dstbtCeil + 10)){
            dstbt.height = newDstbtCeil - dstbt.floor;
            mergedIntensityVec.back() += shiftedIntensity;
            mergedHeightVec.back() += newDstbt.height;
        }else{
            mergedGridData.emplace_back(newDstbt);
            mergedIntensityVec.emplace_back(shiftedIntensity);
            mergedHeightVec.emplace_back(newDstbt.height);
        }

    }
    for(size_t i = 0; i < mergedGridData.size(); i++){
        juefx::Distribution& dstbt = mergedGridData[i];
        dstbt.intensity = mergedIntensityVec[i] / mergedHeightVec[i];
    }
}

}
