#include "mesharea.h"
#include <google/protobuf/io/coded_stream.h>
#include <google/protobuf/io/zero_copy_stream_impl_lite.h>

namespace localize {

MeshArea::MeshArea(const string &meshPath)
    :meshPath_(meshPath)
{
    Vector2i window_size = {3, 3};
    for(int x = -window_size.x(); x <= window_size.x(); x++)
        for(int y = -window_size.y(); y <= window_size.y(); y++){
            if(fabs(x * y) < 6){
                checkNeighbors_.push_back({x, y});
            }
        }
    window_size = {3, 3};
    for(int x = -window_size.x(); x <= window_size.x(); x++)
        for(int y = -window_size.y(); y <= window_size.y(); y++){
            if(fabs(x * y) < 6){
                firstLoadNeighbors_.push_back({x, y});
            }
        }
    window_size = {4, 4};
    for(int x = -window_size.x(); x <= window_size.x(); x++)
        for(int y = -window_size.y(); y <= window_size.y(); y++){
            if(fabs(x * y) < 12){
                loadNeighbors_.push_back({x, y});
            }
        }
}

void MeshArea::init()
{
    ifstream infile(meshPath_, ios::binary);
    if(!infile)
    {
        LOG(WARNING) << meshPath_ << " open error!";
        return;
    }
    //    ---------------- HeaderLength --------------
    char buf[2];
    infile.read(buf, 2);
    uint16_t headerLength = ((uint16_t)buf[1] << 8) + (uint16_t)buf[0];

    //    ---------------- Header --------------
    char* headerData = (char*)malloc(headerLength);
    infile.read(headerData, headerLength);
    Header header;
    header.ParseFromArray(headerData, headerLength);
    free(headerData);
    try {
        meshId_ = stoi(header.mesh_id());
    }
    catch (std::invalid_argument&){
        LOG(INFO) << "header.mesh_id() is not int";
    }
    gridResolution_ = lround(header.high_relolution() * 100);
    blockResolution_ = lround(header.block_relolution() * 100);
    gridCntPerBlockSide_ = blockResolution_ / gridResolution_;
    setResolution(blockResolution_);
    setSideItemCnt(BlockCntPerMeshSide);
    offset_.x() = header.offset_x();
    offset_.y() = header.offset_y();
    offset_.z() = header.offset_z();
    meshLevel_ = header.mesh_level();
    meshCenter_.x() = header.latitude();
    meshCenter_.y() = header.longtitude();
    Vector3d realCenter;
    LocalCartesian proj(meshCenter_.x(), meshCenter_.y(), 0);
    proj.Reverse(offset_.x(), offset_.y(), 0,
                 realCenter.x(), realCenter.y(), realCenter.z());
    proj_.reset(new LocalCartesian(realCenter.x(), realCenter.y(), 0));
//    LOG(INFO) << setprecision(15) << "realCenter: " << realCenter.transpose();


    uint16_t blockIndexLength = header.block_index_length();
    uint32_t blockInfoLength = header.block_info_length();
    uint32_t blockIndexOffset = 2 + headerLength;
    uint32_t blockInfoOffset = blockIndexOffset + blockIndexLength;

    //    ---------------- Distributions --------------
    infile.clear();
    infile.seekg(0, infile.end);
    uint32_t totalSize = infile.tellg();
    infile.clear();
    uint32_t distributionOffset = blockInfoOffset + blockInfoLength;
    infile.seekg(distributionOffset);
    uint32_t distributionSize = totalSize - distributionOffset;
    char* distributionBuf = (char*)malloc(distributionSize);
    infile.read(distributionBuf, distributionSize);
    distributions_.reset(new Distributions);
    ::google::protobuf::io::ArrayInputStream input(distributionBuf, distributionSize);
    ::google::protobuf::io::CodedInputStream decoder(&input);
    decoder.SetTotalBytesLimit(1024 * 1024 * 1024, 64 * 1024 * 1024);
    distributions_->ParseFromCodedStream(&decoder);
    free(distributionBuf);

    //    ---------------- BlockIndex --------------
    infile.clear();
    infile.seekg(blockIndexOffset);
    char* blockIndexData = (char*)malloc(blockIndexLength);
    infile.read(blockIndexData, blockIndexLength);
    BlockIndex blockIndex;
    blockIndex.ParseFromArray(blockIndexData, blockIndexLength);
    free(blockIndexData);
    blockAreaTable_.resize(rowItemCnt_);
    for(auto& col : blockAreaTable_){
        col.resize(colItemCnt_);
    }
    shiftBlockIdOffset_ = (sideItemCnt_ / 2) * (sideItemCnt_ + 1);
    size_t block_describe_size = blockIndex.block_describe_size();
    for(size_t i = 0; i < block_describe_size; i++){
        BlockDescribe blockDescribe = blockIndex.block_describe(i);
        int32_t blockId = blockDescribe.block_id();
//        if(blockId == -315){
//            int a = 0;
//            a++;
//        }
        uint32_t shiftBlockId = blockId + shiftBlockIdOffset_;
        int32_t shiftBlockIndexX = floor((float)shiftBlockId / sideItemCnt_);
        int32_t shiftBlockIndexY = shiftBlockId % sideItemCnt_;
        int32_t blockIndexX = shiftBlockIndexX - shiftSideItemOffset_;
        int32_t blockIndexY = shiftBlockIndexY - shiftSideItemOffset_;
        float blockCenterX = (float(blockIndexX) + 0.5) * blockResolution_ / 100;
        float blockCenterY = (float(blockIndexY) + 0.5) * blockResolution_ / 100;
//        if(blockIndexX == -5 && blockIndexY == 5){
//            int a = 0;
//            a++;
//        }
        BlockContent &blockContent = blockAreaTable_[shiftBlockIndexX][shiftBlockIndexY];
        uint32_t blockInfoFileOffset = blockDescribe.block_info_offset();
        uint32_t blockFileSize = blockDescribe.block_info_size();
        blockContent.block.reset(
                    new BlockArea(meshPath_, distributions_,
                                  blockInfoOffset + blockInfoFileOffset,
                                  blockFileSize, gridCntPerBlockSide_));
        blockContent.block->setItemCnt(shiftBlockIndexX, shiftBlockIndexY);
        blockContent.blockState = CurrMesh;
//        proj_->Reverse(blockCenterX, blockCenterY, 0,
//                       blockContent.blockCenterBLH.B,
//                       blockContent.blockCenterBLH.L,
//                       blockContent.blockCenterBLH.H);
    }
    infile.close();


//    LOG(INFO) << meshId_ <<" init Done";
    inited_ = true;
}

void MeshArea::backendFunc(uint8_t type, Vector2d point)
{
    if(type == LoadAread){
        loadArea(point);
    }
}

void MeshArea::loadArea(const Vector2d &bl)
{
    if(!inited_){
        init();
    }
    if(!blockStateChecked_){
        checkBlockState();
    }
    loadBlockAtAreaByBlh(bl);
}

void MeshArea::load()
{
    if(!inited_){
        init();
    }
    for(size_t x = 0; x < rowItemCnt_; x++){
        for(size_t y = 0; y < colItemCnt_; y++){
            const BlockContent &blockContent = blockAreaTable_[x][y];
            if(blockContent.blockState != CurrMesh){
                continue;
            }
            blockContent.block->load();
        }
    }
}

void MeshArea::checkBlockState()
{
    reSetupNeighBorMesh();
    for(size_t x = 0; x < rowItemCnt_; x++){
        for(size_t y = 0; y < colItemCnt_; y++){
            BlockContent &blockContent = blockAreaTable_[x][y];
            if(blockContent.blockState == CurrMesh){
                continue;
            }
            int32_t blockIndexX = x - BlockCntPerMeshSide / 2;
            int32_t blockIndexY = y - BlockCntPerMeshSide / 2;
            float blockCenterX = (float(blockIndexX) + 0.5) * blockResolution_ / 100;
            float blockCenterY = (float(blockIndexY) + 0.5) * blockResolution_ / 100;
            proj_->Reverse(blockCenterX, blockCenterY, 0,
                           blockContent.blockCenterBLH.B,
                           blockContent.blockCenterBLH.L,
                           blockContent.blockCenterBLH.H);
            ulong meshIdContainBlock = GetMeshFromPosition(
                        blockContent.blockCenterBLH.L, blockContent.blockCenterBLH.B);
            if(meshIdContainBlock == meshId_){
                blockContent.blockState = Empty;
                continue;
            }
            auto iter = neighBorMeshMap_.find(meshIdContainBlock);
            if(iter == neighBorMeshMap_.end()){
                blockContent.blockState = Empty;
                continue;
            }
            blockContent.blockState = NeighborMesh;
            blockContent.neighborMeshId = meshIdContainBlock;
        }
    }
    blockStateChecked_ = true;
    LOG(INFO) <<"checkBlockState Done";
}

void MeshArea::reSetupNeighBorMesh()
{
    auto iter = neighBorMeshMap_.begin();
    for(; iter != neighBorMeshMap_.end(); iter++){
        Isometry3d T;
        Vector2d neghborMeshCenter = iter->second.neighborMesh->getMeshCenter();
        LocalCartesian proj(neghborMeshCenter.y(), neghborMeshCenter.x(), 0);
        proj.Forward(meshCenter_.y(), meshCenter_.x(), 0,
                     T.translation().x(), T.translation().y(), T.translation().z());
        T.linear() = get_rotation(meshCenter_.y(), meshCenter_.x(), 0,
                                  neghborMeshCenter.y(), neghborMeshCenter.x(), 0);
        iter->second.neighborT = T;
    }
}

bool MeshArea::checkBlockAtAreaToLoad(
        const Vector2i &shiftedIndexToLoad)
{
    Vector2i centerIndex(shiftedIndexToLoad.x(), shiftedIndexToLoad.y());
    for(unsigned int i = 0; i < checkNeighbors_.size(); i++){
        Vector2i curr_index = centerIndex + checkNeighbors_[i];
        const BlockContent &blockContent = blockAreaTable_[curr_index.x()][curr_index.y()];
        if(blockContent.blockState == NoneCheck){
            return false;
        }
        if(blockContent.blockState == CurrMesh &&
                !blockContent.block->isLoad()){
            return false;
        }
        if(blockContent.blockState == NeighborMesh){
            return false;
        }
    }
    return true;
}

void MeshArea::loadBlockAtAreaByBlh(const Vector2d &bl)
{
    Vector3d poseInMesh;
    proj_->Forward(bl.x(), bl.y(), 0,
                   poseInMesh.x(), poseInMesh.y(), poseInMesh.z());
    loadBlockAtArea(poseInMesh);
}

void MeshArea::loadBlockAtArea(const Vector3d &position)
{
    uint16_t blockCntLoaded = 0;
    Vector2i centerIndex =
            calShiftedIndex(position.x(), position.y());
    vector<Vector2i> neighbors;
    if(fstLoadMap){
        neighbors = firstLoadNeighbors_;
        fstLoadMap = false;
    }else{
        neighbors = loadNeighbors_;
    }
    for(unsigned int i = 0; i < neighbors.size(); i++){
        Vector2i curr_index = centerIndex + neighbors[i];
        BlockContent &blockContent = blockAreaTable_[curr_index.x()][curr_index.y()];
        if(blockContent.blockState == CurrMesh){
            if(!blockContent.block->isLoad()){
                blockContent.block->load();
                blockCntLoaded++;
            }
            continue;
        }
        if(blockContent.blockState == Empty){
            continue;
        }
        int32_t blockIndexX = curr_index.x() - BlockCntPerMeshSide / 2;
        int32_t blockIndexY = curr_index.y() - BlockCntPerMeshSide / 2;
        float blockCenterX = (float(blockIndexX) + 0.5) * blockResolution_ / 100;
        float blockCenterY = (float(blockIndexY) + 0.5) * blockResolution_ / 100;
        proj_->Reverse(blockCenterX, blockCenterY, 0,
                       blockContent.blockCenterBLH.B,
                       blockContent.blockCenterBLH.L,
                       blockContent.blockCenterBLH.H);
        ulong neighborMeshId = GetMeshFromPosition(
                    blockContent.blockCenterBLH.L, blockContent.blockCenterBLH.B);
        auto iter = neighBorMeshMap_.find(neighborMeshId);
        if(iter == neighBorMeshMap_.end()){
            blockContent.blockState = Empty;
            continue;
        }
        Vector3d blockCenter2Curr(blockCenterX, blockCenterY, 0);
        Vector3d blockCenter2Neighbor = iter->second.neighborT * blockCenter2Curr;
        boost::shared_ptr<MeshArea> &neighbor = iter->second.neighborMesh;
        if(!neighbor->IsInited()){
            neighbor->init();
        }
        Vector2i shiftedIndex;
        if(neighbor->hasActiveBlock(blockCenter2Neighbor, shiftedIndex)){
            blockContent.blockState = NeighborMeshActive;
            blockContent.neighborMeshId = neighborMeshId;
            blockContent.shiftedIndex = shiftedIndex;
            auto blockInNeighbor =
                    neighbor->getBlockByShiftedIndex(shiftedIndex);
            blockInNeighbor->load();
        }else{
            blockContent.blockState = Empty;
        }
    }
    LOG(INFO) << "blockCntLoaded: " <<blockCntLoaded;
}

}
