#ifndef MESHAREA_H
#define MESHAREA_H

#include "manager/squareareamanager.h"
#include "../../interface/voxel_map_define.h"
#include <GeographicLib/LocalCartesian.hpp>
#include "blockarea.h"
#include "common/header.pb.h"
#include "block/block_index.pb.h"
#include "block/block_info.pb.h"
#include "distribution/distribution.pb.h"

using namespace GeographicLib;
using namespace juefx;

namespace localize {

class MeshArea : public SquareAreaManager
{
public:
    enum MeshBackEndType{
        LoadAread = 100
    };
    enum BlockState{
        NoneCheck = 0,
        CurrMesh,
        NeighborMesh,
        NeighborMeshActive,
        Empty
    };
    struct BlockContent{
        BlockState blockState = NoneCheck;
        boost::shared_ptr<BlockArea> block;
        ulong neighborMeshId;
        Vector2i shiftedIndex;
        GnssPoint blockCenterBLH;
    };
    struct MeshContent{
        boost::shared_ptr<MeshArea> neighborMesh;
        Isometry3d neighborT;
    };

    struct SByteId
    {
        int32_t x = 0;
        int32_t y = 0;
    };
public:
    MeshArea(const string& meshPath);

    void init();

    void backendFunc(uint8_t type, Vector2d point = Vector2d::Zero()) override;

    inline bool IsAreaLoaded(const GnssPoint &blh_pose)
    {
        blhToLoad_ = blh_pose;
        if(!inited_){
            return false;
        }
        Vector3d poseInMesh;
        proj_->Forward(blh_pose.B, blh_pose.L, blh_pose.H,
                       poseInMesh.x(), poseInMesh.y(), poseInMesh.z());
        Vector2i shiftedIndexToLoad =
                calShiftedIndex(poseInMesh.x(), poseInMesh.y());
        return checkBlockAtAreaToLoad(shiftedIndexToLoad);
    }

    inline bool IsAreaLoaded(const Vector3d &pose)
    {
        if(!inited_){
            return false;
        }
        double x = pose.x() - offset_.x();
        double y = pose.y() - offset_.y();
        Vector2i shiftedIndexToLoad =
                calShiftedIndex(x, y);
        return checkBlockAtAreaToLoad(shiftedIndexToLoad);
    }

    void setMeshId(ulong meshId){
        meshId_ = meshId;
        meshCenter_ = GetMeshCenter(meshId_);
    }

    ulong getMeshId(){
        return meshId_;
    }

    void insertNeighborMesh(
            ulong meshId, MeshContent &neighborMeshContent){
        neighBorMeshMap_.insert(make_pair(meshId, neighborMeshContent));
    }

    inline boost::shared_ptr<BlockArea>
    getBlockByPoint(float x, float y){
        int32_t blockIndexX = floor(x * 100 / blockResolution_);
        int32_t blockIndexY = floor(y * 100 / blockResolution_);
        uint32_t shiftedIndexX = blockIndexX + shiftSideItemOffset_;
        uint32_t shiftedIndexY = blockIndexY + shiftSideItemOffset_;
        return blockAreaTable_[shiftedIndexX][shiftedIndexY].block;
    }

    inline boost::shared_ptr<BlockArea>
    getBlockByShiftedIndex(const Vector2i &shiftedIndex){
        return blockAreaTable_[shiftedIndex.x()][shiftedIndex.y()].block;
    }

    inline bool hasActiveBlock(const Vector3d &point,
                               Vector2i &shiftedIndex){
        if(!inited_){
            return false;
        }
        shiftedIndex = calShiftedIndex(point.x(), point.y());
        const BlockContent &blockContent =
                blockAreaTable_[shiftedIndex.x()][shiftedIndex.y()];
        if(blockContent.block){
            return true;
        }
        return false;
    }

    inline bool hasActiveBlock(const GnssPoint &blh_pose){
        if(!inited_){
            return false;
        }
        Vector3d poseInMesh;
        proj_->Forward(blh_pose.B, blh_pose.L, blh_pose.H,
                       poseInMesh.x(), poseInMesh.y(), poseInMesh.z());
        Vector2i unused;
        return hasActiveBlock(poseInMesh, unused);
    }

    Vector2d getMeshCenter(){
        return meshCenter_;
    }

    inline void Forward(
            double lat, double lon, double h, double& x, double& y, double& z)
    {
        proj_->Forward(lat, lon, h, x, y, z);
    }

    inline void Reverse(
            double x, double y, double z, double& lat, double& lon, double& h)
    {
        proj_->Reverse(x, y, z, lat, lon, h);

    }

    inline bool IsInited(){
        return inited_;
    }

    float getGridResolution(){
        return (float)gridResolution_ / 100;
    }

    inline juefx::BlockGridIndex GetCellIndex(
            float x, float y) const{
        x -= offset_.x();
        y -= offset_.y();
        juefx::BlockGridIndex ret;
        int32_t blockIndexX = floor(x * 100 / blockResolution_);
        int32_t blockIndexY = floor(y * 100 / blockResolution_);
        uint32_t shiftedIndexX = blockIndexX + shiftSideItemOffset_;
        uint32_t shiftedIndexY = blockIndexY + shiftSideItemOffset_;
        const BlockContent &blockContent =
                blockAreaTable_[shiftedIndexX][shiftedIndexY];
        if(blockContent.blockState == NeighborMeshActive){
            Vector3d currPoint(x, y, 0);
            neighBorMeshId_ = blockContent.neighborMeshId;
            Vector3d neighborPoint =
                    neighBorMeshMap_.at(neighBorMeshId_).neighborT * currPoint;
            x = neighborPoint.x();
            y = neighborPoint.y();
            blockIndexX = floor(x * 100 / blockResolution_);
            blockIndexY = floor(y * 100 / blockResolution_);
        }else{
            neighBorMeshId_ = 0;
        }
        float xInBlock = x - blockIndexX * (float)blockResolution_ / 100;
        float yInBlock = y - blockIndexY * (float)blockResolution_ / 100;
        uint32_t gridIndexX = floor(xInBlock * 100 / gridResolution_);
        uint32_t gridIndexY = floor(yInBlock * 100 / gridResolution_);
        ret.x.blockId = blockIndexX;
        ret.y.blockId = blockIndexY;
        ret.x.gridId = gridIndexX;
        ret.y.gridId = gridIndexY;
        return ret;
    }

    inline juefx::Distribution* GetCellValue(
            uint8_t type, const BlockGridIndex &index, uint8_t &length) const{
        if(neighBorMeshId_ == 0){
            return GetCurrMeshCellValue(type, index, length);
        }else{
            const auto &neighbor =
                    neighBorMeshMap_.at(neighBorMeshId_).neighborMesh;
            return neighbor->GetCurrMeshCellValue(type, index, length);
        }
        return nullptr;
    }

    inline juefx::Distribution* GetCurrMeshCellValue(
            uint8_t type, const BlockGridIndex &index, uint8_t &length) const{
        length = 0;
        int32_t blockIndexX = index.x.blockId;
        int32_t blockIndexY = index.y.blockId;
        uint32_t shiftedIndexX = blockIndexX + shiftSideItemOffset_;
        uint32_t shiftedIndexY = blockIndexY + shiftSideItemOffset_;
        const BlockContent &blockContent =
                blockAreaTable_[shiftedIndexX][shiftedIndexY];
        const auto &block = blockContent.block;
        if(blockContent.blockState == NoneCheck ||
                blockContent.blockState == Empty){
            return nullptr;
        }else if(blockContent.blockState == CurrMesh){
            if(!block->isLoad()){
                return nullptr;
            }
            return block->GetCellValue(type, index, length);
        }
        return nullptr;
    }

    void loadArea(const Vector2d &bl);


    inline void loadArea(const Vector3d &pose)
    {
        if(!inited_){
            init();
        }
        if(!blockStateChecked_){
            checkBlockState();
        }
        loadBlockAtArea(pose);
    }

    void load();

    Vector3f GetOffset(){
        return offset_;
    }

private:

    void checkBlockState();

    inline Vector2i calShiftedIndex(
            double x, double y) const{
        Vector2i blockIndex;
        blockIndex.x() = floor(x * 100 / blockResolution_);
        blockIndex.y() = floor(y * 100 / blockResolution_);
        Vector2i shiftedIndexToLoad;
        shiftedIndexToLoad.x() = blockIndex.x() + shiftSideItemOffset_;
        shiftedIndexToLoad.y() = blockIndex.y() + shiftSideItemOffset_;
        return shiftedIndexToLoad;
    }

    void reSetupNeighBorMesh();

    bool checkBlockAtAreaToLoad(const Vector2i &shiftedIndexToLoad);

    void loadBlockAtAreaByBlh(const Vector2d &bl);

    void loadBlockAtArea(const Vector3d &position);

private:
    string meshPath_;

    bool inited_ = false;

    ulong meshId_;

    uint32_t gridResolution_;   //cm

    uint32_t blockResolution_;   //cm

    uint32_t gridCntPerBlockSide_;

    Vector3f offset_ = Vector3f::Zero();

    uint32_t meshLevel_;

    boost::shared_ptr<LocalCartesian> proj_;

    vector<vector<BlockContent>> blockAreaTable_;

    boost::shared_ptr<Distributions> distributions_;

    uint32_t shiftBlockIdOffset_;

    Vector2d meshCenter_;

    map<ulong, MeshContent> neighBorMeshMap_;

    GnssPoint blhToLoad_;

    vector<GridIndex::ShortId> shiftedIndexVecToLoad_;

    bool blockStateChecked_ = false;

    vector<Vector2i> checkNeighbors_;

    vector<Vector2i> firstLoadNeighbors_;

    vector<Vector2i> loadNeighbors_;

    mutable ulong neighBorMeshId_;

    bool fstLoadMap = true;
};

}
#endif // MESHAREA_H
