#include "dispatch.h"
#include "grid/block.h"

using namespace boost::filesystem;
vector<double> BlockResolutions{BlockResolution};

Dispatch::Dispatch(const string &path)
{
    meshLayer_.reset(new GridLayer(MaxMeshSideLength, BlockResolution));
    meshLayer_->setLayerResolutions(BlockResolutions);
    loadPCD(path);
}

Dispatch::Dispatch(PointCloudExport::Ptr pointcloud)
{
    meshLayer_.reset(new GridLayer(MaxMeshSideLength, BlockResolution));
    meshLayer_->setLayerResolutions(BlockResolutions);
    setPointcloud(pointcloud);
}

bool Dispatch::loadPCD(const string &path)
{
    pointcloud_ = PointCloudExport::Ptr(new PointCloudExport);
    if(exists(path)){
        pcl::io::loadPCDFile(path, *pointcloud_);
        boost::filesystem::path boostPath(path);
        meshId_ = stoi(boostPath.filename().string());
        return true;
    }else{
        return false;
    }
}

void Dispatch::setPointcloud(
        PointCloudExport::Ptr &pointcloud)
{
    pointcloud_ = pointcloud;
}

boost::shared_ptr<GridLayer> Dispatch::getMeshLayer()
{
    return meshLayer_;
}

void Dispatch::configOffset()
{
    double x = 0, y = 0, z = 0;
    for(const auto &point : pointcloud_->points){
        x += point.x;
        y += point.y;
//        z += point.z;
    }
    offset_ = {x / pointcloud_->size(),
               y / pointcloud_->size(),
               z / pointcloud_->size()};
    offset_ = {lround(offset_.x() / GridResolutionHigh) * GridResolutionHigh,
               lround(offset_.y() / GridResolutionHigh) * GridResolutionHigh,
               0
               /*lround(offset_.z() / GridResolutionHigh) * GridResolutionHigh*/};
    for(auto &point : pointcloud_->points){
        point.x -= offset_.x();
        point.y -= offset_.y();
//        point.z -= offset_.z();
    }

}

void Dispatch::dispatch()
{
//    boost::function<void(boost::shared_ptr<LocateGridBase>)> configFunc
//            = boost::bind(&Dispatch::configBlockInfo, this, _1);
    if(!pointcloud_){
        LOG(WARNING) << "NO pointcloud_ !";
        return;
    }

//    boost::function<void(const PointExport &)> dispatchFunc
//            = boost::bind(&Dispatch::dispatchPoint, this, _1);
//    multiThreadProcess<PointExport>(pointcloud_, dispatchFunc);

    for(size_t i = 0; i < pointcloud_->size(); i++){
        const auto &point = pointcloud_->at(i);
        dispatchPoint(point);
    }

    pointcloud_.reset();

    auto gridTabel = meshLayer_->getGridTable();
    vector<boost::shared_ptr<LocateGridBase>> blockGridVec;
    for(auto raw : gridTabel){
        for(auto grid : raw){
            if(!grid){
                continue;
            }
            blockGridVec.push_back(grid);
        }
    }
    boost::function<void(boost::shared_ptr<LocateGridBase>)> configFunc
            = boost::bind(&Dispatch::configBlockInfo, this, _1);
    syncProcess<boost::shared_ptr<LocateGridBase>>(blockGridVec, configFunc);

}

Vector3d Dispatch::getOffset()
{
    return offset_;
}

void Dispatch::setMeshId(ulong meshId)
{
    meshId_ = meshId;
}

ulong Dispatch::getMeshId()
{
    return meshId_;
}

void Dispatch::dispatchPoint(const PointExport &point)
{
//    PointExport p;
//    p.x = 145.75; p.y = 62.14; p.z = -5.83;
//    auto bi = meshLayer_->getGridIndexByPoint(p);
//    auto* blk = reinterpret_cast<BlockGrid*>(
//                meshLayer_->getBlock(bi).get());
//    vector<boost::shared_ptr<GridLayer>> bl =
//            blk->getLayers();
//    auto gi = bl[0]->getGridIndexByPoint(p);

    Vector2i blockIndex = meshLayer_->getGridIndexByPoint(point);
    BlockGrid* block = reinterpret_cast<BlockGrid*>(
                meshLayer_->getBlock(blockIndex).get());
    vector<boost::shared_ptr<GridLayer>> blockLayers =
            block->getLayers();
    for(const auto &blockLayer : blockLayers){
        Vector2i gridIndex = blockLayer->getGridIndexByPoint(point);
        if(gridIndex.x() > 127 || gridIndex.x() < -128){
            continue;
        }
        if(gridIndex.y() > 127 || gridIndex.y() < -128){
            continue;
        }
//        if(blockIndex.x() != 6 || blockIndex.y() != 2 ||
//                gridIndex.x() != -45 || gridIndex.y() != 50){
//            continue;
//        }
//        LOG(INFO) << gridIndex.transpose();
        Grid* grid = reinterpret_cast<Grid*>(blockLayer->getGrid(gridIndex).get());
        if(!grid){
            continue;
        }
        grid->addPoint(point);
    }

}

void Dispatch::configBlockInfo(
        boost::shared_ptr<LocateGridBase> blockGrid){
    BlockGrid* block = reinterpret_cast<BlockGrid*>(blockGrid.get());
    block->configInfo();
}
