#include "blockserializer.h"
#include "common/header.pb.h"
#include "block/block_index.pb.h"
#include "block/block_info.pb.h"
#include "utils/basic_functions.h"
#include <google/protobuf/io/coded_stream.h>
#include <google/protobuf/io/zero_copy_stream_impl_lite.h>

BlockSerializer::BlockSerializer(boost::shared_ptr<GridLayer> gridLayer)
    :meshLayer_(gridLayer)
{

}

bool checkOneDistribution_(
    const locate::OneDistribution &source,
    const locate::OneDistribution &target){
    if(source.floor == target.floor
            && source.z_delta_level == target.z_delta_level
            && source.intensity == target.intensity){
        return true;
    }else{
        return false;
    }
}

bool checkGridDistribution_(
        const locate::GridDistribution &source,
        const locate::GridDistribution &target){
    if(source.oneDistributions.size()
            != target.oneDistributions.size()){
        return false;
    }
    for(size_t i = 0; i < source.oneDistributions.size(); i++){
        if(!checkOneDistribution_(
                    source.oneDistributions[i],
                    target.oneDistributions[i])){
            return false;
        }
    }
    return true;
}

struct DistributionInfo{
    locate::GridDistribution gridDistribution;
    uint32_t cnt = 0;
    uint32_t id = 0;
};

void BlockSerializer::configSerialization()
{
    string serializedStream;
    auto blockTabel = meshLayer_->getGridTable();
    vector<locate::BlockInfo> blockInfoVec;
    map<int64_t, DistributionInfo> distributionMap;
    for(auto raw : blockTabel){
        for(auto blockPtr : raw){
            if(!blockPtr){
                continue;
            }
            BlockGrid* block = reinterpret_cast<BlockGrid*>(blockPtr.get());
            pcl::ModelCoefficients::Ptr blockPlaneParam =
                    block->getplaneParam();
            int32_t blockId = meshLayer_->encodeIndex(block->getIndex());

            locate::BlockInfo blockInfo_;
            blockInfo_.block_id = blockId;
            if(blockPlaneParam->values.size() >= 4){
                blockInfo_.a = blockPlaneParam->values[0];
                blockInfo_.b = blockPlaneParam->values[1];
                blockInfo_.c = blockPlaneParam->values[2];
                blockInfo_.d = blockPlaneParam->values[3];
            }


            vector<boost::shared_ptr<GridLayer>> blockLayers =
                    block->getLayers();
            if(blockLayers.size() == 0){
                return;
            }
            locate::BlockDistribution blockDistribution =
                    block->getblockDistribution();

            for(size_t grid_id = 0; grid_id < blockDistribution.gridDistributions.size(); grid_id++){
                locate::GridDistribution gridDistribution = blockDistribution.gridDistributions[grid_id];
                int64_t key = gridDistribution.getScore();
                if(key == 0){
                    continue;
                }
                blockInfo_.activeGridCnt++;
                auto iter = distributionMap.find(key);
                if(iter == distributionMap.end()){
                    DistributionInfo distributionInfo;
                    distributionInfo.gridDistribution = gridDistribution;
                    distributionInfo.cnt = 1;
                    distributionMap.insert(make_pair(key, distributionInfo));
                }else{
                    iter->second.cnt++;
                }
            }
            blockInfo_.grid_resolution = blockLayers[0]->getResolution();
            blockInfo_.bit_map.resize(pow(BlockResolution / blockInfo_.grid_resolution, 2) / 8);

    //        blockInfo_.grid_feature_index.resize(90000, 0);
            blockInfoVec.push_back(blockInfo_);
        }
    }

    vector<DistributionInfo> distributionInfoVec;
    for(auto iter = distributionMap.begin(); iter != distributionMap.end(); iter++){
        distributionInfoVec.push_back(iter->second);
    }
    sort(distributionInfoVec.begin(),
         distributionInfoVec.end(),
         [](const DistributionInfo& query, const DistributionInfo& target) {
             return query.cnt > target.cnt;
         });


    map<int64_t, DistributionInfo>().swap(distributionMap);
    for(size_t i = 0; i < distributionInfoVec.size(); i++){
        auto distributionInfo = distributionInfoVec[i];
        distributionInfo.id = i;
        distributionMap.insert(make_pair(distributionInfo.gridDistribution.getScore(), distributionInfo));
    }


    size_t index = 0;
    size_t activeCnt = 0;
    for(auto raw : blockTabel){
        for(auto blockPtr : raw){
            if(!blockPtr){
                continue;
            }
            BlockGrid* block = reinterpret_cast<BlockGrid*>(blockPtr.get());
            int32_t blockId = meshLayer_->encodeIndex(block->getIndex());
            vector<boost::shared_ptr<GridLayer>> blockLayers =
                    block->getLayers();

            if(blockLayers.size() == 0){
                return;
            }
            locate::BlockDistribution blockDistribution =
                    block->getblockDistribution();

            uint8_t byte = 0;
            uint8_t bitIndex;
            for(size_t grid_id = 0; grid_id < blockDistribution.gridDistributions.size(); grid_id++){
                activeCnt++;
                locate::GridDistribution gridDistribution = blockDistribution.gridDistributions[grid_id];
                uint32_t byteIndex = grid_id / 8;
                bitIndex = grid_id % 8;
                int64_t key = gridDistribution.getScore();
                auto iter = distributionMap.find(key);
                if(iter != distributionMap.end()){
                    size_t distributionId = iter->second.id;
                    byte |= 1 << bitIndex;
                    blockInfoVec[index].active_grid_feature_index.push_back(distributionId);
                    blockInfoVec[index].whole_grid_feature_index.push_back(distributionId + 1);
                }else{
                    blockInfoVec[index].whole_grid_feature_index.push_back(0);
                }
                if(bitIndex == 7){
                    blockInfoVec[index].bit_map[byteIndex] = byte;
                    byte = 0;
                }
            }
            if(bitIndex != 7){
                blockInfoVec[index].bit_map.back() = byte;
            }
            index++;
            blockPtr.reset();
        }
    }


    /////////// BlockIndex /////////////
    /// \brief blockIndex
    ///

    BlockIndex blockIndex;

    /////////// BlockInfo /////////////
    /// \brief blockInfo
    ///

    string blockInfoBitmapStream, blockInfoIndexStream;
    uint32_t block_offset = 0;


    LOG(INFO) << "blockInfoVec.size(): " << blockInfoVec.size();
    int blockCnt = 0;
    for(auto &blockInfo_ : blockInfoVec){
        BlockInfo blockInfoBitmap;
        for(auto byte : blockInfo_.bit_map){
            blockInfoBitmap.add_bit_map(byte);
        }
        for(auto gridIndex : blockInfo_.active_grid_feature_index){
            blockInfoBitmap.add_grid_feature_index(gridIndex);
        }
        blockInfoBitmap.set_grid_resolution(blockInfo_.grid_resolution);
        blockInfoBitmap.set_a(blockInfo_.a);
        blockInfoBitmap.set_b(blockInfo_.b);
        blockInfoBitmap.set_c(blockInfo_.c);
        blockInfoBitmap.set_d(blockInfo_.d);
        blockCnt++;

        string currBlockInfoBitmap = blockInfoBitmap.SerializeAsString();
        blockInfoBitmapStream += currBlockInfoBitmap;
        BlockDescribe* blockDescribe = blockIndex.add_block_describe();
        blockDescribe->set_block_id(blockInfo_.block_id);
        blockDescribe->set_block_info_size(currBlockInfoBitmap.size());
        blockDescribe->set_block_info_offset(block_offset);
        block_offset = blockInfoBitmapStream.size();

        BlockInfo blockInfoIndex;
        for(auto gridIndex : blockInfo_.whole_grid_feature_index){
            blockInfoIndex.add_grid_feature_index(gridIndex);
//            if(gridIndex >= 2592551){
//                LOG(INFO) << "gridIndex: " << gridIndex;
//            }
        }
        blockInfoIndex.set_grid_resolution(blockInfo_.grid_resolution);
        blockInfoIndex.set_a(blockInfo_.a);
        blockInfoIndex.set_b(blockInfo_.b);
        blockInfoIndex.set_c(blockInfo_.c);
        blockInfoIndex.set_d(blockInfo_.d);
        blockInfoIndexStream += blockInfoIndex.SerializeAsString();
    }
    LOG(INFO) << "blockCnt: " << blockCnt;
    vector<locate::BlockInfo>().swap(blockInfoVec);

    Distributions distributions;
    LOG(INFO) << "distributionInfoVec.size(): " << distributionInfoVec.size();
    int activeGridDistributionCnt = 0;
    for(auto &distributionInfo : distributionInfoVec){
        GridDistribution* gridDistribution = distributions.add_grid_distribution();
        bool active = false;
        for(auto &oneDistribution : distributionInfo.gridDistribution.oneDistributions){
            OneDistribution* protoOneDistribution = gridDistribution->add_one_distribution();
            protoOneDistribution->set_floor(oneDistribution.floor);
            protoOneDistribution->set_z_delta_leval(oneDistribution.z_delta_level);
            protoOneDistribution->set_intensity(oneDistribution.intensity);
            active = true;
        }
        if(active){
            activeGridDistributionCnt++;
        }
    }
    LOG(INFO) << "activeGridDistributionCnt: " << activeGridDistributionCnt;
    vector<DistributionInfo>().swap(distributionInfoVec);
    LOG(INFO) << "grid_distribution_size: " << distributions.grid_distribution_size();

    /////////// Header /////////////
    /// \brief header
    ///

    Header header;
    header.set_save_type(Header_SaveType_DIRECT);
    header.set_mesh_level(15);
    header.set_mesh_id(meshId_);
    header.set_block_relolution(BlockResolution);
    header.set_high_relolution(GridResolutionHigh);
    header.set_low_relolution(GridResolutionLow);
    LOG(INFO) << setprecision(15) << "offset_: " << offset_.transpose();
    header.set_offset_x(offset_.x());
    header.set_offset_y(offset_.y());
    header.set_offset_z(offset_.z());
//    header.set_longtitude(121.131722586496);
//    header.set_latitude(31.1529105050223);
    header.set_latitude(blh_.x());
    header.set_longtitude(blh_.y());
    header.set_height(blh_.z());

    LOG(INFO) << "blockInfoBitmapStream.size(): " << blockInfoBitmapStream.size();
    LOG(INFO) << "blockInfoIndexStream.size(): " << blockInfoIndexStream.size();

    string blockInfoStream;
    if(blockInfoBitmapStream.size() <  blockInfoIndexStream.size()){
        blockInfoStream = blockInfoBitmapStream;
        header.set_save_type(Header_SaveType_BIT);
        LOG(INFO) << "Header_SaveType_BIT";
    }else{
        blockInfoStream = blockInfoIndexStream;
        header.set_save_type(Header_SaveType_DIRECT);
        LOG(INFO) << "Header_SaveType_DIRECT";
    }

    header.set_active_block_count(activeCnt);

    LOG(INFO) << "blockIndex.block_describe_size(): " << blockIndex.block_describe_size();

    string blockIndexStream = blockIndex.SerializeAsString();
    string distributionsStream = distributions.SerializeAsString();

    header.set_block_index_length(blockIndexStream.size());
    header.set_block_info_length(blockInfoBitmapStream.size());

    string headerStream = header.SerializeAsString();

    serializedStream += headerStream;
    LOG(INFO) << "headerStream.size(): " << headerStream.size();
    serializedStream += blockIndexStream;
    LOG(INFO) << "blockIndexStream.size(): " << blockIndexStream.size();
    serializedStream += blockInfoStream;
    LOG(INFO) << "blockInfoStream.size(): " << blockInfoStream.size();
    serializedStream += distributionsStream;
    LOG(INFO) << "distributionsStream.size(): " << distributionsStream.size();

    uint16_t headerSize = headerStream.size();

    std::ofstream ofs(outputPath_, ios::binary);
    if(!ofs){
        LOG(WARNING) << outputPath_ << " load fail!";
        return;
    }
    ofs << (char)headerSize << (char)(headerSize >> 8);
    ofs << serializedStream;
\
    ofs.close();

//    std::ofstream of(outputPath_ + ".offset");
//    of << offset_.x() << "," << offset_.y();
//    of.close();

}

void BlockSerializer::setOffset(const Vector3d &offset)
{
    offset_ = offset;
}

void BlockSerializer::setMeshId(const string &meshId)
{
    meshId_ = meshId;
}

void BlockSerializer::setBLH(const Vector3d &blh)
{
    blh_ = blh;
}

void BlockSerializer::setOutputPath(
        const string &outputPath)
{
    outputPath_ = outputPath;
}
