#include "streaminfo.h"

StreamInfo::StreamInfo(const string &streamPath)
    :streamPath_(streamPath)
{

}
static size_t activeGridCnt = 0, distribCnt = 0;

bool StreamInfo::load()
{
    ifstream infile(streamPath_, ios::binary);
    if(!infile)
    {
        LOG(WARNING) << streamPath_ << " open error!";
        return false;
    }
    char buf[2];
    infile.read(buf, 2);
    headerLength_ = ((uint16_t)buf[1] << 8) + (uint16_t)buf[0];

    char* headerData = (char*)malloc(headerLength_);
    infile.read(headerData, headerLength_);
    header_.ParseFromArray(headerData, headerLength_);
    free(headerData);
    try {
        meshId_ = stoi(header_.mesh_id());
    }
    catch (std::invalid_argument&){
        LOG(INFO) << "header.mesh_id() is not int";
    }
    highResolution = header_.high_relolution();
    lowResolution = header_.low_relolution();
    blockResolution = header_.block_relolution();
    offset_.x() = header_.offset_x();
    offset_.y() = header_.offset_y();
    meshLevel_ = header_.mesh_level();

    meshLayer_.reset(new GridLayer(MaxMeshSideLength, blockResolution));
    meshLayer_->setOrigin(offset_);
    meshLayer_->setLayerResolutions({lowResolution, highResolution});

    proj_.reset(new LocalCartesian(header_.latitude(),
                                   header_.longtitude(),
                                   header_.height()));

    uint16_t blockIndexLength = header_.block_index_length();

    char* blockIndexData = (char*)malloc(blockIndexLength);
    infile.read(blockIndexData, blockIndexLength);
    blockIndex_.ParseFromArray(blockIndexData, blockIndexLength);
    free(blockIndexData);
    size_t block_describe_size = blockIndex_.block_describe_size();
    for(size_t i = 0; i < block_describe_size; i++){
        BlockDescribe blockDescribe = blockIndex_.block_describe(i);
        Internal::BlockDescribe internalBlockDescribe;
        internalBlockDescribe.blockId = blockDescribe.block_id();
        internalBlockDescribe.blockOffset = blockDescribe.block_info_offset();
        internalBlockDescribe.blockSize = blockDescribe.block_info_size();
        blockDescribeVec_.push_back(internalBlockDescribe);
    }

    infile.clear();
    infile.seekg(0, infile.end);
    uint32_t totalSize = infile.tellg();
    infile.clear();
    infile.seekg(2 + headerLength_ +
                 header_.block_index_length() + header_.block_info_length());

    uint32_t distributionSize = totalSize - 2 - headerLength_ -
            header_.block_index_length() - header_.block_info_length();

    char* distributionBuf = (char*)malloc(distributionSize);

    infile.read(distributionBuf, distributionSize);
    infile.close();

    Distributions_.ParseFromArray(distributionBuf, distributionSize);
    free(distributionBuf);
    extractedPointcloudLayers_.push_back(
                PointCloudXYZI::Ptr(new PointCloudXYZI));

    for(size_t i = 0; i < blockDescribeVec_.size(); i++){
        auto blockDescribe = blockDescribeVec_[i];
        parseOneBlock(blockDescribe);

    }

    LOG(INFO) << "activeGridCnt: " << activeGridCnt
              << " distribCnt: : " << distribCnt
              << " avg: " << (float)distribCnt / activeGridCnt;
    calMeshVerts();
    return true;
}

void StreamInfo::parseOneBlock(
        const Internal::BlockDescribe &internalBlockDescribe)
{
    ifstream infile(streamPath_, ios::binary);
    if(!infile)
    {
        LOG(WARNING) << streamPath_ << " open error!";
        return ;
    }

    char* blockInfoData = (char*)malloc(internalBlockDescribe.blockSize);
    uint32_t infoOffset = 2 + headerLength_ + header_.block_index_length() + internalBlockDescribe.blockOffset;
    infile.clear();
    infile.seekg(infoOffset);
    infile.read(blockInfoData, internalBlockDescribe.blockSize);
    BlockInfo blockInfo;
    blockInfo.ParseFromArray(blockInfoData, internalBlockDescribe.blockSize);
    free(blockInfoData);
    infile.close();

//    LOG(INFO) << "blockIndexX: " << blockIndexX;
//    LOG(INFO) << "blockIndexY: " << blockIndexY;
//    LOG(INFO) << "blockCenterX: " << blockCenterX;
//    LOG(INFO) << "blockCenterY: " << blockCenterY;

    if(blockInfo.bit_map_size()){
        parseBitmapBlock(blockInfo,
                         Distributions_,
                         internalBlockDescribe);
    }else{
        parseIndexBlock(blockInfo,
                        Distributions_,
                        internalBlockDescribe);

    }


}

vector<uint32_t> StreamInfo::parseBitMap(
        const ProtoRepeatedUint32 &bitmap,
        const uint32_t size)
{
    vector<uint32_t> activeGridIndexes;
    uint32_t activeGridIndex = 0;
    for(size_t i = 0; i < size; i++){
        uint8_t byte = bitmap.Get(i);
        for(size_t byteCnt = 0; byteCnt < 8; byteCnt++){
            if((byte >> byteCnt) & 1){
                activeGridIndexes.push_back(activeGridIndex);
            }
            activeGridIndex++;
        }
    }
//    LOG(INFO) << "activeGridIndexes.size(): " << activeGridIndexes.size();
    return activeGridIndexes;
}

float StreamInfo::getHighResolution()
{
    return highResolution;
}

float StreamInfo::getLowResolution()
{
    return lowResolution;
}

float StreamInfo::getBlockResolution()
{
    return blockResolution;
}

void StreamInfo::parseBitmapBlock(const BlockInfo &blockInfo,
        const Distributions& distribution,
        const Internal::BlockDescribe &internalBlockDescribe)
{
    float gridResolution = blockInfo.grid_resolution();
    int32_t blockPerMeshItemCnt = round(MaxMeshSideLength / blockResolution);
    int32_t blockId = internalBlockDescribe.blockId;
    int32_t shiftBlockId = blockId +
            (blockPerMeshItemCnt / 2) * (blockPerMeshItemCnt + 1);
    int32_t shiftBlockIndexX = floor((float)shiftBlockId / blockPerMeshItemCnt);
    int32_t shiftBlockIndexY = shiftBlockId % blockPerMeshItemCnt;
    int32_t blockIndexX = shiftBlockIndexX - blockPerMeshItemCnt / 2;
    int32_t blockIndexY = shiftBlockIndexY - blockPerMeshItemCnt / 2;
    float blockCenterX = (float)(blockIndexX + 0.5) * blockResolution;
    float blockCenterY = (float)(blockIndexY + 0.5) * blockResolution;

    vector<uint32_t> activeGridIndexes =
            parseBitMap(blockInfo.bit_map(), blockInfo.bit_map_size());
    activeGridCnt += activeGridIndexes.size();

    auto gridBase =
            meshLayer_->getBlock({blockIndexX, blockIndexY});
    BlockGrid* block = reinterpret_cast<BlockGrid*>(gridBase.get());
    const vector<boost::shared_ptr<GridLayer>>& blockLayers =
            block->getLayers();
    boost::shared_ptr<GridLayer> currBlockLayer;
//    if(blockLayer.grid_resolution() == highResolution){
//        currBlockLayer = blockLayers[1];
//    }else{
//        currBlockLayer = blockLayers[0];
//    }
    currBlockLayer = blockLayers[0];


    for(size_t i = 0; i < activeGridIndexes.size(); i++){
        auto index = activeGridIndexes[i];
        auto grid_feature_index = blockInfo.grid_feature_index(i);
        GridDistribution gridDistribution = distribution.grid_distribution(grid_feature_index);
        uint32_t indexX = index / (uint32_t)(blockResolution / gridResolution);
        uint32_t indexY = index % (uint32_t)(blockResolution / gridResolution);

        Grid* grid = reinterpret_cast<Grid*>(
                    currBlockLayer->getGrid({indexX, indexY}).get());
        if(grid == nullptr){
            continue;
        }

        size_t oneDistributionCnt = gridDistribution.one_distribution_size();
        distribCnt += oneDistributionCnt;
        locate::GridDistribution gridDistribution_;
        for(size_t i = 0; i < oneDistributionCnt; i++){
            OneDistribution oneDistribution = gridDistribution.one_distribution(i);
            locate::OneDistribution oneDistribution_;
            oneDistribution_.z_delta_level = oneDistribution.z_delta_leval();
            oneDistribution_.floor = oneDistribution.floor();
            oneDistribution_.intensity = oneDistribution.intensity();
            gridDistribution_.oneDistributions.push_back(oneDistribution_);
        }
        grid->setGridDistribution(gridDistribution_);

    }
    block->configMultiLayer();
}

void StreamInfo::parseIndexBlock(const BlockInfo &blockInfo,
        const Distributions& distribution,
        const Internal::BlockDescribe &internalBlockDescribe)
{
    float gridResolution = blockInfo.grid_resolution();
    int32_t blockPerMeshItemCnt = round(MaxMeshSideLength / blockResolution);
    int32_t blockId = internalBlockDescribe.blockId;
    int32_t shiftBlockId = blockId +
            (blockPerMeshItemCnt / 2) * (blockPerMeshItemCnt + 1);
    int32_t shiftBlockIndexX = floor((float)shiftBlockId / blockPerMeshItemCnt);
    int32_t shiftBlockIndexY = shiftBlockId % blockPerMeshItemCnt;
    int32_t blockIndexX = shiftBlockIndexX - blockPerMeshItemCnt / 2;
    int32_t blockIndexY = shiftBlockIndexY - blockPerMeshItemCnt / 2;
    float blockCenterX = (float)(blockIndexX + 0.5) * blockResolution;
    float blockCenterY = (float)(blockIndexY + 0.5) * blockResolution;

    BlockGrid* block = reinterpret_cast<BlockGrid*>(
                meshLayer_->getBlock({blockIndexX, blockIndexY}).get());
    vector<boost::shared_ptr<GridLayer>> blockLayers =
            block->getLayers();
    boost::shared_ptr<GridLayer> currBlockLayer;
    if(blockInfo.grid_resolution() == highResolution){
        currBlockLayer = blockLayers[1];
    }else{
        currBlockLayer = blockLayers[0];
    }
    for(size_t i = 0; i < blockInfo.grid_feature_index_size(); i++){
        auto grid_feature_index = blockInfo.grid_feature_index(i);
        if(grid_feature_index == 0){
            continue;
        }
        grid_feature_index--;

        GridDistribution gridDistribution = distribution.grid_distribution(grid_feature_index);
        uint32_t indexX = i / (blockResolution / gridResolution);
        uint32_t indexY = i % (uint32_t)(blockResolution / gridResolution);
        size_t oneDistributionCnt = gridDistribution.one_distribution_size();
        Grid* grid = reinterpret_cast<Grid*>(
                    currBlockLayer->getGrid({indexX, indexY}).get());

        activeGridCnt++;
        distribCnt += oneDistributionCnt;

        locate::GridDistribution gridDistribution_;
        for(size_t i = 0; i < oneDistributionCnt; i++){
            OneDistribution oneDistribution = gridDistribution.one_distribution(i);
            locate::OneDistribution oneDistribution_;
            oneDistribution_.z_delta_level = oneDistribution.z_delta_leval();
            oneDistribution_.floor = oneDistribution.floor();
            oneDistribution_.intensity = oneDistribution.intensity();
            grid->setGridDistribution(gridDistribution_);
        }
    }
    block->configMultiLayer();
}

void StreamInfo::calMeshVerts()
{
    Vector4d lurd = GetMeshBound(meshId_, meshLevel_);
    double unused;
    proj_->Forward(lurd[1], lurd[0], 0, meshVerts_[0].x(), meshVerts_[0].y(), unused);
    proj_->Forward(lurd[1], lurd[2], 0, meshVerts_[1].x(), meshVerts_[1].y(), unused);
    proj_->Forward(lurd[3], lurd[2], 0, meshVerts_[2].x(), meshVerts_[2].y(), unused);
    proj_->Forward(lurd[3], lurd[0], 0, meshVerts_[3].x(), meshVerts_[3].y(), unused);
}

Vector2f StreamInfo::getOffset()
{
    return offset_;
}

uint8_t StreamInfo::checkNearMeshBound(
        const Vector3d &currPos)
{
    uint8_t ret = 0;
    if(currPos.x() - meshVerts_[0].x() < MESH_NEAR_THREHOLD){
        ret |= MESH_LEFT;
    }
    if(meshVerts_[0].y() - currPos.y() < MESH_NEAR_THREHOLD){
        ret |= MESH_UP;
    }
    if(meshVerts_[2].x() - currPos.x() < MESH_NEAR_THREHOLD){
        ret |= MESH_RIGHT;
    }
    if(currPos.y() - meshVerts_[2].y() < MESH_NEAR_THREHOLD){
        ret |= MESH_DOWN;
    }
    return ret;
}

uint8_t StreamInfo::checkAwayMeshBound(const Vector3d &currPos)
{
    uint8_t ret = 0;
    if(currPos.x() - meshVerts_[0].x() > MESH_AWAY_THREHOLD){
        ret |= MESH_LEFT;
    }
    if(meshVerts_[0].y() - currPos.y() > MESH_AWAY_THREHOLD){
        ret |= MESH_UP;
    }
    if(meshVerts_[2].x() - currPos.x() > MESH_AWAY_THREHOLD){
        ret |= MESH_RIGHT;
    }
    if(currPos.y() - meshVerts_[2].y() > MESH_AWAY_THREHOLD){
        ret |= MESH_DOWN;
    }
}

