#ifndef STREAMINFO_H
#define STREAMINFO_H

#include "utils/basic_types.h"
#include "common/header.pb.h"
#include "block/block_index.pb.h"
#include "block/block_info.pb.h"
#include "distribution/distribution.pb.h"
#include <GeographicLib/LocalCartesian.hpp>
#include "grid/gridmanager.h"
#include "grid/block.h"

typedef pcl::PointCloud<pcl::PointXYZI> PointCloudXYZI;
using namespace GeographicLib;

namespace Internal {
struct BlockDescribe{
    int32_t blockId = 0;
    uint32_t blockOffset = 0;
    uint32_t blockSize = 0;
};
}

typedef ::google::protobuf::RepeatedField< ::google::protobuf::uint32 >
ProtoRepeatedUint32;

#define MESH_NEAR_THREHOLD      100
#define MESH_AWAY_THREHOLD      200

#define MESH_LEFT      1
#define MESH_UP        2
#define MESH_RIGHT     4
#define MESH_DOWN      8

class StreamInfo
{
public:
    StreamInfo(const string& streamPath);

    bool load();

    void parseOneBlock(
            const Internal::BlockDescribe &internalBlockDescribe);

    vector<uint32_t> parseBitMap(
            const ProtoRepeatedUint32& bitmap,
            const uint32_t size);

    inline void Forward(
            double lat, double lon, double h, double& x, double& y, double& z)
    {
        proj_->Forward(lat, lon, h, x, y, z);
    }

    inline void Reverse(
            double x, double y, double z, double& lat, double& lon, double& h)
    {
        proj_->Reverse(x, y, z, lat, lon, h);

    }

    float getHighResolution();

    float getLowResolution();

    float getBlockResolution();

   inline  ulong getMeshId()
    {
        return meshId_;
    }

    Vector2f getOffset();

    uint8_t checkNearMeshBound(const Vector3d &currPos);

    uint8_t checkAwayMeshBound(const Vector3d &currPos);
    inline int8_t IsGridInArea(const Vector2i &index){
        return meshLayer_->IsGridActive(index);
    }

    inline const juefx::GridData *getGridData(uint8_t type,
                                const juefx::GridIndex::ByteId &blockIndex,
                                const juefx::GridIndex::ShortId &gridIndex)
    {
        juefx::GridData* ret = NULL;
        LocateGridBase* blockGrid;
        if(!meshLayer_->getGridBase({blockIndex.x, blockIndex.y}, blockGrid)){
            return ret;
        }
        BlockGrid* block = reinterpret_cast<BlockGrid*>(blockGrid);
        return block->getGridData(type, gridIndex);
    }
private:
    void parseBitmapBlock(
            const BlockInfo &blockInfo,
            const Distributions &distribution,
            const Internal::BlockDescribe &internalBlockDescribe);

    void parseIndexBlock(
            const BlockInfo &blockInfo,
            const Distributions &distribution,
            const Internal::BlockDescribe &internalBlockDescribe);

    void calMeshVerts();
private:
    string streamPath_;

    uint16_t headerLength_;

    Header header_;

    BlockIndex blockIndex_;

    vector<Internal::BlockDescribe> blockDescribeVec_;

    Distributions Distributions_;

    vector<PointCloudXYZI::Ptr> extractedPointcloudLayers_;

    Vector2f offset_ = Vector2f::Zero();

    boost::shared_ptr<LocalCartesian> proj_;

    float highResolution = 0;

    float lowResolution = 0;

    float blockResolution = 0;

    boost::shared_ptr<GridLayer> meshLayer_;

    ulong meshId_;

    uint32_t meshLevel_;

    Vector2d meshVerts_[4];
};

#endif // STREAMINFO_H
