#include "streaminfo_restore.h"
#include <google/protobuf/io/coded_stream.h>
#include <google/protobuf/io/zero_copy_stream_impl_lite.h>

StreamInfoLoad::StreamInfoLoad(const string &streamPath)
    :streamPath_(streamPath)
{

}

bool StreamInfoLoad::load()
{
    ifstream infile(streamPath_, ios::binary);
    if(!infile)
    {
        LOG(WARNING) << streamPath_ << " open error!";
        return false;
    }
    char buf[2];
    infile.read(buf, 2);
    headerLength_ = ((uint16_t)buf[1] << 8) + (uint16_t)buf[0];
    LOG(INFO) << "headerLength: " << headerLength_;

    char* headerData = (char*)malloc(headerLength_);
    infile.read(headerData, headerLength_);
    header_.ParseFromArray(headerData, headerLength_);
    free(headerData);
    meshId_ = header_.mesh_id();
    highResolution = header_.high_relolution();
    lowResolution = header_.low_relolution();
    blockResolution = header_.block_relolution();
    offset_.x() = header_.offset_x();
    offset_.y() = header_.offset_y();
    offset_.z() = header_.offset_z();
    LOG(INFO) << setprecision(15) << "offset_: " << offset_.transpose();
    origin_.x() = offset_.x();
    origin_.y() = offset_.y();

    vector<double> blockResolutions{blockResolution};
    blockManager_.reset(new GridManager(MaxMeshSideLength, blockResolutions));
    blockManager_->setOrigin(origin_);
    vector<boost::shared_ptr<GridLayer>> meshLayers =
            blockManager_->getLayers();
    meshLayer_ = meshLayers[0];
    meshLayer_->setLayerResolutions({lowResolution, highResolution});

    proj_.reset(new LocalCartesian(header_.latitude(),
                                   header_.longtitude(),
                                   header_.height()));

    uint16_t blockIndexLength = header_.block_index_length();
    LOG(INFO) << "blockIndexLength: " << blockIndexLength;
    char* blockIndexData = (char*)malloc(blockIndexLength);
    infile.read(blockIndexData, blockIndexLength);
    blockIndex_.ParseFromArray(blockIndexData, blockIndexLength);
    free(blockIndexData);
    size_t block_describe_size = blockIndex_.block_describe_size();
    for(size_t i = 0; i < block_describe_size; i++){
        BlockDescribe blockDescribe = blockIndex_.block_describe(i);
        Internal::BlockDescribe internalBlockDescribe;
        internalBlockDescribe.blockId = blockDescribe.block_id();
        internalBlockDescribe.blockOffset = blockDescribe.block_info_offset();
        internalBlockDescribe.blockSize = blockDescribe.block_info_size();
        blockDescribeVec_.push_back(internalBlockDescribe);
    }


    LOG(INFO) << "block_describe_size: " << block_describe_size;

    infile.clear();
    infile.seekg(0, infile.end);
    uint32_t totalSize = infile.tellg();
    LOG(INFO) << "infile.tellg(): " << infile.tellg();
    infile.clear();
    infile.seekg(2 + headerLength_ +
                 header_.block_index_length() + header_.block_info_length());

    uint32_t distributionSize = totalSize - 2 - headerLength_ -
            header_.block_index_length() - header_.block_info_length();

    char* distributionBuf = (char*)malloc(distributionSize);

    LOG(INFO) << "distributionSize: " << distributionSize;
    infile.read(distributionBuf, distributionSize);
    infile.close();


    ::google::protobuf::io::ArrayInputStream input(distributionBuf, distributionSize);
    ::google::protobuf::io::CodedInputStream decoder(&input);
    decoder.SetTotalBytesLimit(1024 * 1024 * 1024, 64 * 1024 * 1024);
    Distributions_.ParseFromCodedStream(&decoder);
    LOG(INFO) << "grid_distribution_size: " << Distributions_.grid_distribution_size();
    free(distributionBuf);

    extractedPointcloudLayers_.reset(new PointCloudXYZI);

//    parseOneBlock(blockDescribeVec_[100]);

    for(size_t i = 0; i < blockDescribeVec_.size(); i++){
        auto blockDescribe = blockDescribeVec_[i];
        parseOneBlock(blockDescribe);
    }

    string savePath = streamPath_.substr(0, streamPath_.find_last_of("/") + 1);
    pcl::io::savePCDFileBinary(
                savePath + "/extracted.pcd", *extractedPointcloudLayers_);

    return true;
}

void StreamInfoLoad::parseOneBlock(
        const Internal::BlockDescribe &internalBlockDescribe)
{
    ifstream infile(streamPath_, ios::binary);
    if(!infile)
    {
        LOG(WARNING) << streamPath_ << " open error!";
        return ;
    }

    char* blockInfoData = (char*)malloc(internalBlockDescribe.blockSize);
    uint32_t infoOffset = 2 + headerLength_ + header_.block_index_length() + internalBlockDescribe.blockOffset;
    infile.clear();
    infile.seekg(infoOffset);
    infile.read(blockInfoData, internalBlockDescribe.blockSize);
    BlockInfo blockInfo;
    blockInfo.ParseFromArray(blockInfoData, internalBlockDescribe.blockSize);
    free(blockInfoData);
    infile.close();

//    LOG(INFO) << "blockIndexX: " << blockIndexX;
//    LOG(INFO) << "blockIndexY: " << blockIndexY;
//    LOG(INFO) << "blockCenterX: " << blockCenterX;
//    LOG(INFO) << "blockCenterY: " << blockCenterY;

    if(blockInfo.bit_map_size()){
        *extractedPointcloudLayers_ +=
                *parseBitmapBlock(blockInfo,
                                  Distributions_,
                                  internalBlockDescribe);
    }else{
        *extractedPointcloudLayers_ +=
                *parseIndexBlock(blockInfo,
                                  Distributions_,
                                  internalBlockDescribe);

    }
}

vector<uint32_t> StreamInfoLoad::parseBitMap(
        const ProtoRepeatedUint32 &bitmap,
        const uint32_t size)
{
    vector<uint32_t> activeGridIndexes;
    uint32_t activeGridIndex = 0;
    for(size_t i = 0; i < size; i++){
        uint8_t byte = bitmap.Get(i);
        for(size_t byteCnt = 0; byteCnt < 8; byteCnt++){
            if((byte >> byteCnt) & 1){
                activeGridIndexes.push_back(activeGridIndex);
            }
            activeGridIndex++;
        }
    }
//    LOG(INFO) << "activeGridIndexes.size(): " << activeGridIndexes.size();
    return activeGridIndexes;
}

void StreamInfoLoad::Forward(
        double lat, double lon, double h, double &x, double &y, double &z)
{
    proj_->Forward(lat, lon, h, x, y, z);
}

void StreamInfoLoad::Reverse(
        double x, double y, double z, double &lat, double &lon, double &h)
{
    proj_->Reverse(x, y, z, lat, lon, h);

}

float StreamInfoLoad::getHighResolution()
{
    return highResolution;
}

float StreamInfoLoad::getLowResolution()
{
    return lowResolution;
}

float StreamInfoLoad::getBlockResolution()
{
    return blockResolution;
}

boost::shared_ptr<GridManager> StreamInfoLoad::getBlockManager()
{
    return blockManager_;
}

PointCloudXYZI::Ptr StreamInfoLoad::parseBitmapBlock(
        const BlockInfo &blockInfo,
        const Distributions &distribution,
        const Internal::BlockDescribe &internalBlockDescribe)
{
    PointCloudXYZI::Ptr blockPointcloud =
            PointCloudXYZI::Ptr(new PointCloudXYZI);
    float gridResolution = blockInfo.grid_resolution();
    int32_t blockPerMeshItemCnt = round(MaxMeshSideLength / blockResolution);
    int32_t blockId = internalBlockDescribe.blockId;
    int32_t shiftBlockId = blockId +
            (blockPerMeshItemCnt / 2) * (blockPerMeshItemCnt + 1);
    int32_t shiftBlockIndexX = floor((float)shiftBlockId / blockPerMeshItemCnt);
    int32_t shiftBlockIndexY = shiftBlockId % blockPerMeshItemCnt;
    int32_t blockIndexX = shiftBlockIndexX - blockPerMeshItemCnt / 2;
    int32_t blockIndexY = shiftBlockIndexY - blockPerMeshItemCnt / 2;
    float blockCenterX = (float)(blockIndexX + 0.5) * blockResolution;
    float blockCenterY = (float)(blockIndexY + 0.5) * blockResolution;

//    if(blockIndexX != 22 || blockIndexY != 13){
//        return blockPointcloud;
//    }
    vector<uint32_t> activeGridIndexes =
            parseBitMap(blockInfo.bit_map(), blockInfo.bit_map_size());

    BlockGrid* block = reinterpret_cast<BlockGrid*>(
                meshLayer_->getBlock({blockIndexX, blockIndexY}).get());

    vector<boost::shared_ptr<GridLayer>> blockLayers =
            block->getLayers();
    boost::shared_ptr<GridLayer> currBlockLayer;
//    if(blockLayer.grid_resolution() == highResolution){
//        currBlockLayer = blockLayers[1];
//    }else{
//        currBlockLayer = blockLayers[0];
//    }
    currBlockLayer = blockLayers[0];

    int32_t gridPerBlockItemCnt = round(blockResolution / gridResolution);
    for(size_t i = 0; i < activeGridIndexes.size(); i++){
        auto index = activeGridIndexes[i];
        auto grid_feature_index = blockInfo.grid_feature_index(i);
        GridDistribution gridDistribution = distribution.grid_distribution(grid_feature_index);

//        int32_t shiftGridId = index +
//                (gridPerBlockItemCnt / 2) * (gridPerBlockItemCnt + 1);
        int32_t shiftGridIndexX = floor((float)index / gridPerBlockItemCnt);
        int32_t shiftGridIndexY = index % gridPerBlockItemCnt;
        int32_t gridIndexX = shiftGridIndexX - gridPerBlockItemCnt / 2;
        int32_t gridIndexY = shiftGridIndexY - gridPerBlockItemCnt / 2;

        Grid* grid = reinterpret_cast<Grid*>(
                    currBlockLayer->getGrid({gridIndexX, gridIndexY}).get());

        size_t oneDistributionCnt = gridDistribution.one_distribution_size();
        float centerX = ((float)gridIndexX + 0.5) * gridResolution;
        float centerY = ((float)gridIndexY + 0.5) * gridResolution;
        locate::GridDistribution gridDistribution_;
        for(size_t i = 0; i < oneDistributionCnt; i++){
            OneDistribution oneDistribution = gridDistribution.one_distribution(i);
            locate::OneDistribution oneDistribution_;
            oneDistribution_.z_delta_level = oneDistribution.z_delta_leval();
            oneDistribution_.floor = oneDistribution.floor();
            oneDistribution_.intensity = oneDistribution.intensity();
            for(int32_t j = 0;
                j < oneDistribution.z_delta_leval();
                j++){
                int32_t indexZ = j + oneDistribution.floor();
                pcl::PointXYZI point;
                PointExport pointExport;
                point.x = centerX + blockCenterX + offset_.x();
                point.y = centerY + blockCenterY + offset_.y();
                point.z = ((float)indexZ + 0.5) * ZResolution + offset_.z();
                point.intensity = oneDistribution.intensity() * 8;
                blockPointcloud->push_back(point);

                pointExport.x = centerX + blockCenterX + offset_.x();
                pointExport.y = centerY + blockCenterY + offset_.y();
                pointExport.z = ((float)indexZ + 0.5) * ZResolution;
                pointExport.intensity = oneDistribution.intensity() * 8;
                grid->addPoint(pointExport);
            }
            gridDistribution_.oneDistributions.push_back(oneDistribution_);
        }
        grid->setGridDistribution(gridDistribution_);

    }

    return blockPointcloud;
}

PointCloudXYZI::Ptr StreamInfoLoad::parseIndexBlock(
        const BlockInfo &blockInfo,
        const Distributions &distribution,
        const Internal::BlockDescribe &internalBlockDescribe)
{
    PointCloudXYZI::Ptr blockPointcloud =
            PointCloudXYZI::Ptr(new PointCloudXYZI);
    float gridResolution = blockInfo.grid_resolution();
    int32_t blockId = internalBlockDescribe.blockId;
    int32_t blockPerMeshItemCnt = round(MaxMeshSideLength / blockResolution);
    int32_t shiftBlockId = blockId +
            (blockPerMeshItemCnt / 2) * (blockPerMeshItemCnt + 1);
    int32_t shiftBlockIndexX = floor((float)shiftBlockId / blockPerMeshItemCnt);
    int32_t shiftBlockIndexY = shiftBlockId % blockPerMeshItemCnt;
    int32_t blockIndexX = shiftBlockIndexX - blockPerMeshItemCnt / 2;
    int32_t blockIndexY = shiftBlockIndexY - blockPerMeshItemCnt / 2;
    float blockCenterX = (float)(blockIndexX + 0.5) * blockResolution;
    float blockCenterY = (float)(blockIndexY + 0.5) * blockResolution;

    BlockGrid* block = reinterpret_cast<BlockGrid*>(
                meshLayer_->getBlock({blockIndexX, blockIndexY}).get());
    vector<boost::shared_ptr<GridLayer>> blockLayers =
            block->getLayers();
    boost::shared_ptr<GridLayer> currBlockLayer;
    if(blockInfo.grid_resolution() == highResolution){
        currBlockLayer = blockLayers[1];
    }else{
        currBlockLayer = blockLayers[0];
    }
    for(size_t i = 0; i < blockInfo.grid_feature_index_size(); i++){
        auto grid_feature_index = blockInfo.grid_feature_index(i);
        if(grid_feature_index == 0){
            continue;
        }
        grid_feature_index--;

        GridDistribution gridDistribution = distribution.grid_distribution(grid_feature_index);
        uint32_t indexX = i / (blockResolution / gridResolution);
        uint32_t indexY = i % (uint32_t)(blockResolution / gridResolution);

        size_t oneDistributionCnt = gridDistribution.one_distribution_size();
        Grid* grid = reinterpret_cast<Grid*>(
                    currBlockLayer->getGrid({indexX, indexY}).get());
        float centerX = ((float)indexX + 0.5) * gridResolution - blockResolution / 2;
        float centerY = ((float)indexY + 0.5) * gridResolution - blockResolution / 2;
        locate::GridDistribution gridDistribution_;
        for(size_t i = 0; i < oneDistributionCnt; i++){
            OneDistribution oneDistribution = gridDistribution.one_distribution(i);
            locate::OneDistribution oneDistribution_;
            oneDistribution_.z_delta_level = oneDistribution.z_delta_leval();
            oneDistribution_.floor = oneDistribution.floor();
            oneDistribution_.intensity = oneDistribution.intensity();
            for(int32_t j = 0;
                j < oneDistribution.z_delta_leval();
                j++){
                int32_t indexZ = j + oneDistribution.floor();
                pcl::PointXYZI point;
                PointExport pointExport;
                point.x = centerX + blockCenterX/* + offset_.x()*/;
                point.y = centerY + blockCenterY/* + offset_.y()*/;
                point.z = ((float)indexZ + 0.5) * ZResolution;
                point.intensity = oneDistribution.intensity() * 8;
                blockPointcloud->push_back(point);

                pointExport.x = centerX + blockCenterX + offset_.x();
                pointExport.y = centerY + blockCenterY + offset_.y();
                pointExport.z = ((float)indexZ + 0.5) * ZResolution;
                pointExport.intensity = oneDistribution.intensity() * 8;
                grid->addPoint(pointExport);
            }
            grid->setGridDistribution(gridDistribution_);
        }
    }
    return blockPointcloud;
}

string StreamInfoLoad::getMeshId()
{
    return meshId_;
}

Vector2f StreamInfoLoad::getOrigin()
{
    return origin_;
}

Vector3f StreamInfoLoad::getOffset()
{
    return offset_;
}
