#include "streammanager.h"
using namespace boost::filesystem;

StreamManager::StreamManager()
{

}

bool StreamManager::SetMapPath(
        std::string path_name){
    if(exists(path_name)){
        streamDir = path_name;
        return true;
    }
    return false;
}



void StreamManager::checkAndLoad(ulong mesh)
{
    vector<ulong> loadingMashVec;
    {
        boost::shared_lock<boost::shared_mutex> readLock(loadingMashMtx_);
        loadingMashVec = loadingMashVec_;
    }
    auto loadingIter = find(loadingMashVec.begin(),
                     loadingMashVec.end(),
                     mesh);
    if(loadingIter != loadingMashVec.end()){
        return;     //is loading
    }
    {
        boost::unique_lock<boost::shared_mutex> writeLock(loadingMashMtx_);
        loadingMashVec_.push_back(mesh);
    }
    loadMesh(mesh);
}

bool StreamManager::loadMesh(ulong mesh)
{
    if(IsMeshLoaded(mesh)){
        return true;
    }
    string meshStreamPath = streamDir + "/" + to_string(mesh) + ".stream";
    if(!exists(meshStreamPath)){
        return false;
    }
    boost::shared_ptr<StreamInfo> meshStream =
            boost::shared_ptr<StreamInfo>(new StreamInfo(meshStreamPath));
    auto start = system_clock::now();
    bool loadSuccess = meshStream->load();
    auto end = system_clock::now();
    auto duration = duration_cast<milliseconds>(end - start);
    LOG(INFO) << "load " << mesh << " take "
              << double(duration.count()) / 1000 << " seconds";
    if(!loadSuccess){
        return false;
    }
    {
        boost::unique_lock<boost::shared_mutex> writeLock(meshStreaMtx_);
        meshStreamMap_.insert(make_pair(mesh, meshStream));
    }
    {
        boost::unique_lock<boost::shared_mutex> writeLock(loadingMashMtx_);
        auto loadingIter = find(loadingMashVec_.begin(),
                         loadingMashVec_.end(),
                         mesh);
        if(loadingIter != loadingMashVec_.end()){
            loadingMashVec_.erase(loadingIter);
        }
    }
    if(!fstMeshLoaded){
        highResolution_ = meshStream->getHighResolution();
        lowResolution_ = meshStream->getLowResolution();
        blockResolution_ = meshStream->getBlockResolution();
        blockCntPerMeshSide_ = MaxMeshSideLength / blockResolution_;
        highGridCntPerBlockSide_ = blockResolution_ / highResolution_;
        lowGridCntPerBlockSide_ = blockResolution_ / lowResolution_;
        fstMeshLoaded = true;
    }
    return true;
}

void StreamManager::checkMeshBound(
        const Vector3d currPos)
{
    checkNeighborToLoad(currPos);
    checkNeighborToFree(currPos);
}

vector<ulong> StreamManager::getNeighborsByMask(
        ulong currMeshId, uint8_t mask)
{
    bool nearLeft = false;
    bool nearUp = false;
    bool nearRight = false;
    bool nearDown = false;
    if ((mask & MESH_LEFT) != 0)
        nearLeft = true;
    if ((mask & MESH_UP) != 0)
        nearUp = true;
    if ((mask & MESH_RIGHT) != 0)
        nearRight = true;
    if ((mask & MESH_DOWN) != 0)
        nearDown = true;

    vector<ulong> neighborMesh;
    if(nearLeft){
        neighborMesh.push_back(
                    GetNeighborGeoHash(currMeshId, NeighborType::Left));
        if(nearUp){
            neighborMesh.push_back(
                        GetNeighborGeoHash(currMeshId, NeighborType::LeftUp));
        }else if(nearDown){
            neighborMesh.push_back(
                        GetNeighborGeoHash(currMeshId, NeighborType::LeftDown));
        }
    }
    if(nearUp){
        neighborMesh.push_back(
                    GetNeighborGeoHash(currMeshId, NeighborType::Up));
        if(nearRight){
            neighborMesh.push_back(
                        GetNeighborGeoHash(currMeshId, NeighborType::RightUp));
        }
    }
    if(nearRight){
        neighborMesh.push_back(
                    GetNeighborGeoHash(currMeshId, NeighborType::Right));
        if(nearDown){
            neighborMesh.push_back(
                        GetNeighborGeoHash(currMeshId, NeighborType::RightDown));
        }
    }
    if(nearDown){
        neighborMesh.push_back(
                    GetNeighborGeoHash(currMeshId, NeighborType::Down));
    }
    return neighborMesh;
}

void StreamManager::checkNeighborToLoad(
        const Vector3d currPos)
{
    boost::shared_ptr<StreamInfo> currMeshStream;
    {
        boost::shared_lock<boost::shared_mutex> readLock(currMashMtx_);
        currMeshStream = currMeshStream_;
    }
    int8_t neighborMask = currMeshStream->checkNearMeshBound(currPos);
    ulong currMeshId = currMeshStream->getMeshId();
    vector<ulong> neighborMesh = getNeighborsByMask(currMeshId, neighborMask);

    map<ulong, boost::shared_ptr<StreamInfo>> meshStreamMap;
    {
        boost::shared_lock<boost::shared_mutex> readLock(meshStreaMtx_);
        meshStreamMap = meshStreamMap_;
    }
    vector<ulong> neighborMeshToLoad;
    for(ulong neighbor : neighborMesh){
        auto iter = meshStreamMap.find(neighbor);
        if(iter == meshStreamMap.end()){
            neighborMeshToLoad.push_back(neighbor);
            LOG(INFO) << neighbor <<" to load";
        }
    }
    if(neighborMeshToLoad.size() == 0){
        return;
    }
    if(loadingNeighbors){
        return;
    }
    loadingNeighbors = true;
    for(ulong neighbor : neighborMeshToLoad){
        loadMesh(neighbor);
    }
    loadingNeighbors = false;
}

void StreamManager::checkNeighborToFree(
        const Vector3d currPos)
{
    boost::shared_ptr<StreamInfo> currMeshStream;
    {
        boost::shared_lock<boost::shared_mutex> readLock(currMashMtx_);
        currMeshStream = currMeshStream_;
    }
    int8_t neighborMask = currMeshStream->checkAwayMeshBound(currPos);
    ulong currMeshId = currMeshStream->getMeshId();
    vector<ulong> neighborMesh = getNeighborsByMask(currMeshId, neighborMask);

    map<ulong, boost::shared_ptr<StreamInfo>> meshStreamMap;
    {
        boost::shared_lock<boost::shared_mutex> readLock(meshStreaMtx_);
        meshStreamMap = meshStreamMap_;
    }
    vector<ulong> neighborMeshToFree;
    for(ulong neighbor : neighborMesh){
        auto iter = meshStreamMap.find(neighbor);
        if(iter != meshStreamMap.end()){
            neighborMeshToFree.push_back(neighbor);
        }
    }
    if(neighborMeshToFree.size() == 0){
        return;
    }
    if(freeingNeighbors){
        return;
    }
    freeingNeighbors = true;
    for(ulong neighbor : neighborMeshToFree){
        map<ulong, boost::shared_ptr<StreamInfo>>::iterator iter;
        bool isIn = false;
        {
            boost::shared_lock<boost::shared_mutex> readLock(meshStreaMtx_);
            auto iter = meshStreamMap_.find(neighbor);
            isIn = iter != meshStreamMap_.end();
            currMeshStream = iter->second;
        }
        if(isIn){
            {
                boost::unique_lock<boost::shared_mutex> writeLock(meshStreaMtx_);
                meshStreamMap_.erase(iter);
            }
            currMeshStream.reset();
        }
    }
    freeingNeighbors = false;
}

bool StreamManager::getNeighborTypeByIndex(
        const juefx::GridIndex::SByteId &index, NeighborType &neighborType) const
{
    if(index.x == -1){
        if(index.y == -1){
            neighborType = LeftDown;
        }else if(index.y == 0){
            neighborType = Left;
        }else if(index.y == 1){
            neighborType = LeftUp;
        }else{
            return false;
        }
    }else if(index.x == 0){
        if(index.y == -1){
            neighborType = Down;
        }else if(index.y == 1){
            neighborType = Up;
        }else{
            return false;
        }
    }else if(index.x == 1){
        if(index.y == -1){
            neighborType = RightDown;
        }else if(index.y == 0){
            neighborType = Right;
        }else if(index.y == 1){
            neighborType = RightUp;
        }else{
            return false;
        }
    }else{
        return false;
    }
    return true;
}

bool StreamManager::getCurrMeshGridResolution(
        uint8_t type, float &resolution) const
{
    if(type == 0){
        resolution = currMeshStream_->getHighResolution();
    }else if(type == 1){
        resolution = currMeshStream_->getLowResolution();
    }else{
        return false;
    }
    if(resolution == 0){
        LOG(WARNING) << "gridResolution == 0!";
        return false;
    }
    return true;
}


float StreamManager::getHighResolution() const
{
    return highResolution_;
}

float StreamManager::getLowResolution() const
{
    return lowResolution_;
}

float StreamManager::getBlockResolution() const
{
    return blockResolution_;
}
